#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#

;; mode is one of:
;;	'tail	value is to be returned (ie, expr is in tail position)
;;	'value	value is needed, but not in tail position
;;	'effect	value is not needed
;;	'top 	top-level form

(define (untop mode)
    (if (eq? mode 'top)
	'tail
	mode))

(define (compile expr lexical-envt dynamic-envt mode)
    (cond ((symbol? expr) (compile/symbol expr lexical-envt dynamic-envt mode))
    	  ((pair? expr) (compile/list expr lexical-envt dynamic-envt mode))
	  ((self-evaluating? expr) (make-const expr mode))
	  (#t (error/syntax "Unrecognized expression form " expr))))
    
(define (self-evaluating? expr)
    (or (number? expr)
	(char? expr)
	(string? expr)
	(boolean? expr)
	(null? expr)))

(define (find envt name)
    (let ((bdg-info (lookup envt name)))
	(if bdg-info
	    bdg-info
	    (let ((v (make <TopLevelVar> name: name)))
	        (if (fluid-ref *warn-unbound-var-ref* #t)
		    (warning "reference to unbound variable ~a" name))
		(bind! (the-top-level envt) v)
		v))))

(define (compile/symbol symbol lexical-envt dynamic-envt mode)
    (compile-ref (find lexical-envt symbol) lexical-envt dynamic-envt mode))

(define (compile/list expr lex-envt dyn-envt mode)
    (if (symbol? (car expr))
	(compile-head (find lex-envt (car expr)) 
		      expr 
		      lex-envt 
		      dyn-envt 
		      mode)
	(compile/combo expr lex-envt dyn-envt mode)))

;; This function compiles a form that is known to be a
;; combination.  It optimizes ((lambda (formals) body) args)
;; into (let (formals args) body)

(define (compile/combo form lex-envt dyn-envt mode)
    (let ((fn (compile (car form) lex-envt dyn-envt 'value))
    	  (args (compile/multi (cdr form) lex-envt dyn-envt 'value)))
	(if (eq? (class-of fn) <ic-lambda>)
	    ;; re-compile the LAMBDA as a LET
	    ;; (we have to recompile, because the body was
	    ;;  originally compiled in TAIL mode, which it
	    ;;  will no longer be in)
	    (let ((formals (cadar form))
	    	  (body (cddar form)))
		(if (and (pair? formals) 
			 (eq? (car formals) 'quote))
		    (begin
			(set! formals (car body))
			(set! body (cdr body))))
		(let ((new-envt (make-lexical-envt 
					(make-lex-vars formals lex-envt)
					lex-envt 
					dyn-envt)))
		    (make <ic-bind>
			    envt: new-envt
			    inits: args
			    rest?: (compute-has-rest formals)
			    body: (compile/body body 
			    			new-envt 
						new-envt 
						(untop mode))))) 	   
	    ;; otherwise, make it a normal call
	    (make <ic-call> 
		    function: fn
		    args: args
		    mode: mode))))

(define (compile/multi items lex-envt dyn-envt mode)
    (let loop ((i items) (r '()))
	(cond 
	    ((symbol? i) (let ((binding (lookup lex-envt i)))
			    (if (substitution? binding)
				(begin
				    (set! lex-envt (envt binding))
				    (loop (expr binding) r))
				(error/syntax
				    "Rest symbol ~a not a substitution" i))))
	    ((null? i) (let ((args (reverse r)))
			(make <ic-multi> arg-list: args
					 mode: mode)))
	    ((pair? i) (loop (cdr i)
			     (cons (compile (car i) lex-envt dyn-envt 'value)
				   r)))
	    (#t (error/syntax "Illegal args form: ~a" items)))))

(define (make-no-values mode)
    (make <ic-multi> arg-list: '() mode: mode))

(define (compile/body body lex-envt dyn-envt mode)

    (define (rewrite-as-letrec defines stmts)
	(cons* 'letrec
	       (map
		  (lambda (def)
		    (if (pair? (cadr def))
			
			; (define (x y) ...)
			(list
			    (caadr def)
			    (cons* 'lambda
				    (cdadr def)
				    (cddr def)))

			; (define x ...)
			(cdr def)))
		  defines)
	       stmts))
    (define (compile-stmts stmts)
	(let loop ((i stmts) (r '()))
	    (if (null? i)
		(cond 
		    ((null? r) 
			(make-no-values mode))
		    ((null? (cdr r)) 
			(car r))
		    (else 
			(make <ic-seq> 
				    return-types: (return-types (car r))
				    stmt-list: (reverse r)
				    mode: mode)))
		(if (symbol? i)
		    (let ((bdg (lookup lex-envt i)))
			(if (substitution? bdg)
			    (begin
				(set! lex-envt (envt bdg))
				(loop (expr bdg) r))
			    (error/syntax
				"Rest symbol ~a not a substitution" i)))
		    (loop (cdr i)
			    (cons (compile (car i) 
					    lex-envt
					    dyn-envt
					    (if (null? (cdr i)) 
						(untop mode) 
						'effect))
				r))))))
			    
    ;;  Note -- Due to the way we are doing macros, I think there
    ;;  is something slightly broken:  You can't define a macro
    ;;  which contains a body which accepts internal defines!
    ;;  Solution:  We should search for internal defines AS we process
    ;;  the items in the sequence.  OR, we should make two passes
    ;;  over the sequence, descending into substitutions looking for
    ;;  defines.
    
    ;;  filters out the defines from a sequence and returns
    ;;  a pair of the list of defines and the other statements
    
    (define (filter-internal-defines expr-list)
	(let loop ((define-list '()) 
		    (stmt-list '())
		    (last-stmt #f)
		    (exprs expr-list))
	    (if (pair? exprs)
		(if (and (pair? (car exprs)) (eq? (caar exprs) 'define))
		    (loop (cons (car exprs) define-list)
			    stmt-list
			    last-stmt
			    (cdr exprs))
		    (let ((last (cons (car exprs) '())))
			(if last-stmt
			    (set-cdr! last-stmt last))
			(loop define-list
				(if last-stmt stmt-list last)
				last
				(cdr exprs))))
		(begin
		    (if last-stmt
			(begin
			    (set-cdr! last-stmt exprs)
			    (cons (reverse define-list) stmt-list))
			(cons (reverse define-list) exprs))))))

    (let ((parts (filter-internal-defines body)))
	(if (null? (car parts))
	    (compile-stmts (cdr parts))
	    (compile/letrec #f 
			    (rewrite-as-letrec (car parts) (cdr parts))
			    lex-envt
			    dyn-envt
			    mode))))




;; this compiles all lambda's, define's, etc.
;; name is any literal object that will be associated with the
;; procedure for debugging purposes

(define (current-procedure-name)
    (let ((bdg (%fluid-assq '*procedure-name*)))
	(if bdg
	    (cdr bdg)
	    #f)))

(define (current-place-name)
    (let loop ((s (%dynamic-state-ref)) (r '()))
	(if (pair? s)
	    (if (and (pair? (car s))
		     (eq? (caar s) '*place*))
		(loop (cdr s) (cons (cdar s) r))
		(loop (cdr s) r))
	    r)))

(define (compile/procedure name formals-spec body lex-envt dyn-envt 
			   . see-shadows)
    (fluid-let ((*procedure-name* name)
    		(*place* name))
	(let ((new-envt (make-lexical-envt 
			    (make-lex-vars formals-spec lex-envt)
			    lex-envt 
			    dyn-envt))
	      (old-seeing-shadows *compile-seeing-shadows*)
	      (icode #f))
	    (if (pair? see-shadows)
		(set! *compile-seeing-shadows* (car see-shadows)))
	    (set! icode 
		(make <ic-procedure>
		    name: name
		    body: (make <ic-bind>
				envt: new-envt
				inits: #f	;; special
				rest?: (compute-has-rest formals-spec)
				body: (compile/body body 
						    new-envt 
						    new-envt 
						    'tail))))
	    (set! *compile-seeing-shadows* old-seeing-shadows)
	    icode)))

(define (set-compile-seeing-shadows! v)
    (set! *compile-seeing-shadows* v))

(define (set-compile-optimized-loops! v)
    (set! *compile-optimized-loops* v))

		
