#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;; DEFINERs are special kinds of SPECIAL FORMs, with the
;; property that they create new bindings for top-level
;; variables (or other kinds of objects, like macros)
;; A given binding can only have one definer for it in
;; all of input.  If a binding is not marked write-protected,
;; then it can have it value changed, of course...

(define (car-most form)
    (if (pair? form)
	(car-most (car form))
	form))

;; general-purpose function to grab a binding
;; for a `define'-type construct (which
;; may involve the allocation of a root slot).

;; recognizes things like:
;;    (define [root: { "name" | ("name" module slot) }] foo ...
;;    (define [root: { "name" | ("name" module slot) }] (foo arg ...)
;;    (define [root: { "name" | ("name" module slot) }] ((foo bar) arg...))
;; returns a cons whose 
;;    car is the binding
;;    cdr is the rest of the form

(define (grab-binding lex-envt dyn-envt form)
    (let ((rest-of-form (cdr form))
	  (bdg #f))
      (if (and (pair? rest-of-form)
	       (eq? (car rest-of-form) 'root:))
	  (let ((desc (cadr rest-of-form))
		(bdg-name (car-most (cddr rest-of-form))))
	    (cons
	     (if (pair? desc)
		(let ((c-name (car desc))
		      (module (cadr desc))
		      (slot (caddr desc)))
		  (grab-root-binding bdg-name c-name module slot))
		(grab-root-binding bdg-name
				   desc 
				   (name *current-module*) 
				   (alloc-root-slot desc bdg-name)))
	     (cddr rest-of-form)))
	  (cons (grab-tlv-binding (car-most rest-of-form))
		rest-of-form)))

    (define (check-owner-type bdg class)
      (let ((old-owner (table-insert! *defined-by* 
					  bdg 
					  *current-source-file*)))
	(if (and bdg (not (instance? bdg class)))
	    (error/semantic "~a is a ~a, tried to redefine as a ~a"
			    (name bdg)
			    (class-of bdg)
			    class))
	(if (and old-owner
		 (not (eq? old-owner *current-source-file*)))
	    (warning "~a originally defined by ~a, now by ~a"
		     (name bdg)
		     old-owner
		     *current-source-file*)))
      bdg)

    (define (register-self-owner bdg)
      (table-insert! *defined-by* bdg *current-source-file*))

    (define (grab-root-binding name c-name module slot)
      (let ((bdg (lookup (the-top-level lex-envt) name)))
	 (if bdg
	     (begin
	       (check-owner-type bdg <RootVar>)
	       (set-c-name! bdg c-name)
	       (set-module! bdg module)
	       (set-slot! bdg slot)
	       bdg)
	     (let ((bdg (make <RootVar> name: name
			                c-name: c-name
					module: module
					slot: slot)))
	       (bind! (the-top-level lex-envt) bdg)
	       (register-self-owner bdg)
	       bdg))))

    (define (grab-tlv-binding name)
      (let ((bdg (lookup (the-top-level lex-envt) name)))
	 (if bdg
	     (begin
	       (check-owner-type bdg <TopLevelVar>)
	       bdg)
	     (let ((bdg (make <TopLevelVar> name: name)))
	       (bind! (the-top-level lex-envt) bdg)
	       (register-self-owner bdg)
	       bdg)))))


(define (c-compile/define sf form lex-envt dyn-envt mode)
    (if (not (eq? mode 'top))
	(error/syntax "define not at top-level: ~a" form))
    (let* ((bdg&rest (grab-binding lex-envt dyn-envt form))
	   (bdg (car bdg&rest))
	   (form (cdr bdg&rest)))
      ;(if (instance? (car bdg&rest) <TopLevelVar>)
      ;	  (set-value! (car bdg&rest) (cons '#uninit (cdr bdg&rest))))
      (if (pair? (car form))
	  (c-compile/define-fn bdg form lex-envt dyn-envt)
	  (c-compile/define-var bdg form lex-envt dyn-envt)))
    #f)	;; return #f ==> cause eval to ignore this form

; note that you can say something like
;   (define ((foo bar) x y) ...)
; to make a function of two arguments with a compound
; name.  `bdg' has already been bound to the tlv foo
; but we will compute name as '(foo bar)
;
; (at this point, form starts with the name-arg list)

(define (deferred-value name formals body lex-envt dyn-envt)
    (delay
	(xwrap-procedure 
	    (compile/procedure name 
				formals
				body
				lex-envt
				dyn-envt
				#t))))
				
(define (c-compile/define-fn bdg form lex-envt dyn-envt)
    (let ((name (caar form))
	  (formals (cdar form))
	  (body (cdr form)))
	(ct-format #t "compiling function: ~a\n" name)
	
	(set-value! bdg 
	    (make-closure 
		'() 
		(compile-template name formals body lex-envt dyn-envt)))
	(if (instance? bdg <RootVar>)
	    (add-load-cmd-set-var bdg (value bdg)))
					
	;; compile to CTC
	;; (This a lazy process, because at
	;;  compile time the vast majority of CTC is never
	;;  executed)
	
	; note:  I think we should be mapping on the templates
	; instead of the closures
	
	(push! *shadows* 
	    (cons (value bdg)
		  (deferred-value name formals body lex-envt dyn-envt)))))

; %code is sort of like lambda, but it creates
; a template (not a closure), and always generates
; compiled code
; (when a lambda occurs in a top-level form,
;  it is closure-threaded code)

(define (c-compile/%code sf form lex-envt dyn-envt mode)
    (let ((name (cadr form))
	  (formals (caddr form))
	  (body (cdddr form)))
	
	;; we would like to compile to CTC also, but
	;; that is a mapping for closures, not templates
	;; (maybe I should have made %code generate closures
	;;  instead!  But I wanted the ability to hand-craft
	;;  closures (ie, hacking environments).  On the other
	;;  hand, it's rarely useful to hand-hack environments
	;;  of automatically-compiled code, because you can't
	;;  set up the c-t envt anyway!)
	(make-const 
	    (compile-template name formals body lex-envt dyn-envt) 
	    mode)))

;; when defining a top-level var,
;; there are several possibilities:
;;   1.  The value can be COMPUTED and STORED in the TLV,
;;	 to be loaded in the natural way
;;	  appropriate for:
;;		(define *foo* (vector 1 2 3))
;;   2.  The value can be set to #unint, and a procedure
;;	 emitted (XCTC) that will set the correct value after
;;	 loading.
;;	  appropriate for:
;;		(define *console-input* (%CIP-make))
;; (2) always works, but (1) is much more efficient.
;; On the other hand, it turns out (1) is more useful for
;; compile-time operations, so that's what we're doing.

;; note also that we SIDE-EFFECT the old value of the
;; binding if it is the same size and class, and the
;; class is COPY-SAFE? (currently, only class <<Class>>
;; is copy-safe?)

(define (copy-safe? x)
    (eq? x <<Class>>))

(define (ct-picky-set bdg new-value)
    (let ((old-value (value bdg)))
      (if (and (eq? (class-of old-value) (class-of new-value))
	       (gvec? old-value)
	       (gvec? new-value)   ;; this SHOULD follow from above two
	       (eq? (%gvec-length new-value) (%gvec-length old-value)))
	  (if (copy-safe? (class-of new-value))
	      (let loop ((i (%gvec-length new-value)))
		(if (not (eq? i 0))
		    (begin
		      (%gvec-write old-value
				   (sub1 i)
				   (%gvec-read new-value (sub1 i)))
		      (loop (sub1 i)))))
	      (set-value! bdg new-value))
	  (begin
	    (if (and (gvec? old-value)
		     (gvec? new-value)
		     (not (and (pair? old-value)
			       (eq? (car old-value) '#uninit))))
		(warning "~a changed: was a ~a (size ~d), now a ~a (size ~d)"
			 (name bdg)
			 (class-name (class-of old-value))
			 (%gvec-length old-value)
			 (class-name (class-of new-value))
			 (%gvec-length new-value))
		(if (and (not (eq? (class-of old-value) (class-of new-value)))
			 (not (and (pair? old-value)
				   (eq? (car old-value) '#uninit)))
			 (not (eq? old-value '#unbound)))
		    (warning "~a changed: was a ~a, now a ~a"
			     (name bdg)
			     (class-name (class-of old-value))
			     (class-name (class-of new-value)))))
	    (set-value! bdg new-value)))))

(define (c-compile/define-var bdg form lex-envt dyn-envt)
  (ct-format #t "compiling declaration: ~a\n" (car form))
  (if (pair? (cdr form))
      (let ((name (car form))
	    (expr (cadr form)))
	(ct-picky-set bdg (eval expr lex-envt))
	(if (instance? bdg <RootVar>)
	    (add-load-cmd-set-var bdg (value bdg))))
      (if (not (instance? bdg <RootVar>))
	  (error/syntax "vacuous (define ~a...) not permitted" (car form)))))


(define (add-load-cmd-root-set root-bdg value)
    (push! *source-load-commands*
	    (cons (slot root-bdg) value)))

(define (add-load-cmd-set envt name value)
    (let ((tlv (ensure-tl-var (the-top-level envt) name)))
	(push! *source-load-commands* (cons tlv value))))

(define (add-load-cmd-set-var var value)
    (if (instance? var <RootVar>)
	(push! *source-load-commands* 
	    (vector
		(module-descr (cdr (assq (module var) *modules-table*)))
		(slot var)
		value))
	(if (instance? var <TopLevelVar>)
	    (push! *source-load-commands* (cons var value))
	    (error/internal "add-load-command-set-var: Invalid bdg: ~a" var))))

(define (c-compile/%define-glue-template name arg-list monotones literals)
    (template-within-part
		name
		(length monotones)
		(lambda (c-name)
		    (ct-format #t "/****** raw glue `~a' ******/\n\n" name)
		    (let loop ((i 0) (args arg-list))
			(if (pair? args)
			    (begin
				(ct-format #t "#define ~a REG~d\n"
						(car args)
						i)
				(loop (add1 i) (cdr args)))))
		    (newline)
		    (display-some-monotones
			name
			c-name
			monotones
			(lambda (m)
			    (display (car m))))
		    (for-each
			(lambda (arg)
			    (ct-format #t "#undef ~a\n" arg))
			arg-list)
		    (newline))
		literals))

;; (%template (ctcprims 1097) 0)

(define (c-compile/%template sf form lex-envt dyn-envt mode)
    (let* ((part-info (cadr form))
    	   (part-descr (make <PartDescr> 
				module-name: (symbol->string (car part-info))
				part-tag: (cadr part-info)))
	   (index (caddr form))
	   (literals '()))
	(make-const (make-template index part-descr literals) mode)))

			
(define (c-compile/%define-glue-closure name arg-list mons lits envt)
    (make-closure envt
    		  (c-compile/%define-glue-template name arg-list mons lits)))

(define (c-compile/%define-glue sf form lex-envt dyn-envt mode)
    (if (not (eq? mode 'top))
	(error/syntax "%define-glue not at top-level: ~a" form))
    (let* ((bdg&rest (grab-binding lex-envt dyn-envt form))
    	   (bdg (car bdg&rest))
	   (rest-of-form (cdr bdg&rest)))
	(if *pre-load*
	    (set-value! bdg (cons '%glue (car rest-of-form)))
	    (let ((decl (car rest-of-form))
		    (body (cdr rest-of-form))
		    (literals '())
		    (envt '()))
    
		(if (not *current-part*)
		    (new-part!))
    
		(if (eq? (car body) 'envt:)
		    (begin
			(if (not (pair? decl))
			    (error/syntax 
				"~a: `envt:' doesn't make sense for template"
				decl))
			(set! envt
			    (map
				(lambda (l)
				    (eval l (the-top-level lex-envt)))
				(cadr body)))
			(set! body (cddr body))))
    
		(if (eq? (car body) 'literals:)
		    (begin
			(set! literals
			    (map
				(lambda (l)
				    (if (and (pair? l)
						(eq? (car l) '&))
					(ensure-tl-var lex-envt
							(cadr l))
					(eval l 
						(the-top-level lex-envt))))
				(cadr body)))
			(set! body (cddr body))))
				
		(if (and (null? (cdr body))
			    (not (pair? (car body))))
		    (set! body (list 
				    (list 
				    (string-append
					(c-ify (object->string (name bdg)))
					"_"
					(number->string 
					    (length 
						(link-table 
						    *current-part*)))
					"_00")
				    (car body)))))
    
		(if (pair? decl)
		    (begin
			(ct-format #t 
				    "declaring raw glue: function ~a\n" 
				    (car decl))
			(let ((cl (c-compile/%define-glue-closure (car decl) 
								(cdr decl) 
								body 
								literals
								envt)))
			    (set-value! bdg cl)))
		    (begin
			(ct-format #t "declaring raw glue: template ~a\n" decl)
			(let ((tmpl (c-compile/%define-glue-template decl 
								    '() 
								    body 
								    literals)))
			    (set-value! bdg tmpl)))))))
    #f)

(define (c-compile/define-raw sf form lex-envt dyn-envt mode)
    (if (not (eq? mode 'top))
	(error/syntax "define-raw not at top-level: ~a" form))
    (if (not *pre-load*)
	(let ((name (cadr form))
	      (monotones (cddr form)))

	    (ct-format #t "declaring raw monotones: ~a\n" name)
	    (add-load-cmd-set 
		lex-envt
		name
		(template-within-part
			name
			(length monotones)
			(lambda (c-name)
			    (ct-format #t 
				"/** raw function `~a' **/\n" 
				name)
			    (display-some-monotones
				name
				c-name
				monotones
				(lambda (m)
				    (display (car m)))))
			'()))
	    #f)))
	    
(define (c-compile/prologue sf form lex-envt dyn-envt mode)
    (if (not (eq? mode 'top))
	(error/syntax "prologue not at top-level: ~a" form))
    (if (not *pre-load*)
	(begin
	    (set! *private-interface*
		(append *private-interface* (cdr form)))
	    #f)))

(define (c-compile/run sf form lex-envt dyn-envt mode)
    (if (not (eq? mode 'top))
	(error/syntax "run not at top-level: ~a" form))
    (if (not *pre-load*)
	(let ((icode (compile/procedure
			    'run
			    '()
			    (cdr form) 
			    lex-envt
			    dyn-envt
			    #f)))
	    (set! *source-load-commands*
		(cons
		    (xwrap-procedure icode)
		    *source-load-commands*))
	    #f)))

(define (c-compile/run-named sf form lex-envt dyn-envt mode)
    (if (not (eq? mode 'top))
	(error/syntax "run-named not at top-level: ~a" form))
    (if (not *pre-load*)
	(set! *source-load-commands*
		(cons
		    (eval (cadr form) lex-envt)
		    *source-load-commands*)))
    #f)


;; this looks for a  <TopLevelVar> or <RootVar> binding for the
;; given name.  If there is none, creates a <TopLevelVar> bdg

(define (ensure-tl-var tle name)
    (let ((bdg (lookup tle name)))
	(if bdg
	    (if (or (instance? bdg <TopLevelVar>)
	    	    (instance? bdg <RootVar>))

		; this constitutes a RE-DEFINITION, which is OK
		; (but if we are pre-loading, then it is a textually
		;  evident re-definition, so emit a warning)
		(begin
		    (if *pre-load*
			(warning "Top-level ~a `~a' is a redefinition" 
				 (class-of bdg)
				 name))
		    bdg)

		; this is an attempt to DEFINE a name that
		; is not already of the appropriate type, 
		; which is ILLEGAL
		(error/semantic "Top-level ~a is not a root or tlv" name))

	    ; this is the first definition, so make a new one
	    
	    (let ((bdg (make <TopLevelVar> name: name)))
		(bind! tle bdg)
		bdg))))

(define (include-dest-file src-f)
    (let ((dst-abs-dir (append-dirs (current-directory)
    				    (dest-dir *current-module*))))
	(make <file-name> directory: dst-abs-dir
			    filename: (filename src-f)
			    extension: (extension src-f))))

(define (file-from-to dst-f src-f)
    (make <file-name> directory: (dir-from-to (directory dst-f)  
    					      (directory src-f))
		      filename: (filename dst-f)
		      extension: (extension dst-f)))

(define (local-file name)
    (let ((f (string->file name))
    	  (srcd (append-dirs (current-directory) 
	  		     (directory (pathname *current-source-file*)))))
	(if (directory f)
	    (make <file-name>
		    directory: (append-dirs srcd (directory f))
		    filename: (filename f)
		    extension: (extension f))
	    (make <file-name>
		    directory: srcd
		    filename: (filename f)
		    extension: (extension f)))))
	
(define (copy-a-file src-f dst-f)
    (let ((src-relative (file-from-to dst-f src-f))
          (src-name (pathname->string src-f))
	  (dst-name (pathname->string dst-f)))
	(if (not (file-exists? src-name))
	    (error/semantic "File ~s not found\n" src-name))
	(ct-format #t "Copying ~a to ~a\n" src-name dst-name)
	(system (ct-format #f "rm -f ~a" dst-name))
	(if *use-soft-links*
	    (system (ct-format #f "ln -s ~a ~a"
				    (pathname->string src-relative)
				    dst-name))
	    (system (ct-format #f "cp ~a ~a"
				    src-name
				    dst-name)))))

(define (c-compile/include-c sf form lex-envt dyn-envt mode)
    (if (not (eq? mode 'top))
	(error/syntax "include-c not at top-level: ~a" form))
    (if (not *pre-load*)
	(let* ((src-f (local-file (cadr form)))
	       (dst-f (include-dest-file src-f)))
	    (copy-a-file src-f dst-f)
	    (push! *make-commands* (cons 'c-file (file-within-dir src-f)))
	    #f)))

(define (c-compile/include-h sf form lex-envt dyn-envt mode)
    (if (not (eq? mode 'top))
	(error/syntax "include-h not at top-level: ~a" form))
    (if (not *pre-load*)
	(let* ((src-f (local-file (cadr form)))
	       (dst-f (include-dest-file src-f)))
	    (copy-a-file src-f dst-f)
	    (push! *make-commands* (cons 'h-file (file-within-dir src-f)))
	    #f)))

(define (c-compile/import sf form lex-envt dyn-envt mode)
    (let* ((importing (cadr form))
	    (target (assq importing *modules-table*)))
	(if (not target)
	    (error/semantic "Module '~a' not found" importing))
	(push! *imports* importing)
	(for-each
	    (lambda (bdg)
		(let ((local-bdg (lookup lex-envt (name bdg))))
		    (if local-bdg
			(if *pre-load*
			    (warning 
				"Name ~a imported from ~a already bound"
				(name bdg)
				importing))
			(bind! (the-top-level lex-envt) 
				bdg))))
	    (map export->import 
		    (unified-contributions 
			(exports-table (cdr target))))))
    #f)

(define (c-compile/export sf form lex-envt dyn-envt mode)
    (if (not (eq? mode 'top))
	(error/syntax "export not at top-level: ~a" form))
    (set! *exports*
	   (append *exports* (map compile-1-export (cdr form))))
    #f
    (define (compile-1-export descr)
       (if (symbol? descr)
	   (let ((bdg (lookup lex-envt descr)))
	     (if (not bdg)
		 (error/syntax "Export variable (~a) not bound\n" descr))
	     (list bdg descr #f #f))
	   (let* ((name (car descr))
		  (bdg (lookup lex-envt name)))
	     (if (not bdg)
		 (error/syntax "Export variable (~a) not bound\n" name))
	     (let ((shared? #f)
		   (read-only? #f)
		   (remote-name name))
		(for-each
		    (lambda (x)
			(cond ((eq? x ':shared) (set! shared? #t))
			      ((eq? x ':read-only) (set! read-only? #t))
			      (else (set! remote-name x))))
		    (cdr descr))
		(list bdg remote-name shared? read-only?))))))
		 
; install an association into the shadow list
; e.g.:
;	(define-macro (my-car x) ...)
;	(define-shadow my-car car)

(define (make-gensym-fn)
    (let ((i 0))
	(lambda ()
	    (set! i (add1 i))
	    (string->symbol 
		(string-append "temp[" 
				(number->string i) 
				"]")))))

(define $gensym (make-gensym-fn))

(define (%splat . args)
    (display "splat: ")
    (write (car args))
    (for-each (lambda (x)
    		 (display ", ")
		 (write x))
	    (cdr args))
    (newline))

(define (%tl-ref name)
    (eval name *self*))

(define (c-compile/define-shadow sf form lex-envt dyn-envt mode)
    (let ((bdg (lookup lex-envt (cadr form)))
    	  (rhs (%tl-ref (caddr form))))
	(if bdg
	    (if (instance? bdg <Macro>)
		(set-shadow! bdg rhs)
		(if (instance? bdg <TopLevelVar>)
		    (push! *shadows* (cons (value bdg) (delay rhs)))
		    (error/semantic "cannot shadow `~a', wrong kind of bdg"
		    	            (cadr form))))
	    (error/syntax "cannot shadow `~a', not bound" (cadr form)))
	#f))		

(define (c-compile/%shadow sf form lex-envt dyn-envt mode)
    (make-const (%tl-ref (cadr form)) mode))

(define (c-compile/define-system sf form lex-envt dyn-envt mode)
    (let ((name (eval (cadr form) lex-envt))
    	  (value (eval (caddr form) lex-envt)))
	(if (not (string? name))
	    (error/syntax "define-system: Not a string: ~s" name))
	(push! *system-defs* (cons name value)))
#|
	(if (assoc name *system-defs*)
	    (set-cdr! (assoc name *system-defs*) value)
	    (push! *system-defs* (cons name value)))
|#
    #f)

(define (c-compile/snarf sf form lex-envt dyn-envt mode)
    ;; snarf some bindings directly (well, clone them)
    (for-each
	(lambda (name)
	    (if (pair? name)
		(let ((bdg (lookup *self* (car name))))
		    (if bdg
			(bind! dyn-envt bdg)
			(warning 
			    "snarf: ~a not bound in snooped environment" 
			    name)))
		(let ((bdg (lookup *self* name)))
		    (if bdg
			(bind! dyn-envt (clone bdg))
			(warning 
			    "snarf: ~a not bound in snooped environment" 
			    name)))))
	(cdr form)))

(define (c-compile/%top-level-envt sf form lex-envt dyn-envt mode)
    (make-const '$$bogus mode))
#|
    (make-const
	(select
	    (lambda (bdg) 
		(eq? (class-of bdg) <TopLevelVar>))
	    (table-values->list (table (the-top-level lex-envt))))
	mode)
|#

(define (install-ezc-forms! envt)
    (bind! envt (make <SpecialForm> name: 'snarf 
				    compiler: c-compile/snarf))
    (bind! envt (make <SpecialForm> name: 'define-shadow 
				    compiler: c-compile/define-shadow))
    (bind! envt (make <SpecialForm> name: 'define-system 
				    compiler: c-compile/define-system))
    (bind! envt (make <SpecialForm> name: '%shadow 
				    compiler: c-compile/%shadow))
    (bind! envt (make <SpecialForm> name: '%code 
				    compiler: c-compile/%code))
    (bind! envt (make <SpecialForm> name: '%top-level-environment 
				    compiler: c-compile/%top-level-envt))
    (bind! envt (make <SpecialForm> name: 'define 
				    compiler: c-compile/define))
    (bind! envt (make <SpecialForm> name: '%define-glue 
				    compiler: c-compile/%define-glue))
    (bind! envt (make <SpecialForm> name: '%template 
				    compiler: c-compile/%template))
    (bind! envt (make <SpecialForm> name: 'define-raw 
				    compiler: c-compile/define-raw))
    (bind! envt (make <SpecialForm> name: 'prologue 
				    compiler: c-compile/prologue))
    (bind! envt (make <SpecialForm> name: 'export 
				    compiler: c-compile/export))
    (bind! envt (make <SpecialForm> name: 'import 
				    compiler: c-compile/import))
    (bind! envt (make <SpecialForm> name: 'include-c 
				    compiler: c-compile/include-c))
    (bind! envt (make <SpecialForm> name: 'include-h 
				    compiler: c-compile/include-h))
    (bind! envt (make <SpecialForm> name: 'run 
				    compiler: c-compile/run))
    (bind! envt (make <SpecialForm> name: 'run-named
				    compiler: c-compile/run-named))
    (bind! envt (make <SpecialForm> name: 'load 
				    compiler: c-compile/load))
    (bind! envt (make <SpecialForm> name: '%expr 
				    compiler: compile/prim-expr))
    (bind! envt (make <SpecialForm> name: '%cast
				    compiler: compile/cast))
    (bind! envt (make <SpecialForm> name: '%call 
				    compiler: compile/prim-call))
    (bind! envt (make <TopLevelVar> name: '%version 
				    value: %version ; "v2.0.5f, 94.05.21"
				    )))

(define (display-seperated items sep)
    (if (pair? items)
	(begin
	    (display (car items))
	    (let loop ((i (cdr items)))
		(if (pair? i)
		    (begin
			(display sep)
			(display (car i))
			(loop (cdr i))))))))
