#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;  Generate ASSEMBLY code
;  (passing the generation mode as an argument)

(define-generic-fn code-gen)

; If the mode is 'value then a stmt/expr tuple is returned
; otherwise, only a stmt is returned

(define (m s e)		; make a stmt/expr tuple
    (vector s e))
    
(define (s code)	; extract the stmt part
    (vector-ref code 0))

(define (e code)	; extract the expr part
    (vector-ref code 1))

; The code generation modes are:
;	value
;	effect
;	tail
;	(bind min rest?)
;	(reg-bind min rest? first-reg)

; The meaning of the "bind" modes is that the code
; is to have its value or values be the initializers
; of the given binding (either on the heap or in registers,
; according as whether it is (bind ...) or (reg-bind ...))


; code context management...

(define (make-code-ctx)
    (vector '() 0))

(define (code-ctx-literals* code-ctx)
    (vector-ref code-ctx 0))

(define (code-ctx-literals code-ctx)
    (reverse (code-ctx-literals* code-ctx)))
    

; Normally, the dynamic state holds the current code context
; (this is so that it doesn't have to be passed so deeply in 
;  the call chain)

(define (current-code-context)
    (let ((bdg (%fluid-assq '*code-context*)))
	(if bdg
	    (cdr bdg)
	    (error/internal "*code-context* not bound"))))
	    
(define (alloc-label)
    (let* ((code-ctx (current-code-context))
           (i (vector-ref code-ctx 1)))
	(vector-set! code-ctx 1 (add1 i))
	i))
	
(define (coalesce-literal thing)
    (let ((current-literals (code-ctx-literals* (current-code-context))))
	(if (eq? (class-of thing) <TopLevelVar>)
	    (memq thing current-literals)
	    (if (eq? (class-of thing) <Template>)
		#f
		(member thing current-literals)))))

(define (alloc-literal thing)
    (let ((already (coalesce-literal thing)))
	(if already
	    (sub1 (length already))
	    (let ((code-ctx (current-code-context)))
		(vector-set! code-ctx 0
		    (cons thing
			  (code-ctx-literals* code-ctx)))
		(sub1 (length (code-ctx-literals* code-ctx)))))))


; CODE-GEN-PROCEDURE is the external entry point into this
; code generation mechanism

(define (code-gen-procedure self ct-envt code-ctx) ;; (self <ic-procedure>)
    (fluid-let ((*code-context* code-ctx))
	(code-gen (body self) ct-envt '() 'tail)))

; SIMPLE? returns #t if the value-compilation of the given
; icode results in side-effect free and inlinable code
; It is used to notice when the remaining arguments in a lazy-args
; aren't going to affect an expr

(define (simple? icode)
    (let ((c (class-of icode)))
	(or (eq? c <ic-lex-ref>)
	    (eq? c <ic-tl-ref>)
	    (eq? c <ic-const>))))

; LEX-VAR constructs an effective address (ea) for a lexical
; variable.  The type depends on whether or not the variable
; is found in the registers or not.

(define (lex-var ct-envt in-regs var)
    (let ((r (memq var in-regs)))
	(if r
	    `(reg ,(sub1 (length r)) ,(name var))
	    (let ((lex-addr (find-lex-addr ct-envt var)))
		`(lex-var ,(car lex-addr) ,(cdr lex-addr) ,(name var))))))

; TL-VAR constructs an eff.addr. for a top-level variable.

(define (tl-var  var)
    (list 'tl-var (alloc-literal var) (name var)))

; NORMALIZE-BIND-EXPRS takes a list of expressions and a
; desired structure specification (min+rest?) and massages
; the exprs into the desired form.  It will pad the exprs
; with #f's if necessary, and drop expressions if needed,
; and construct a list as needed for rest?.  The returned
; list of expressions always has length (+ min (if rest? 1 0))

; NOTE: The input, exprs, is a list of typed exprs (t-expr)
; but the output is a list of plain (untyped) expressions of
; type <Object*>

(define (normalize-bind-exprs exprs min rest?)
    
    (define (collect exprs)
	(if (null? exprs)
	    "NIL_OBJ"
	    (list "cons" 
		  (obj-expr (car exprs))
		  (collect (cdr exprs)))))

    (let loop ((min min) (exprs exprs))
	(if (eq? min 0)
	    (if rest?
		(list (collect exprs))
		'())
	    (if (null? exprs)
		(cons "FALSE_OBJ" 
		      (loop (sub1 min) '()))
		(cons (obj-expr (car exprs)) 
		      (loop (sub1 min) (cdr exprs)))))))

; MAKE-BINDER takes a list of expressions and a binding mode (ie
; '(bind ...) or '(reg-bind ...)) and constructs the appropriate
; binder, filling out the expressions or truncating them to fit.
; Currently, there aren't any assurances that the expressions 
; will be evaluated in any particular order, but that will probably
; change (because if we generate explicit calls to gvec_write to 
; build up the new envt, we can assure the order just as easily as
; for the reg-bind)

(define (make-binder t-exprs mode)
    (let ((norm (normalize-bind-exprs t-exprs (cadr mode) (caddr mode))))
	(if (eq? (car mode) 'bind)
	    (list (cons 'bind norm))
	    (let loop ((result '()) (src norm) (dst (cadddr mode)))
		(if (null? src)
		    (reverse result)
		    (loop (cons `(set! (reg ,dst #f) ,(car src)) 
		    	        result)
		    	  (cdr src)
			  (add1 dst)))))))
    
(define (code-gen-simple stmt expr mode)
    (case mode
	((value)	(m stmt expr))
	((effect)	stmt)
	((tail)		(append stmt
				`((set! (reg 0 #f) ,(obj-expr expr))
				  (return 1))))
	(else 		(append stmt (make-binder (list expr) mode)))))

(define (code-gen-ref rvalue mode)
    (code-gen-simple '() (make-t-expr <Object*> (list 'ref rvalue)) mode))
    
(define (code-gen-set lvalue rvalue mode in-regs)
    (case mode
	((value)	(let ((temp `(reg ,(length in-regs) #f)))
			    (m (append (s rvalue) 
			    	       `((set! ,temp ,(obj-expr (e rvalue)))
				         (set! ,lvalue (ref ,temp))))
			       `(ref ,temp))))
	((effect)	(append (s rvalue) 
				`((set! ,lvalue ,(obj-expr (e rvalue))))))
	((tail)		(append (s rvalue) 
			    	`((set! ,lvalue ,(obj-expr (e rvalue)))
; CHECK THIS:
				 ; how do we know it's in REG0!?
				  (return 1))))
	(else 		(append (s rvalue)
				`((set! ,lvalue ,(obj-expr (e rvalue))))
; CHECK THIS:
				; shouldn't this use a temp var...?
		 	        (make-binder (list (e rvalue)) mode)))))

(define-method code-gen ((self <ic-const>) ct-envt in-regs mode)
    (let ((v (value self)))
	(code-gen-simple 
	    '() 
	    (cond ((eq? v #t) yes-const)
	    	  ((eq? v #f) no-const)
		  ((integer? v) 
			(make-t-const <raw-int> (number->string v) v))
		  ((real? v) 
			(make-t-const <raw-float> (number->string v) v))
		  ((char? v)
			(make-t-const <Object*> 
				      (list "MAKE_ASCII_CHAR" 
				            (char->integer v)) 
				      v))
		  ((null? v) (make-t-const <Object*> "NIL_OBJ" '()))
		  ((eq? v '#none) (make-t-const <Object*> "NOVALUE_OBJ" v))
		  ((eq? v '#undef) (make-t-const <Object*> "UNDEFINED_OBJ" v))
		  ((eq? v '#uninit) (make-t-const <Object*> 
		  				"UNINITIALIZED_OBJ" v))
		  ((eq? v '#unbound) (make-t-const <Object*> "UNBOUND_OBJ" v))
		  (#t (make-t-expr <Object*> 
		  		   (list 'literal 
					 (alloc-literal v) v))))
	    mode)))

(define *num-anonymous* 0)

(define (new-anonymous)
    (set! *num-anonymous* (add1 *num-anonymous*))
    (list 'anonymous *num-anonymous*))

(define-method code-gen ((self <ic-lambda>) ct-envt in-regs mode)
    (let* ((code (proc self))
	   (code-context (make-code-ctx))
	   (name (if (name code) 
		     (name code) 
		     (new-anonymous)))
	   (insns (code-gen-procedure code ct-envt code-context))
	   (t (template-within-part name
			            (count-vinsns insns)
			            (write-vinsns name insns)
			            (code-ctx-literals code-context))))
	(code-gen-simple '() 
			 (make-t-expr <Closure*> 
				     (list 'closure (alloc-literal t)))
			 mode)))

(define-method code-gen ((self <ic-lex-ref>) ct-envt in-regs mode)
    (code-gen-ref (lex-var ct-envt in-regs (var self)) mode))

(define-method code-gen ((self <ic-lex-set>) ct-envt in-regs mode)
    (code-gen-set (lex-var ct-envt in-regs (var self))
    		  (code-gen (rhs self) ct-envt in-regs 'value)
		  mode 
		  in-regs))

(define-method code-gen ((self <ic-root-ref>) ct-envt in-regs mode)
    (code-gen-ref `(root ,(c-name (var self))) mode))

(define-method code-gen ((self <ic-root-set>) ct-envt in-regs mode)
    (code-gen-set `(root ,(c-name (var self)))
    		  (code-gen (rhs self) ct-envt in-regs 'value)
		  mode 
		  in-regs))


(define-method code-gen ((self <ic-tl-ref>) ct-envt in-regs mode)
    (code-gen-ref (tl-var (var self)) mode))

(define-method code-gen ((self <ic-tl-set>) ct-envt in-regs mode)
    (code-gen-set (tl-var (var self))
    		  (code-gen (rhs self) ct-envt in-regs 'value)
		  mode 
		  in-regs))


(define-method code-gen ((self <ic-prim-expr>) ct-envt in-regs mode)
    (code-gen-simple '() 
    		     (make-t-expr (car (return-types self)) (expr self)) 
		     mode))


(define-method code-gen ((self <ic-seq>) ct-envt in-regs mode)
    (let loop ((result '()) (stmts (stmt-list self)))
	(if (null? stmts)
	    (begin
		(warning "<ic-seq> with no statements")
		(code-gen-simple '() 
				 (make-t-expr <Object*> "NOVALUE_OBJ") 
				 mode))
	    (let ((code (code-gen (car stmts) 
	    			  ct-envt
				  in-regs
				  (if (null? (cdr stmts))
				      mode
				      'effect))))
		(if (null? (cdr stmts))
		    (if (eq? mode 'value)
			(m (append result (s code)) (e code))
			(append result code))
		    (loop (append result code) (cdr stmts)))))))


(define (use-regs in-regs num)
    (if (eq? num 0)
	in-regs
	(cons #f (use-regs in-regs (sub1 num)))))

(define (shift-regs num-regs first-src-reg first-dst-reg)
    (if (eq? first-src-reg first-dst-reg)
    	'()
	(let loop ((s first-src-reg) 
		    (d first-dst-reg) 
		    (n num-regs) 
		    (r '()))
	    (if (eq? n 0)
		(if (< first-src-reg first-dst-reg)
		    r
		    (reverse r))
		(loop (add1 s)
			(add1 d)
			(sub1 n)
			(cons `(set! (reg ,d #f) (ref (reg ,s #f)))
			    r))))))

;; pop enough environments to get from 'from-envt' to 'to-envt'

(define (pop-envts from-envt to-envt)
    (let loop ((f from-envt) (r '()))
	(if (eq? f to-envt)
	    r
	    (loop (cdr f) (cons '(unbind) r)))))

;; the functionality of CODE-GEN-TAIL-COMBO is
;; extended to support the possibility that in-regs != '()
;; (because we may still reg-allocate if the only combos
;;  in a body are in tail position)
;; this simply involves reg-binding to regs N... where N=(length in-regs)
;; and installing an appropriate argument shifter
;; (generates the function part to reg N+M, also) --
;; the shift occurs JUST BEFORE the apply

(define (code-gen-tail-combo combo ct-envt in-regs)
    (let* ((n (num-args (args combo)))
	   (args-code (code-gen (args combo) 
			        ct-envt
			        in-regs
			        `(reg-bind ,n #f ,(length in-regs))))
	   (fn-code (code-gen (function combo) 
	   		ct-envt 
			(use-regs in-regs n)
			'value)))
	(append args-code
		(s fn-code)
		(shift-regs n (length in-regs) 0)
		`((apply ,n ,(obj-expr (e fn-code)))))))

(define (code-gen-combo code ct-envt in-regs before-restore)
    (let ((l (alloc-label)))
	(append `((save ,(length in-regs) ,l))
		(code-gen-tail-combo code ct-envt '())
		`((label ,l))
		before-restore
		`((restore ,(length in-regs))))))

(define (make-runtime-bind mode fn-args?)

    (define (make-shift-regs)
	(let loop ((num (if (caddr mode)
			    (add1 (cadr mode))
			    (cadr mode)))
		    (src 0)
		    (dst (cadddr mode))
		    (result '()))
	    (if (eq? num 0)
		result
		(loop (sub1 num)
			(add1 src)
			(add1 dst)
			(cons `(set! (reg ,dst #f) (ref (reg ,src #f)))
			    result)))))

    (define (make-heap-bind)
	(let loop ((num (if (caddr mode)
			    (add1 (cadr mode))
			    (cadr mode)))
		    (src 0)
		    (result '()))
	    (if (eq? num 0)
		(list (cons 'bind (reverse result)))
		(loop (sub1 num)
			(add1 src)
			(cons `(ref (reg ,src #f))
			    result)))))

    (append
	(if fn-args?
	    (if (caddr mode)
		`((check>= ,(cadr mode)))
		`((check= ,(cadr mode))))
	    `((set-false< ,(cadr mode))))
	(if (caddr mode)
	    `((collect> ,(cadr mode)))
	    '())
	(if (eq? (car mode) 'reg-bind)
	    (make-shift-regs)
	    (make-heap-bind))))

(define-method code-gen ((self <ic-call>) ct-envt in-regs mode)

    (case mode
	((tail) (code-gen-tail-combo self ct-envt in-regs))
	((value) (let ((r (list 'reg (length in-regs) #f)))
		    (m (code-gen-combo self 
		    		      ct-envt 
				      in-regs 
				     
				      `((set! ,r (ref (reg 0 #f)))))
		       (make-t-expr <Object*> `(ref ,r)))))
	((effect) (let ((r (list 'reg (length in-regs) #f)))
		    (code-gen-combo self 
		    		      ct-envt 
				      in-regs 
				     
				      `())))
	(else   (code-gen-combo self 
				ct-envt 
				in-regs 
				(make-runtime-bind mode #f)))))

;; suppress the (RT) creation of a binding environment
;; when there is nothing to bind.  Note that, if it is a
;; procedure, we go ahead and emit the check

(define-method code-gen ((self <ic-bind>) ct-envt in-regs mode)
    (if (null? (bindings (envt self)))
	(if (inits self)
	    (code-gen (body self) ct-envt in-regs mode)
	    (append (make-runtime-bind '(reg-bind 0 #f 0) #t)
	    	    (code-gen (body self) ct-envt in-regs mode)))
	(code-gen-bind self ct-envt in-regs mode #f)))

;; various cases to consider w.r.t. (unbind)
;;	<ic-bind> for tail vs. non-tail
;;	last thing in <ic-bind> is a combo -- the
;;	  restore on return from IT will then restore the envt
;;        or not, because the SAVE doesn't come until the actual
;;	  call, but that could be changed, and has been contemplated,
;;        but I think it was rejected because it wound up generating
;;	  an additional SAVE
;; body-label is used to give a label to the code just at the beginning
;; of the body/end of the initializers (before the bindings are BOUND)
;; (it is used for loops)

(define (code-gen-bind self ct-envt in-regs mode body-label)
    (let* ((reg-alloc? (not (or (does-apply? (body self)) 
    				(does-lambda? (body self)))))
    	   (dbl-bind? (and reg-alloc? 
	   		   (pair? mode) 
			   (eq? (car mode) 'reg-bind)))
	   (bindings (bindings (envt self)))
	   (body (code-gen (body self)
			    (if reg-alloc?
			    	ct-envt
				(cons bindings ct-envt))
			    (if reg-alloc?
			    	(append (reverse bindings) in-regs)
				in-regs)
			    mode))
	   (bind-mode   (if reg-alloc?
			    (list 'reg-bind 
				(num-args self) 
				(rest? self) 
				(if dbl-bind?
				    (+ (cadddr mode) (length in-regs))
				    (length in-regs)))
			    (list 'bind (num-args self) (rest? self))))
	   (binder  (if (inits self)
			(let ((b (code-gen (inits self)
					    ct-envt
					    (if dbl-bind? 
						(use-regs in-regs 
							  (length bindings))
						in-regs)
					    bind-mode)))
			    (if body-label
				(append b 
				        `((jump ,(+ (length bindings) 
						    (length in-regs))
						,body-label)
					  (label ,body-label)))
				b))
			(make-runtime-bind bind-mode #t))))
	(if (eq? mode 'value)
	    (let ((save-reg `(reg ,(length in-regs) #f)))
		(m (append binder 
			   (s body)
			   `((set! ,save-reg ,(obj-expr (e body))))
			   (if reg-alloc? '() '((unbind))))
		   (make-t-expr <Object*> `(ref ,save-reg))))
	    (if (or (eq? mode 'tail) reg-alloc?)
		(append binder body)
		(append binder body '((unbind)))))))



(define (code-gen-multi-bind-regs inits ct-envt in-regs mode)

    (define (make-list exprs)
	(if (null? exprs)
	    "NIL_OBJ"
	    (list "cons" (obj-expr (car exprs)) (make-list (cdr exprs)))))
	    
    (let loop ((left (cadr mode))	; number of reg bindings left
    	       (result '())		; resulting statements
	       (dst (cadddr mode))	; current dest reg
	       (regs in-regs)		; current reg bindings
	       (src inits))		; source forms
	(if (null? src)
	    ; we're out of source forms
	    (if (eq? left 0)
		; and all register bindings (except possibly the "rest")
		; have been filled
		(if (caddr mode) ; rest?
		    (append result
		    	    `((set! (reg ,dst #f) "NIL_OBJ")))
		    result)
		; there are some register bindings waiting to be filled
		(loop (sub1 left)
		      (append result `((set! (reg ,dst #f) #f)))
		      (add1 dst)
		      (cons #f regs)
		      '()))
	    ; we're NOT out of source forms, so generate code for
	    ; the current one in the current (regs) context
	    (let ((code (code-gen (car src) ct-envt regs 'value))
	    	  (reg `(reg ,dst #f)))
		(if (eq? left 0)
		    ; all register bindings are filled
		    ; (but may still need to remember in a reg if #rest)
		    ; (we should really use code-gen-multi-lazy-args 
		    ;  or similar technology here)
		    ; (OOPS! The #rest list is getting assembled in 
		    ;  the wrong direction!)
		    (if (caddr mode) ; rest?
			(let ((rmdr (code-gen-lazy-args src
						 	ct-envt
							regs)))
			    (append result
				    (s rmdr)
				    `((set! (reg ,dst #f) 
				    	    ,(make-list (e rmdr))))))
			(loop 0
			      (append result 
			      	      (code-gen (car src) 
				      		ct-envt 
						regs 
						'effect))
			      dst
			      regs
			      (cdr src)))
		    ; there are still some register bindings that need
		    ; to be filled (this is actually the COMMON case)
		    (loop (sub1 left)
			  (append result 
				  (s code) 
				  `((set! ,reg ,(obj-expr (e code)))))
			  (add1 dst)
			  (cons #f regs)
			  (cdr src)))))))

(define-method code-gen ((self <ic-multi>) ct-envt in-regs mode)
    (case mode
	((value) 
	    (if (null? (arg-list self))
		;; should this be FALSE_OBJ instead..?
		(m '() (make-t-expr <Object*> "NOVALUE_OBJ"))
		(let ((code (code-gen-lazy-args (arg-list self)
							    ct-envt 
							    in-regs)))
		    (m (s code) (car (e code))))))
	((effect) 
	    (if (null? (arg-list self))
		'()
		(let ((code (code-gen-lazy-args (arg-list self)
							    ct-envt 
							    in-regs)))
		    (s code))))
	((tail)
	    (if (null? (arg-list self))
		`((return 0))
		(let* ((num (length (arg-list self)))
			(rb-mode (list 'reg-bind 
					num
					#f
					(length in-regs)))
			(code (code-gen-multi-bind-regs (arg-list self)
							ct-envt
							in-regs
							rb-mode)))
		    (append code
			    (let loop ((dst num) 
					(src (+ num (length in-regs)))
					(result `((return ,num))))
				(if (eq? dst 0)
				    result
				    (loop (sub1 dst)
					  (sub1 src)
					  (cons `(set! (reg ,(sub1 dst) 
					  		    #f)
						       (ref (reg ,(sub1 src) 
						                 #f)))
						result))))))))
	(else
	    (if (eq? (car mode) 'bind)
		(let ((code (code-gen-lazy-args (arg-list self)
						      ct-envt 
						      in-regs)))
		    (append (s code)
			    (make-binder (e code) mode)))
		(code-gen-multi-bind-regs (arg-list self)
					  ct-envt
					  in-regs
					  mode)))))

(define-method code-gen ((self <ic-if>) ct-envt in-regs mode)

    (define (before-label stmts0)
	(let loop ((stmts stmts0) (result '()))
	    (if (null? stmts)
		stmts0
		(if (and (pair? (car stmts))
			    (eq? (caar stmts) 'label))
		    (reverse result)
		    (loop (cdr stmts) (cons (car stmts) result))))))

    (define (after-label stmts)
	(if (null? stmts)
	    '()
	    (if (and (pair? (car stmts))
	    	     (eq? (caar stmts) 'label))
		stmts
		(after-label (cdr stmts)))))

    (let* ((n (length in-regs))
    	   (l (alloc-label))
	   (gen-br (case mode
	   		((effect) (lambda (ic)
				    (append (code-gen ic 
						      ct-envt 
						      in-regs 
						      'effect)
					    `((jump ,n ,l)))))
	   		((tail) (lambda (ic)
				    (code-gen ic 
					      ct-envt 
					      in-regs 
					      'tail)))
	   		((value) (lambda (ic)
				     (let ((c (code-gen ic 
						        ct-envt 
						        in-regs 
						        'value)))
					(append (s c)
						`((set! (reg ,n #f) 
							,(obj-expr (e c)))
						  (jump ,(add1 n) ,l))))))
			(else (lambda (ic)
				(append (code-gen ic
						  ct-envt
						  in-regs
						  mode)
					`((jump ,(if (eq? (car mode) 'reg-bind)
						     (+ (cadddr mode)
						     	(cadr mode)
							(if (caddr mode) 1 0))
						     n)
						,l)))))))
	   (test (code-gen (condition self) ct-envt in-regs 'value))
	   (true-br (gen-br (if-true self)))
	   (false-br (gen-br (if-false self)))
	   (result (append (s test)
	   		   `((if ,(raw-bool-expr (e test))
			   	 ,(cons 'seq (before-label true-br))
				 ,(cons 'seq (before-label false-br))))
			   (after-label true-br)
			   (after-label false-br)
			   (if (eq? mode 'tail)
			       '()
			       `((label ,l))))))
	(if (eq? mode 'value)
	    (m result (make-t-expr <Object*> `(ref (reg ,n #f))))
	    result)))

(define-method code-gen ((self <ic-call-prim>) ct-envt in-regs mode)

    ; should probably use map for this...
    
    (define (coerce-args args types)
	(if (null? args)
	    (if (null? types)
		'()
		(error/syntax "Too few args to primitive ~a" (function self)))
	    (if (null? types)
		(error/syntax "Too many args to primitive ~a" (function self))
		(cons
		    (expr (coerce (car args) (car types)))
		    (coerce-args (cdr args) (cdr types))))))

    (let* ((args (code-gen-lazy-args (arg-list (args self))
				     ct-envt 
				     in-regs))
	   (the-call (cons (function self) 
			   (coerce-args (e args) (arg-types self)))))
	(if (null? (return-types self))
	    ; calling a procedure...
	    (case mode
		((effect) (append (s args) `((do ,the-call))))
		(else (error/internal 
			"ic-call-prim(~a): not impl for procedures: ~a" 
			(function self) 
			mode)))
	    ; calling a function...
	    (let ((call-t-expr (make-t-expr (car (return-types self))
					    the-call)))
		(case mode
		    ((value) (m (s args) call-t-expr))
		    ((tail) (append
				(s args)
				`((set! (reg 0 #f) ,(obj-expr call-t-expr))
				    (return 1))))
		    (else (error/internal 
			    "ic-call-prim(~a): not impl for functions: ~a" 
			    (function self) 
			    mode)))))))

(define (code-gen-lazy-args arg-list ct-envt in-regs)

    (define (all? list)
	(if (null? list)
	    #t
	    (if (car list)
		(all? (cdr list))
		#f)))
	    
    (let loop ((result-exprs '()) 
	       (result-stmts '())
	       (regs in-regs)
	       (args arg-list)
	       (simplicity (map simple? arg-list)))
	(if (null? args)
	    (m result-stmts (reverse result-exprs))
	    (if (or (eq? (class-of (car args)) <ic-const>)
	    	    (all? (cdr simplicity)))
		; don't put it in a register
		(let ((c (code-gen (car args) ct-envt regs 'value)))
		    (loop (cons (e c) result-exprs)
		    	  (append result-stmts (s c))
			  regs
			  (cdr args)
			  (cdr simplicity)))
		; put it in a register
		(let ((r `(reg ,(length regs) #f))
		      (c (code-gen (car args) ct-envt regs 'value)))
		    (loop (cons (make-t-expr <Object*> `(ref ,r)) result-exprs)
		    	  (append result-stmts 
			  	  (s c)
				  `((set! ,r ,(obj-expr (e c)))))
			  (cons #f regs)
			  (cdr args)
			  (cdr simplicity)))))))

;; loop optimization...

(define code-gen-loop 'not-implemented)
(define code-gen-jump 'not-implemented)

(define-method code-gen ((self <ic-loop>) ct-envt in-regs mode)
    (code-gen-loop self ct-envt in-regs mode))
    
(define-method code-gen ((self <ic-jump>) ct-envt in-regs mode)
    (code-gen-jump self ct-envt in-regs mode))
    
