/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "runtim.h"
#include <string.h>
#include <stdlib.h>

#define OSLIB_NOT_IMPLEMENTED (-100)

static void oslib_error( const char *fn, int rc )
{
    scheme_error( "OS library error in ~s: rc ~d",
    		  2, make_string(fn), int2fx(rc) );
}

const char *os_getenv( const char *key )
{
    return NULL;
}

obj os_scan_directory( const char *path )
{
    oslib_error( "os_scan_directory", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

obj os_getwd( void )
{
    oslib_error( "os_getwd", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

obj os_getpid( void )
{
    oslib_error( "os_getpid", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

obj os_invoke( obj command )
{
    oslib_error( "os_invoke", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

obj os_error_string( int rc )
{
    oslib_error( "os_error_string", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

obj os_get_resource_usage( obj i_class )
{
    oslib_error( "os_get_resource_usage", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

obj os_file_status( const char *path, obj t_class )
{
    oslib_error( "os_file_status", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

bool os_file_exists_p( const char *path )
{
    oslib_error( "os_file_exists_p", OSLIB_NOT_IMPLEMENTED );
}

void os_setwd( const char *path )
{
    oslib_error( "os_setwd", OSLIB_NOT_IMPLEMENTED );
}

obj os_fork( void )
{
    oslib_error( "os_fork", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

int os_wait( obj child )
{
    oslib_error( "os_wait", OSLIB_NOT_IMPLEMENTED );
    return 0;
}

obj os_fork_wait( void )
{
    oslib_error( "os_fork_wait", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

void os_sleep( UINT_32 usec )
{
    oslib_error( "os_sleep", OSLIB_NOT_IMPLEMENTED );
}

obj os_time_now( obj t_class )
{
    oslib_error( "os_time_now", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

/* convert a unix-standard pathname to local OS format */

/*
    Rob Browning. Fall, 1993.
*/

/*
This seems like a good short-term solution, except (and I hate to say this)
we should use a static buffer to assuage the possibility of a memory leak.
And we should call it unix_path_to_native_path to be more consistent with
naming convention and style.

Also, don't forget to handle the unix "..", which turns into something like
"::".   For example (and this is something that will happen):

unix_path_to_native_path( "../foo/bar.scm" ) ==> "::foo:bar.scm"
unix_path_to_native_path( "quux/../foo/bar.scm" ) ==> ":quux::foo:bar.scm"
unix_path_to_native_path( "/tmp/temp.scm" ) ==> "tmp:temp.scm"

NOTE:  a reference to a DIRECTORY will always end in "/"
       (ie, via (chdir ...) and possibly in the future (opendir ...))
       a root directory ("/") is never used

../   =>  ::
../../ => :::
../foo/./bar  => ::foo:bar
../foo/../bar ==> ::foo::bar
../foo/../bar/ ==> ::foo::bar:
/foo/bar/../baz ==> foo:bar::baz
../foo ==> ::foo
foo ==> foo       (":foo" might be OK, but not preferable)
foo/bar ==> :foo:bar
foo/bar/./baz/../bleen ==> :foo:bar:baz::bleen

*/

void unix_path_to_native_path(char* path) {
        char* orig_str = (char *) malloc(strlen(path)+1);
        int i=0,j=0,in;

        strcpy(orig_str,path);

        /* Check to see if just a file name */
        if(!strpbrk(orig_str,"/")) return;

        if(orig_str[i]!='/') {
                path[j++]=':';
                while(1) {
                        if(orig_str[i]=='/') i++;
                        if(strncmp(&(orig_str[i]),"./",2)==0) {
                                i+=2;
                        }
                        else if(strncmp(&(orig_str[i]),"../",3)==0) {
                                path[j++]=':';
                                i+=3;
                        }
                        else break;
                }
        }
        else i++;

        while(orig_str[i]!='\0') {
                if(orig_str[i]=='/') {
                        path[j++]=':';
                        i++;
                        while(1) {
                                if(orig_str[i]=='/') i++;
                                if(strncmp(&(orig_str[i]),"./",2)==0) {
                                        i+=2;
                                }
                                else if(strncmp(&(orig_str[i]),"../",3)==0) {
                                        path[j++]=':';
                                        i+=3;
                                }
                                else break;
                        }
                }
                else path[j++]=orig_str[i++];
        }
        path[j]='\0';

        free(orig_str);
}

FILE *os_fopen(const char* filename,const char* mode) {
        FILE* opened_file;
        char* translated_name = (char *) malloc(strlen(filename)+12);
        /* The 12 is just a safety margin; I think that 2 would be sufficient */
        /* The only time a mac path could be larger is when the path is
relative. */
        /* The mac adds a : on the front then. */

        if(!translated_name) {
                fprintf( stderr, 
		         "EasyScheme: malloc failed in fopen_unix_path.\n");
                exit(-1);
        }
        strcpy(translated_name,filename);
        unix_path_to_native_path(translated_name);
        opened_file = fopen(translated_name,mode);
        free(translated_name);
        return opened_file;
}

obj os_path( obj path )
{
/* The 12 is just a safety margin; I think that 2 would be sufficient */
/* The only time a mac path could be larger is when the path is relative. */
/* The mac adds a : on the front then. */
char* translated_name = (char *) malloc(string_length(path)+12);
obj str;

    if(!translated_name)
	oslib_error( "os_path", -1 );
	
    strcpy( translated_name, string_text(path) );
    unix_path_to_native_path( translated_name );
    str = make_string( translated_name );
    free( translated_name );
    return str;
}
