/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include <LowMem.h>
#include <string.h>
#include <stdlib.h>
#include "runtim.h"
#include "scmtime.h"
#include "osglue.h"

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE (!FALSE)
#endif

#define MAXPATHLEN (1024)
#define OSLIB_NOT_IMPLEMENTED (-100)

void PathNameFromDirID(long dirID, short vRefNum, StringPtr fullPathName);
void PathNameFromWD(long vRefNum, StringPtr pathName);
void pstrcat(StringPtr dst, StringPtr src);
void pstrinsert(StringPtr dst, StringPtr src);

static void oslib_error( const char *fn, int rc )
{
    scheme_error( "OS library error in ~s: rc ~d",
              2, make_string(fn), int2fx(rc) );
}

const char *os_getenv( const char *key )
{
    return NULL;
}

obj os_time_now( obj t_class )
{
    unsigned long time_in_secs;
    struct scheme_time time_in_scheme;
    int err;

    /* Accesses chip clock directly */

    if ((err = ReadDateTime(&time_in_secs)) != noErr)
    oslib_error( "ReadDateTime", err );

    /* Acceses global variable Time that may be less accurate
       in the presence of interrupts.  This cannot fail. */
    /* GetDateTime(&time_in_secs); */

    time_in_scheme.sec = time_in_secs;
    time_in_scheme.usec = 0 /* Microseconds() not sure how to use this call
yet. */;
    return make_time( time_in_scheme, t_class );
}

obj os_scan_directory( const char *path )
{
    oslib_error( "os_scan_directory", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

obj os_getwd( void )
{       short vol = - LMGetSFSaveDisk();
    long dir = LMGetCurDirStore();
    char temp[MAXPATHLEN+1];

    PathNameFromDirID(dir,vol,(StringPtr) temp);

    return make_string(temp);
}

obj os_getpid( void )
{
    oslib_error( "os_getpid", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

obj os_get_resource_usage(  obj i_class  )
{
    oslib_error( "os_get_resource_usage", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

obj os_file_status( const char *path, obj t_class)
{
    oslib_error( "os_file_status", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

void os_setwd( const char *path )
{
    oslib_error( "os_setwd", OSLIB_NOT_IMPLEMENTED );
}

void os_sleep( UINT_32 usec )
{       EventRecord theEvent;
    UINT_32 sleep = (usec * 50000)/3; /* (usec * 1000000)/60 */
    UINT_32 wakeup = TickCount() + sleep;

    while(TickCount() < wakeup)
            WaitNextEvent(0 /*eventMask*/,&theEvent,sleep,NULL);
}

/* convert a unix-standard pathname to local OS format */

/*
    Rob Browning. Fall, 1993.
*/

/*
This seems like a good short-term solution, except (and I hate to say this)
we should use a static buffer to assuage the possibility of a memory leak.
And we should call it unix_path_to_native_path to be more consistent with
naming convention and style.

Also, don't forget to handle the unix "..", which turns into something like
"::".   For example (and this is something that will happen):

unix_path_to_native_path( "../foo/bar.scm" ) ==> "::foo:bar.scm"
unix_path_to_native_path( "quux/../foo/bar.scm" ) ==> ":quux::foo:bar.scm"
unix_path_to_native_path( "/tmp/temp.scm" ) ==> "tmp:temp.scm"

NOTE:  a reference to a DIRECTORY will always end in "/"
       (ie, via (chdir ...) and possibly in the future (opendir ...))
       a root directory ("/") is never used

../   =>  ::
../../ => :::
../foo/./bar  => ::foo:bar
../foo/../bar ==> ::foo::bar
../foo/../bar/ ==> ::foo::bar:
/foo/bar/../baz ==> foo:bar::baz
../foo ==> ::foo
foo ==> foo       (":foo" might be OK, but not preferable)
foo/bar ==> :foo:bar
foo/bar/./baz/../bleen ==> :foo:bar:baz::bleen

*/

void unix_path_to_native_path(char* path);
void unix_path_to_native_path(char* path) {
    char* orig_str = (char *) malloc(strlen(path)+1);
    int i=0,j=0,in;

    strcpy(orig_str,path);

    /* Check to see if just a file name */
    if(!strpbrk(orig_str,"/")) return;

    if(orig_str[i]!='/') {
            path[j++]=':';
            while(1) {
                    if(orig_str[i]=='/') i++;
                    if(strncmp(&(orig_str[i]),"./",2)==0) {
                            i+=2;
                    }
                    else if(strncmp(&(orig_str[i]),"../",3)==0) {
                            path[j++]=':';
                            i+=3;
                    }
                    else break;
            }
    }
    else i++;

    while(orig_str[i]!='\0') {
            if(orig_str[i]=='/') {
                    path[j++]=':';
                    i++;
                    while(1) {
                            if(orig_str[i]=='/') i++;
                            if(strncmp(&(orig_str[i]),"./",2)==0) {
                                    i+=2;
                            }
                            else if(strncmp(&(orig_str[i]),"../",3)==0) {
                                    path[j++]=':';
                                    i+=3;
                            }
                            else break;
                    }
            }
            else path[j++]=orig_str[i++];
    }
    path[j]='\0';

    free(orig_str);
}

FILE *os_fopen(const char* filename,const char* mode);
FILE *os_fopen(const char* filename,const char* mode) {
    FILE* opened_file;
    char* translated_name = (char *) malloc(strlen(filename)+12);
    /* The 12 is just a safety margin; I think that 2 would be sufficient */
    /* The only time a mac path could be larger is when the path is
relative. */
    /* The mac adds a : on the front then. */

    if(!translated_name) {
            fprintf(stderr, "Easy Scheme: Allocation failed in
fopen_unix_path.\n");
            exit(-1);
    }
    strcpy(translated_name,filename);
    unix_path_to_native_path(translated_name);
    opened_file = fopen(translated_name,mode);
    free(translated_name);
    return opened_file;
}

obj os_path( obj path );
obj os_path( obj path )
{
/* The 12 is just a safety margin; I think that 2 would be sufficient */
/* The only time a mac path could be larger is when the path is relative. */
/* The mac adds a : on the front then. */
char* translated_name = (char *) malloc(strlen((char *)path)+12);
obj str;

    if(!translated_name)
    oslib_error( "os_path", -1 );

    strcpy( translated_name, (char *)path );
    unix_path_to_native_path( translated_name );
    str = make_string( translated_name );
    free( translated_name );
    return str;
}

// Getting a Full Pathname
// Utilities from the Apple Q&A Stack showing how to get the full
// pathname of a file.  Note that this is NOT the recommended way
// to specify a file to Toolbox routines.  These routines should be
// used for displaying full pathnames only.

// Assumes inclusion of <MacHeaders>
#define haveAUX() 0

/*
 * Pascal string utilities
 */
/*
 * pstrcat - add string 'src' to end of string 'dst'
 */
void pstrcat(StringPtr dst, StringPtr src)
{
    /* copy string in */
    BlockMove(src + 1, dst + *dst + 1, *src);
    /* adjust length byte */
    *dst += *src;
}

/*
 * pstrinsert - insert string 'src' at beginning of string 'dst'
 */
void pstrinsert(StringPtr dst, StringPtr src)
{
    /* make room for new string */
    BlockMove(dst + 1, dst + *src + 1, *dst);
    /* copy new string in */
    BlockMove(src + 1, dst + 1, *src);
    /* adjust length byte */
    *dst += *src;
}

void PathNameFromDirID(long dirID, short vRefNum, StringPtr fullPathName)
{
    DirInfo block;
    //CInfoPBRec    block;
    Str255  directoryName;
    OSErr   err;

    fullPathName[0] = '\0';

    block.ioDrParID = dirID;
    block.ioNamePtr = directoryName;
    do {
	block.ioVRefNum = vRefNum;
	block.ioFDirIndex = -1;
	block.ioDrDirID = block.ioDrParID;
	err = PBGetCatInfo(((CInfoPBRec *)(&block)), FALSE);
	if (haveAUX()) {
	    if (directoryName[1] != '/')
		/* If this isn't root (i.e. "/"), append a slash ('/') */
		pstrcat(directoryName, (StringPtr)"\p/");
	}
	else
	    pstrcat(directoryName, (StringPtr)"\p:");
	pstrinsert(fullPathName, directoryName);
    } while (block.ioDrDirID != 2);
}

/*
PathNameFromWD:
Given an HFS working directory, this routine returns the full pathname that
corresponds to it. It does this by calling PBGetWDInfo to get the VRefNum and
DirID of the real directory. It then calls PathNameFromDirID, and returns its
result.

*/
void PathNameFromWD(long vRefNum, StringPtr pathName)
{
    WDPBRec myBlock;
    OSErr   err;
    /*
    PBGetWDInfo has a bug under A/UX 1.1.  If vRefNum is a real
     vRefNum and not a wdRefNum, then it returns garbage.
    Since A/UX has only 1 volume (in the Macintosh sense) and
    only 1 root directory, this can occur only when a file has been
    selected in the root directory (/).
    So we look for this and hardcode the DirID and vRefNum.
    */
    if ((haveAUX()) && (vRefNum == -1))
                    PathNameFromDirID(2, -1, pathName);
    else {
                    myBlock.ioNamePtr = nil;
                    myBlock.ioVRefNum = vRefNum;
                    myBlock.ioWDIndex = 0;
                    myBlock.ioWDProcID = 0;
    /*
     Change the Working Directory number in vRefnum into a real
    vRefnum and DirID. The real vRefnum is returned in ioVRefnum,
     and the real DirID is returned in ioWDDirID.
    */
                    err = PBGetWDInfo(&myBlock, FALSE);
                    if (err != noErr)
                                    return;
                    PathNameFromDirID(myBlock.ioWDDirID,
myBlock.ioWDVRefNum,
                                    pathName);
    }
}


/***** The Timer *****/






obj os_invoke( obj command )
{
    oslib_error( "os_invoke", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

obj os_error_string( int rc )
{
    oslib_error( "os_error_string", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

obj os_fork( void )
{
    oslib_error( "os_fork", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}

int os_wait( obj child )
{
    oslib_error( "os_wait", OSLIB_NOT_IMPLEMENTED );
    return 0;
}

obj os_fork_wait( void )
{
    oslib_error( "os_fork_wait", OSLIB_NOT_IMPLEMENTED );
    return FALSE_OBJ;
}




