/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "intrs.h"
#ifdef __cplusplus
extern "C" {
#endif
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>

#if defined(PLATFORM_NEXT) && !defined(_POSIX_SOURCE)
typedef int pid_t;
#endif

#ifdef SUBPLATFORM_SOLARIS
#include <dirent.h>
#include <sys/rusage.h>
#else
#include <sys/dir.h>
#endif

#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#ifdef PLATFORM_DEC
#include <sys/param.h>
#endif
#include <sys/wait.h>
#include <pwd.h>

extern int errno;

#ifndef MAXPATHLEN
/* if, for whatever reason, we haven't found a definition
   for MAXPATHLEN yet, create one */
#define MAXPATHLEN (1024)
#endif /* MAXPATHLEN */

#ifdef __cplusplus
#ifdef PLATFORM_AIX
 int setitimer( int, struct itimerval *, struct itimerval * );
 int getitimer( int, struct itimerval * );
 int getrusage( int, struct rusage * );
 long random( void );
#endif /* PLATFORM_AIX */

#ifdef PLATFORM_LINUX
int wait( void * );
#endif /* PLATFORM_LINUX */

#ifdef PLATFORM_NEXT
/*int wait( void * ); */
char *getwd( char * );
extern int      chdir(const char *);
extern int      getpid(void);
extern int	fork(void);
#endif /* PLATFORM_NEXT */
}
#else
#include <unistd.h>
#endif

#include <string.h>
#include "runtim.h"
#include "scmtime.h"

/* #include "vinsns.h" */

#ifdef __cplusplus
#define PLAIN_C(decl) extern "C" { decl }
#else
#define PLAIN_C(decl) extern decl
#endif

#if defined(PLATFORM_AIX) || defined(SUBPLATFORM_SOLARIS)
typedef struct dirent directory_entry;
#else
typedef struct direct directory_entry;
#endif

#if !defined(PLATFORM_LINUX) && !defined(PLATFORM_NEXT)
PLAIN_C(int errno;)
PLAIN_C(char *getenv( const char * );)
PLAIN_C(char *getcwd( char *buf, size_t size );)
PLAIN_C(pid_t getpid( void );)
PLAIN_C(int system( const char *cmd );)
PLAIN_C(int gettimeofday( struct timeval *, struct timezone * );)
PLAIN_C(unsigned usleep( unsigned );)
PLAIN_C(int scandir( char *path, 
		     directory_entry *(*nl[]), 
		     int (*)(), 
		     int (*)() ); )
#endif /* !PLATFORM_LINUX && !PLATFORM_NEXT */

#if defined(PLATFORM_NEXT) || (defined(PLATFORM_SUNOS) && !defined(SUBPLATFORM_SOLARIS))
typedef int ssize_t;
#endif


static void oslib_error( const char *fn, int rc )
{
    scheme_error( "OS library error in ~s: rc ~d",
    		  2, make_string(fn), int2fx(rc) );
}

static obj os_time( struct timeval *tv, obj t_class )
{
struct scheme_time t;

    t.sec = tv->tv_sec;
    t.usec = tv->tv_usec;
    return make_time( t, t_class );
}


const char *os_getenv( const char *key )
{
const char *v;

    OUT_CALL( v = getenv( key ); );
    return v;
}

#ifdef __cplusplus
#define SCANDIR_NULL_PROC (int (*)(...))NULL
#else
#define SCANDIR_NULL_PROC NULL
#endif

obj os_scan_directory( const char *path )
{
directory_entry *ent, **entries;
int i, n;
obj list;

    OUT_CALL(
	n = scandir( (char *)path, 
		     &entries, 
		     SCANDIR_NULL_PROC, 
		     SCANDIR_NULL_PROC );
	if (n < 0)
	    list = FALSE_OBJ;
	else
	{
	    list = NIL_OBJ;
	    for (i=n; i>0; )
	    {
		ent = entries[--i];
		list = cons( make_string( ent->d_name ), list );
		free( ent );
	    }
	    free( entries );
	}
    );
    return list;
}

obj os_getwd( void )
{
char *p, temp[MAXPATHLEN];

#ifdef BSD_UNIX
    OUT_CALL( p = getwd( temp ); );
#else
    OUT_CALL( p = getcwd( temp, 1024 ); ); 
#endif
    if (!p)
	oslib_error( "os_getwd", errno );
    return make_string(p);
}

obj os_getpid( void )
{
    return int2fx( getpid() );
}

obj os_invoke( obj command )
{
int rc;

    OUT_CALL( rc = system( string_text( command ) ); );
    return int2fx( rc );
}

obj os_error_string( int rc )
{
extern int sys_nerr;
extern char *sys_errlist[];

    if (rc >= 0 && rc < sys_nerr)
	return make_string( sys_errlist[rc] );
    else
	return FALSE_OBJ;
}

obj os_get_resource_usage( obj i_class )
{
struct rusage info;

    OUT_CALL( getrusage( RUSAGE_SELF, &info ); );
    return make4( vector_class, os_time( (struct timeval *)&info.ru_utime, 
					 i_class ),
    				os_time( (struct timeval *)&info.ru_stime, 
					 i_class ),
				int2fx( info.ru_msgsnd ),
				int2fx( info.ru_msgrcv ) );
}

obj os_make_directory( const char *path )
{
int rc;

    OUT_CALL( rc = mkdir(path,0777); );
    if (rc)
      return FALSE_OBJ;
    else
      return TRUE_OBJ;
}

obj os_file_seek( obj file, UINT_32 offset )
{
off_t p;

    assert( OBJ_ISA_FIXNUM(file) );
    p = lseek( fx2int( file ), offset, SEEK_SET );
    return int2fx(p);
}

obj os_file_read( obj file, obj bvec_class, UINT_32 bytes )
{
ssize_t n;
obj buf;

    assert( OBJ_ISA_FIXNUM(file) );
    if (EQ(bvec_class,string_class))
      buf = bvec_alloc( bytes+1, bvec_class );
    else
      buf = bvec_alloc( bytes, bvec_class );
    n = read( fx2int(file), PTR_TO_DATAPTR(buf), bytes );
    if ((int)n <= 0)
      {
	return FALSE_OBJ;
      }
    else if (n != bytes)
      {
	obj nbuf; 

	/* trim it down */
	if (EQ(bvec_class,string_class))
	  nbuf = bvec_alloc( n+1, bvec_class );
	else
	  nbuf = bvec_alloc( n, bvec_class );
	memcpy( PTR_TO_DATAPTR(nbuf), PTR_TO_DATAPTR(buf), n );
	return nbuf;
      }
    else
      return buf;
}

obj os_file_write( obj file, obj data )
{
    ssize_t a, n;
    void *p;
  
    if (STRING_P(data))
      {
	a = string_length(data);
	p = (void *)string_text(data);
      }
    else if (OBJ_ISA_PTR(data))
      {
	assert( OBJ_ISA_PTR(data) );
	p = PTR_TO_DATAPTR(data);
	a = SIZEOF_PTR(data);
      }
    else
      {
	p = &data;
	a = sizeof(obj);
      }

    n = write( fx2int(file), p, a );
    if (n != a)
      return FALSE_OBJ;
    else
      return TRUE_OBJ;
}

obj os_open_file( const char *path,
		  bool read_p, bool write_p, bool create_p, 
		  bool append_p, bool nonblock_p )
{
    int flags, fd;

    flags = (read_p ? (write_p ? O_RDWR : O_RDONLY) : (write_p ? O_WRONLY : 0))
            | (append_p ? O_APPEND : 0)
#ifdef O_NONBLOCK
	    | (nonblock_p ? O_NONBLOCK : 0)
#endif
	    | (create_p ? O_CREAT : 0);
    fd = open( path, flags, 0666 );
    if (fd < 0)
      {
	return FALSE_OBJ;
      }
    else
      {
	return int2fx(fd);
      }
}

bool os_file_exists_p( const char *path )
{
int rc;
struct stat info;

    OUT_CALL( rc = stat( path, &info ); );
    return rc ? NO : YES;
}

obj os_file_status( const char *path, obj t_class )
{
int rc;
struct stat info;

    OUT_CALL( rc = stat( path, &info ); );
    if (rc)
	oslib_error( "os_file_status", errno );
    return make6( vector_class, 
		    int2fx( info.st_ino ),
		    make_time_sec( info.st_mtime, t_class ),
		    make_time_sec( info.st_ctime, t_class ),
		    make_time_sec( info.st_atime, t_class ),
		    int2fx( info.st_size ),
		    int2fx( info.st_mode ) );
}

void os_setwd( const char *path )
{
int rc;

    OUT_CALL( rc = chdir( path ); );
    if (rc)
	oslib_error( "os_setwd", errno );
}

obj os_fork( void )
{
int pid;

    OUT_CALL( pid = fork(); );
    if (pid == 0)
	return FALSE_OBJ;	/* return #f to the child */
    else
	return int2fx( pid );
}

#if defined(PLATFORM_AIX) || defined(SUBPLATFORM_SOLARIS)
#define WAIT_ARG_CAST (int *)
#else
#if defined(PLATFORM_SUNOS)
#define WAIT_ARG_CAST (void *)
#else
#define WAIT_ARG_CAST (union wait *)
#endif
#endif

#ifndef PLATFORM_LINUX
PLAIN_C( pid_t wait WAIT_ARG_CAST; )
#endif

#define SIGNAL_IGNORE (void (*)(int))SIG_IGN

int os_wait( obj child )
{
int pid, child_stat;
void (*old_handler)(int);

    pid = fx2int( child );
	/* ignore ^C... */
    old_handler = (void (*)(int))signal( SIGINT, SIGNAL_IGNORE );
    pid = wait( WAIT_ARG_CAST &child_stat );
    signal( SIGINT, old_handler );
    return child_stat;
}

obj os_fork_wait( void )
{
obj x;

    x = os_fork();
    if (!EQ(x,FALSE_OBJ))
	x = int2fx( os_wait( x ) );
    return x;
}

/* convert a unix-standard pathname to local OS format */

obj os_path( obj path )
{
    return path;
}

/* open a file using unix-std pathnames */

FILE *os_fopen( const char *path, const char *mode )
{
const char *home;

    if (path[0] == '~' && path[1] == '/' && (home = getenv("HOME")))
    {
    char temp[ MAXPATHLEN ];
    
	strcpy( temp, home );
	strcat( temp, path+1 );
	return fopen( temp, mode );
    }
    else
	return fopen( path, mode );
}

obj os_time_now( obj t_class )
{
struct timeval tv;

    if (gettimeofday( &tv, NULL ) != 0)
    {
	oslib_error( "gettimeofday", errno );
    }
    return os_time( &tv, t_class );
}

/* the real-time timer... */

#ifdef TIMER_IS_MONOTONE_COUNTER
extern UINT_32 system_timeout;
#endif

void os_set_timer( UINT_32 msec )
{
#ifdef TIMER_IS_MONOTONE_COUNTER
    system_timeout = msec;
#else
struct itimerval newt;

    newt.it_interval.tv_sec = 0;
    newt.it_interval.tv_usec = 0;
    newt.it_value.tv_sec = msec / 1000;
    newt.it_value.tv_usec = (msec % 1000) * 1000;
    setitimer( ITIMER_REAL, &newt, NULL );
#endif
}

UINT_32 os_get_time_remaining( void )
{
#ifdef TIMER_IS_MONOTONE_COUNTER
    return system_timeout;
#else
struct itimerval t;

    getitimer( ITIMER_REAL, &t );
    return (t.it_value.tv_sec * 1000) + (t.it_value.tv_usec / 1000);
#endif
}


UINT_32 os_halt_timer( void )
{
#ifdef TIMER_IS_MONOTONE_COUNTER
UINT_32 t = system_timeout;

    system_timeout = 0;
    return t;
#else
struct itimerval newt, old;

    newt.it_interval.tv_sec = 0;
    newt.it_interval.tv_usec = 0;
    newt.it_value.tv_sec = 0;
    newt.it_value.tv_usec = 0;
    setitimer( ITIMER_REAL, &newt, &old );
    interrupt_flag[INTR_FLAG_TIMEOUT] = FALSE_OBJ;
    return (old.it_value.tv_sec * 1000) + (old.it_value.tv_usec / 1000);
#endif
}

#ifndef TIMER_IS_MONOTONE_COUNTER
static void timeout_handler( int sig )
{
    interrupt_flag[INTR_FLAG_TIMEOUT] = TRUE_OBJ;
    interrupt_pending = YES;
#ifdef NEED_SIGNAL_REINSTALL
    signal( SIGALRM, timeout_handler );
#endif
}
#endif

void init_os( void )
{
#ifndef TIMER_IS_MONOTONE_COUNTER
    signal( SIGALRM, timeout_handler );
#endif
}

obj os_user_config_file( obj basename )
{
char temp[MAXPATHLEN];

    sprintf( temp, "~/.%src", string_text(basename) );
    return make_string(temp);
}

obj os_library_path( void )
{
    return cons( make_string( "~/lib/es" ),
    		 cons( make_string( "/usr/local/lib/es" ),
			NIL_OBJ ) );
}

/*  
    certain functions rely on the range of random
    numbers being fairly large (they will spin
    indefinitely otherwise).
    We'd use random() if it were standard...
*/

UINT_32 os_random( void )
{
    return (rand() << 16) ^ rand();
}

#if !defined(PLATFORM_LINUX)
#ifdef __cplusplus
extern "C" { unsigned usleep( unsigned ); }
#else
extern unsigned usleep( unsigned );
#endif
#endif

void os_sleep( UINT_32 usec )
{
    usleep( usec );
}

obj os_get_user_info( obj usr )
{
struct passwd *pw;

    pw = STRING_P(usr) ? getpwnam( string_text(usr) ) 
    		       : getpwuid( fx2int(usr) );
    return make4( vector_class,
    		  make_string( pw->pw_name ),
		  int2fx( pw->pw_uid ),
		  int2fx( pw->pw_gid ),
    		  make_string( pw->pw_dir ) );
}

