/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#ifdef __cplusplus
extern "C" {
#endif
#include <signal.h>
#include <sys/types.h>
#include <sys/dir.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <unistd.h>
#ifdef __cplusplus
}
#endif

#include <string.h>
#include "runtim.h"
#include "scmtime.h"
#include "intrs.h"

/* #include "vinsns.h" */

#ifdef __cplusplus
#define PLAIN_C(decl) extern "C" { decl }
#else
#define PLAIN_C(decl) extern decl
#endif

PLAIN_C(int errno;)
PLAIN_C(char *getenv( const char * );)
PLAIN_C(char *getcwd( char *buf, size_t size );)
PLAIN_C(int getpid( void );)
PLAIN_C(int system( const char *cmd );)
PLAIN_C(int wait( int * );)
PLAIN_C(int gettimeofday( struct timeval *, struct timezone * );)
PLAIN_C(void usleep( unsigned );)
PLAIN_C(int scandir( char *path, 
		     struct direct *(*nl[]), 
		     int (*)(), 
		     int (*)() ); )

static void oslib_error( const char *fn, int rc )
{
    scheme_error( "OS library error in ~s: rc ~d",
    		  2, make_string(fn), int2fx(rc) );
}

static obj os_time( struct timeval tv, obj t_class )
{
struct scheme_time t;

    t.sec = tv.tv_sec;
    t.usec = tv.tv_usec;
    return make_time( t, t_class );
}


const char *os_getenv( const char *key )
{
const char *v;

    OUT_CALL( v = getenv( key ); );
    return v;
}

#ifdef __cplusplus
#define SCANDIR_NULL_PROC (int (*)(...))NULL
#else
#define SCANDIR_NULL_PROC NULL
#endif

obj os_scan_directory( const char *path )
{
struct direct *ent, **entries;
int i, n;
obj list;

    OUT_CALL(
	n = scandir( (char *)path, 
		     &entries, 
		     SCANDIR_NULL_PROC, 
		     SCANDIR_NULL_PROC );
	if (n < 0)
	    list = FALSE_OBJ;
	else
	{
	    list = NIL_OBJ;
	    for (i=n; i>0; )
	    {
		ent = entries[--i];
		list = cons( make_string( ent->d_name ), list );
		free( ent );
	    }
	    free( entries );
	}
    );
    return list;
}

obj os_getwd( void )
{
char *p, temp[1024];

    OUT_CALL( 
	p = getcwd( temp, 1024 ); 
    );
    if (!p)
	oslib_error( "os_getwd", errno );
    return make_string(p);
}

obj os_getpid( void )
{

    return int2fx( getpid() );
}

obj os_invoke( obj command )
{
int rc;

    OUT_CALL( rc = system( string_text( command ) ); );
    return int2fx( rc );
}

obj os_error_string( int rc )
{
extern int errno, sys_nerr;
extern char *sys_errlist[];
char temp[30];

    if (errno < sys_nerr)
	return make_string( sys_errlist[errno] );
    else
    {
	sprintf( temp, "unix error %d", errno );
	return make_string( temp );
    }
}

obj os_get_resource_usage( obj i_class )
{
struct rusage info;

    OUT_CALL( getrusage( RUSAGE_SELF, &info ); );
    return make4( vector_class, os_time( info.ru_utime, i_class ),
    				os_time( info.ru_stime, i_class ),
				int2fx( info.ru_msgsnd ),
				int2fx( info.ru_msgrcv ) );
}

bool os_file_exists_p( const char *path )
{
int rc;
struct stat info;

    OUT_CALL( rc = stat( path, &info ); );
    return rc ? NO : YES;
}

obj os_file_status( const char *path, obj t_class )
{
int rc;
struct stat info;

    OUT_CALL( rc = stat( path, &info ); );
    if (rc)
	oslib_error( "os_file_status", errno );
    return make6( vector_class, 
		    int2fx( info.st_ino ),
		    make_time_sec( info.st_mtime, t_class ),
		    make_time_sec( info.st_ctime, t_class ),
		    make_time_sec( info.st_atime, t_class ),
		    int2fx( info.st_size ),
		    int2fx( info.st_mode ) );
}

void os_setwd( const char *path )
{
int rc;

    OUT_CALL( rc = chdir( path ); );
    if (rc)
	oslib_error( "os_setwd", errno );
}

obj os_fork( void )
{
int pid;

    OUT_CALL( pid = fork(); );
    if (pid == 0)
	return FALSE_OBJ;	/* return #f to the child */
    else
	return int2fx( pid );
}

#ifdef __cplusplus
#ifdef PLATFORM_DEC
#define SIGNAL_IGNORE
#else
#define SIGNAL_IGNORE (void (*)(int))SIG_IGN
#endif
#else
#define SIGNAL_IGNORE SIG_IGN
#endif

int os_wait( obj child )
{
int pid, child_stat;
void (*old_handler)(int);

    pid = fx2int( child );
	/* ignore ^C... */
    old_handler = signal( SIGINT, (void (*)(int))SIG_IGN );
    pid = wait( &child_stat );
    signal( SIGINT, old_handler );
    return child_stat;
}

obj os_fork_wait( void )
{
obj x;

    x = os_fork();
    if (!EQ(x,FALSE_OBJ))
	x = int2fx( os_wait( x ) );
    return x;
}

void os_sleep( UINT_32 usec )
{
    usleep( usec );
}

/* convert a unix-standard pathname to local OS format */

obj os_path( obj path )
{
    return path;
}

/* open a file using unix-std pathnames */

FILE *os_fopen( const char *path, const char *mode )
{
    return fopen( path, mode );
}

obj os_time_now( obj t_class )
{
struct timeval tv;

    if (gettimeofday( &tv, NULL ) != 0)
    {
	oslib_error( "gettimeofday", errno );
    }
    return os_time( tv, t_class );
}


void os_set_timer( UINT_32 msec )
{
#ifdef TIMER_IS_MONOTONE_COUNTER
    system_timeout = msec;
#else
struct itimerval newt;

    newt.it_interval.tv_sec = 0;
    newt.it_interval.tv_usec = 0;
    newt.it_value.tv_sec = msec / 1000;
    newt.it_value.tv_usec = (msec % 1000) * 1000;
    setitimer( ITIMER_REAL, &newt, NULL );
#endif
}

UINT_32 os_get_time_remaining( void )
{
#ifdef TIMER_IS_MONOTONE_COUNTER
    return system_timeout;
#else
struct itimerval t;

    getitimer( ITIMER_REAL, &t );
    return (t.it_value.tv_sec * 1000) + (t.it_value.tv_usec / 1000);
#endif
}


UINT_32 os_halt_timer( void )
{
#ifdef TIMER_IS_MONOTONE_COUNTER
UINT_32 t = system_timeout;

    system_timeout = 0;
    return t;
#else
struct itimerval newt, old;

    newt.it_interval.tv_sec = 0;
    newt.it_interval.tv_usec = 0;
    newt.it_value.tv_sec = 0;
    newt.it_value.tv_usec = 0;
    setitimer( ITIMER_REAL, &newt, &old );
    interrupt_flag[INTR_FLAG_TIMEOUT] = FALSE_OBJ;
    return (old.it_value.tv_sec * 1000) + (old.it_value.tv_usec / 1000);
#endif
}

#ifndef TIMER_IS_MONOTONE_COUNTER
static void timeout_handler( int sig )
{
    interrupt_flag[INTR_FLAG_TIMEOUT] = TRUE_OBJ;
    interrupt_pending = YES;
}
#endif

void init_os( void )
{
#ifndef TIMER_IS_MONOTONE_COUNTER
    signal( SIGALRM, timeout_handler );
#endif
}
