// Revised:  16 Nov 1993    Mike Hewett
//
// Added:     colorset::allocate_one_object()
// Modified:  colorset::snarf_more_memory to call allocate_one_object()
//
// ---------------------------------------------------------------
// FILE:  colorset.cc
//
// This file contains those functions which are part of the color_set class
// (see the file color_set.hh for a description) that can't be inlined
// (some because they are too large, and some because we take a pointer to
// them).

#include <assert.h>
#include <iostream.h>
#include "gcserver.h"
#include "objmgr.hh"
#include "colorset.hh"
#ifdef INLINES
#include "inlines.hh"
#endif

//***********************************************************************
// color_set::color_set
// This contructor assumes that color_set is called from the youngest colorset
// to the older. 

color_set::color_set(int gen_num, int size_class_num) {
    number_of_objects = 0;
    number_of_non_black = 0;
    number_of_free = 0;
    tail_object.set_next(NULL);
    number_of_objects_per_snarf = 0;
    number_of_snarfed_objects_allocated = number_of_objects_per_snarf + 1;


// initialize the color set pointers

    gray = &tail_object;
    scan = &tail_object;
    free = &tail_object;
    old_free = &tail_object;
    white = &tail_object;
    tail = &tail_object;

    size_class = size_class_num;

    // Currently, different object managers are allocated for color_sets with
    // the same size in different generations. It should be the same, if it
    // can. However, at this stage, gc object is still under construction, and 
    // it is not easy to access different generation using get_gen() or
    // get_color_set(). Maybe, the friendship should be given to this class
    // from gc and generation classes.

    my_object_manager = new object_manager(0,size_class);

    UINT_32 object_size = 1 << size_class;
    // This size is set to the maximum object size including the header.
    size = object_size;
    my_generation = gen_num;
}

//****************************************************************
//* color_set::serve_objects_when_no_free 
//*
//* This function is called when there is no free object left in the color
//* set, and prepare free objects in the set for allocation. If gc is turned
//* off, it calls gc_flip and reclaim the white objects. If no more object
//* can be reclaimed, it fetches more memory from the system. 
void color_set::serve_objects_when_no_free(void){ 
#ifdef GC_TURN_OFF_ABLE
    // The garbage collector only gets turned off if we are allocating
    // black with a single generation.  GC is turned off after finishing
    // tracing and before flipping object lists. So if there is no free
    // objects left, first flip the GC and turn white objects into free.
    // When allocating white, the garbage collector flips
    // as soon as the collection finishes.
    if(gc_turned_off) {
	gc.gc_flip();
    }
#endif

#ifdef GENERATIONAL    
    for(int i = 1; i < NUMBER_OF_GENERATIONS; i++){
	// For each generation, it reclaims older free objects. If any objects
	// could be reclaimed, fetch_old_free returns non zero value. In that
	// case we return.
	if(gc.get_gen(i)->get_color_set(size_class)->fetch_old_free()){
	    return;
	}
    }
#endif
    // Check if there is at least some minimum amount of free memory left
    // in the system.  If not, then snarf some more.
    if (number_of_free < 1){
	snarf_more_memory();
    }
}

#ifdef ALLOCATE_BLACK
//***************************************************************************
// color_set::reclaim_garbage
//
// This routine reclaims the garbage memory for the system.
// The memory is reclaimed in an order such that the next memory
// to be allocated is memory that has never been shaded, followed by
// memory that was black, but has just now become garbage, followed
// by the memory that had been reclaimed on previous cycles of the system.
//
// This routine does not move free objects to another generation, even if the
// system is generational. It simply reclaim garbage white into free list, and
// olf free objects stays there. When allocator finds no more free objects is
// available in the younger generations, it calls fetch_old_free() and move
// the old free objects to the youngest generation for allocation.
//
// This is the version for the collector configured to allocate black.

void color_set::reclaim_garbage()
{	
    gc_object_base *temp;
    assert(scan == gray);

#ifndef NDEBUG

	// traverse the newly found garbage objects just to make sure there
	// aren't any live objects in there.  Used for debugging only.

	for(temp = white;temp!=tail;temp=temp->get_next()) {
	    assert(!temp->is_shaded());
	}
#endif

    if (white != tail) {    // There are garbage objects
	if (free == gray) {             // There are no live objects
	    if (free != white) {        // There are free objects
		// reclaim the garbage.
		white->get_previous()->set_next(tail);
		tail->get_previous()->set_next(free);
		free->set_previous(tail->get_previous());
		tail->set_previous(white->get_previous());
		white->set_previous(NULL);
		gray = white;
		scan = white;
		free = white;
		white = tail;
	    } else { // There are no live or free objects, just garbage objects
		// reclaim the garbage
		gray = white;
		scan = white;
		free = white;
		white = tail;
	    }
	} else {                        // There are live objects
	    if (free != white) {        // There are free objects
		// reclaim the garbage
		temp = free->get_previous();
		white->get_previous()->set_next(tail);
		tail->get_previous()->set_next(free);
		free->set_previous(tail->get_previous());
		tail->set_previous(white->get_previous());
		white->set_previous(temp);
		temp->set_next(white);
		free = white;
		white = tail;
	    } else { // There are no free objects
		// reclaim the garbage
		white = tail;
	    }
	}
    }

    // whiten the live objects.
	    
    if(free != gray) {  // there are live objects, so whiten them.
	if(free != tail) { // there are free objects (everything is still live)
	    temp = free->get_previous();
	    temp->set_next(tail);
	    free->set_previous(NULL);
	    gray->set_previous(tail->get_previous());
	    tail->get_previous()->set_next(gray);
	    tail->set_previous(temp);
	    white = gray;
	    gray = free;
	    scan = free;
	} else { // there are no free objects, so just move white to gray
	    white = gray;
	    free = gray;
	}
    } else { // else there are no live objects so do nothing.
	assert(free == gray);
    }
    // Now the number of free objects is the number of the non black objects
    // at the end of the former GC cycle. And the number of non black objects
    // is the number of all objects.
    number_of_free = number_of_non_black;
    number_of_non_black = number_of_objects;

#ifndef NDEBUG
    INT_32 count = 0;
    for(temp= free;temp!= white;temp=temp->get_next()){
	count++;
    }
    assert(count == number_of_free);
#endif
}
#endif // ALLOCATE_BLACK


#ifdef ALLOCATE_WHITE
//****************************************************************************
// color_set::reclaim_garbage
//
// This routine reclaims the garbage memory for the system.
// The memory is reclaimed in an order such that the next memory
// to be allocated is memory that has never been shaded, followed by
// memory that was black, but has just now become garbage, followed
// by the memory that had been reclaimed on previous cycles of the system.
//
// This is the version for allocating objects white.

void color_set::reclaim_garbage()
{	
    assert(scan == gray);
    if (white != free) {           // there is some garbage
	if (white != old_free) {   // there are some was_black_now_garbage
	    if(old_free != free) { // there are some never_been_shaded
		// The following should swap the position of
		// the was_black_now_garbage set
		// and the never_shaded set within the garbage collector's list
                //
		// first, save the two ends of the was black now garbage list.
		// One end is pointed to by the white pointer, and the other
		// end is pointed to by temp.
		gc_object_base *temp = old_free->get_previous();
	    
		if (white != gray) {   // There are some non-garbage objects

		    // splice out the was black, now garbage sub-list.

		    white->get_previous()->set_next(old_free);
		    old_free->set_previous(white->get_previous());
		    
		    // now, splice in the was black now garbage into the
		    // correct place

		    white->set_previous(free->get_previous());
		    temp->set_next(free);
		    free->get_previous()->set_next(white);
		    free->set_previous(temp);
		    white = old_free;
		} else {  
		    // there are no non-garbage objects, but there are both
		    // was_black_now_garbage and never_shaded

		    old_free->set_previous(NULL);
		    
		    // now, splice in the was black now garbage into the
		    // correct place

		    white->set_previous(free->get_previous());
		    temp->set_next(free);
		    free->get_previous()->set_next(white);
		    free->set_previous(temp);
		    white = old_free;
		    gray = old_free;
		    scan = old_free;
		}
	    } // else there is no never_been_shaded, so do nothing because
	    // the list is already in the correct order.
	} // else there are no was_black_now_garbage, so do nothing because
	// the list is already in the correct order.
#ifndef NDEBUG

	// traverse the newly found garbage objects just to make sure there
	// aren't any live objects in there.  Used for debugging only.

	for(gc_object_base *temp = white;temp!=free;temp=temp->get_next()) {
	    assert(!temp->is_shaded());
	}
#endif

	// now, reclaim all of the memory

	free = white;
	old_free = free;
    } else {
	// else do nothing, because there was no garbage.
    }    
    // whiten the live objects.
    white = gray;

    // Now the number of free objects is the number of the non black objects
    // at the end of the former GC cycle. And the number of non black objects
    // is the number of all objects.
    number_of_free = number_of_non_black;
    number_of_non_black = number_of_objects;

#ifndef NDEBUG
    INT_32 count = 0;
    for(gc_object_base *temp= free;temp!= tail;temp=temp->get_next()){
	count++;
    }
    assert(count == number_of_free);
#endif
}
#endif // ALLOCATE_WHITE

// **********************************************************************
// fetch_old_free
// 
// This routine moves old free objects to the youngest generation.
// Until the last moment, the old free objects remain in the older generation. 
// When it turns out to be in the case that no more free younger objects are
// available, this member functions is called to this color set, and the
// free object is put back to the youngest generation. If any objects is moved
// to the youngest generation, it returns 1. Otherwise, 0.
//
#ifdef ALLOCATE_BLACK
int color_set::fetch_old_free(void){
    if(free == white)
	return(0);
    else{
	gc_object_base *free_head,*free_tail;
	free_head = free;
	free_tail = white->get_previous();
	if(gray != free){// If there are objects before free.
	    free->get_previous()->set_next(white);
	    white->set_previous(free->get_previous());
	}else{ //There is no objects before free.
	    white->set_previous(NULL);
	}
	if(free == scan){
	    if(free == gray){
		gray = white;
	    }
	    scan = white;
	}
	free = white;

	int num = number_of_free;
	number_of_objects -= number_of_free;
	number_of_non_black -= number_of_free;
	number_of_free = 0;
	gc.get_gen(0)->get_color_set(size_class)
	    ->insert_free_list(free_head,free_tail,num);
	return(1);
    }
}
#endif
#ifdef ALLOCATE_WHITE
int color_set::fetch_old_free(void){
    if(free == tail)
	return(0);
    else{
	gc_object_base *free_head,*free_tail;
	free_head = free;
	free_tail = tail->get_previous();
	if(free != gray){ //If there are some objects before free.
	    free->get_previous()->set_next(tail);
	    tail->set_previous(free->get_previous());
	}else{ // We have no objects before free.
	    tail->set_previous(NULL);
	}
	if(free == old_free){
	    if(free == white){
		if(free == scan){
		    if(free == gray){
			gray = tail;
		    }
		    scan = tail;
		}
		white = tail;
	    }
	    old_free = tail;
	}
	free = tail;
	int num = number_of_free;
	number_of_objects -= number_of_free;
	number_of_non_black -= number_of_free;
	number_of_free = 0;
	gc.get_gen(0)->get_color_set(size_class)
	    ->insert_free_list(free_head,free_tail,num);
	return(1);
    }
}
#endif

//**********************************************************************
// insert_free_list
// 
// Insert free object list into this colorset. We assume that there is no
// free objects in this color set. We get the head and tail of the free
// list and the number of free objects to be inserted The routine inserts
// into the proper place and adjusts the counter. 
#ifdef ALLOCATE_BLACK
void color_set::insert_free_list(gc_object_base *free_head,
				 gc_object_base *free_tail,
				 int number_of_inserted)
{
    assert(free == white);
    assert(number_of_free == 0);
    if(free != gray){ // If there are some objects before free.
	free->get_previous()->set_next(free_head);
	free_head->set_previous(free->get_previous());
	free->set_previous(free_tail);
	free_tail->set_next(free);
    }else{ // We have no objects before free.
	free_head->set_previous(NULL);
	free_tail->set_next(free);
	free->set_previous(free_tail);
    }
    if(scan == free){
	if(gray == free){
	    gray = free_head;
	}
	scan = free_head;
    }
    free = free_head;

    number_of_objects += number_of_inserted;
    number_of_non_black += number_of_inserted;
    number_of_free = number_of_inserted;
}
#endif // ALLOCATE_BLACK
#ifdef ALLOCATE_WHITE
void color_set::insert_free_list(gc_object_base *free_head,
				 gc_object_base *free_tail,
				 int number_of_inserted)
{
    assert(free == tail);
    assert(number_of_free == 0);
    if(free != gray){ // If there are some objects before free.
	free->get_previous()->set_next(free_head);
	free_head->set_previous(free->get_previous());
	free->set_previous(free_tail);
	free_tail->set_next(free);
    }else{ // We have no objects before free.
	free_head->set_previous(NULL);
	free_tail->set_next(free);
	free->set_previous(free_tail);
    }
    if(old_free == free){
	if(white == free){
	    if(scan == free){
		if(gray == free){
		    gray = free_head;
		}
		scan = free_head;
	    }
	    white = free_head;
	}
	old_free = free_head;
    }
    free = free_head;

    number_of_objects += number_of_inserted;
    number_of_non_black += number_of_inserted;
    number_of_free = number_of_inserted;
}
#endif // ALLOCATE_WHITE

#ifdef DEBUG_LEVEL2
//***********************************************************************
// color_set::sanity_check
//
// This routine is included for debugging purposes.  Any time you
// want to make sure that the pointers into the color set are in the
// correct order, you simply call this routine, and it will print
// out the relative positions of the pointers.

void color_set::sanity_check() {
    INT_32 flag = 0;
    gc_object_base *footmp;

    cout << hex << " gray:" << gray << " scan:" << scan << " white:" << white
	 << " old_free:" << old_free << " free:" << free << " tail:" << tail 
	 << dec << endl;
    for(footmp = gray;footmp!=NULL;footmp=footmp->get_next()) {
	if (footmp == white){
	    cout << "white" << endl;
	    flag = 1;
	}
	if (footmp == gray) {
	    cout << "gray" << endl;
	    flag = 1;
	}
	if (footmp == scan) {
	    cout << "scan" << endl;
	    flag = 1;
	}
	if (footmp == old_free) {
	    cout << "old_free" << endl;
	    flag = 1;
	}
	if (footmp == free) {
	    cout << "free" << endl;
	    flag = 1;
	}
	if (footmp == tail) {
	    cout << "tail" << endl;
	    flag = 1;
	}
	if (flag) cout << "-------------------" << endl;
	flag = 0;
    }
    cout << "exiting sanity check" << endl;
}
#endif DEBUG_LEVEL2

//***********************************************************************
// color_set::allocate_one_object
// 
// This routine allocates space from a free page for one object
// by linking the object into the appropriate freelist.

void
color_set::allocate_one_object(UINT_32 object_size)
{
    gc_object_base *cur_obj, *start;

//    cout << "allocating one object: memory_area = " << (long)memory_area
//	 << "     object_size = " << object_size << endl ;

//    cout << "    snarfed:  " << number_of_snarfed_objects_allocated
//         << " of " << number_of_objects_per_snarf << endl;

//    cout << "    free = " << number_of_non_black
//	 << "    to_allocate = " << number_of_free << endl ;

    number_of_non_black++;
    number_of_free++;
    
    assert(number_of_snarfed_objects_allocated < number_of_objects_per_snarf);

    // Start is set to the (one) object allocated.
    start = (gc_object_base *)local_high_water_mark;

    // create the new object here
    cur_obj = (gc_object_base *) local_high_water_mark;

    // increment memory area to next object
    local_high_water_mark += object_size;

    assert((local_high_water_mark-1) <= max_incremental_memory_extent);

// This should be eliminated J.S.
//    cur_obj->set_next((gc_object_base *)local_high_water_mark);
    cur_obj->set_containing_list(this);
#ifdef GENERATIONAL
    cur_obj->set_gen_num(0);
    cur_obj->set_step(0);
#endif

    number_of_objects++;
    number_of_snarfed_objects_allocated++;

#ifdef ALLOCATE_BLACK
    // Link in the new memory.

    // In this incremental page-chopping scheme, cur-obj == start, so
    //   the following code, which came from the page-at-a-time
    //   chopping scheme, can probably be compressed.  I'll leave it
    //   as is for the time being.     mh:  11/17/93

    if (gray == tail) { // no objects in this size class yet
//cout << "gray == tail" << endl;
	cur_obj->set_next(tail);
	tail->set_previous(cur_obj);
	start->set_previous(NULL);
	gray = start;
	scan = start;
	free = start;
	assert(white == tail);  // white should already equal tail
	assert(tail == tail); // tail should already equal tail
    } else if (free == tail) { // no free, but some objects
//cout << "free == tail" << endl;
	assert(white == tail); // in this case, there can be no whites.
	start->set_previous(free->get_previous());
	free->get_previous()->set_next(start);
	cur_obj->set_next(tail);
	tail->set_previous(cur_obj);
	if(scan == free) {
	    scan = start; // Blacks must be to the left of the frees
	}
	free = start;
    } else if (gray == free) { // no black or gray objects.
//cout << "gray == free" << endl;
	assert(white != tail); // there must be some white objects
	start->set_previous(NULL);
	gray = start;
	scan = start;
	cur_obj->set_next(free);
	free->set_previous(cur_obj);
	free = start;
    } else if (free == white) { // no free, some white, some live
//cout << "free == white" << endl;
	assert(white != tail); // there must be some white
	assert(gray != free); // there must be some live
	start->set_previous(free->get_previous());
	free->get_previous()->set_next(start);
	cur_obj->set_next(white);
	white->set_previous(cur_obj);
	if (scan == free) {
	    scan = start;
	}
	free = start;
    } else {
	assert(0); // we should never get to this conditional
    }

#endif
#ifdef ALLOCATE_WHITE
    if (gray == tail) { // no objects in this size class yet
//cout << "WHITE: gray == tail" << endl;
	cur_obj->set_next(tail);
	tail->set_previous(cur_obj);
	start->set_previous(NULL);
	gray = start;
	scan = start;
	white = start;
	free = start;
	old_free = start;
    } else { // there are objects in this size class
//cout << "WHITE: objects in this size class" << endl;
	start->set_previous(tail->get_previous());
	tail->get_previous()->set_next(start);
	cur_obj->set_next(tail);
	tail->set_previous(cur_obj);
	if (scan == tail) {
//cout << "scan == tail" << endl;
	    scan = start;
	    white = start;
	    old_free = start;
	    free = start;
	} else if (white == tail) {
//cout << "white == tail" << endl;
	    white = start;
	    old_free = start;
	    free = start;
	} else if (old_free == tail) {
//cout << "old_free == tail" << endl;
	    old_free = start;
	    free = start;
	} else if (free == tail) {
//cout << "free == tail" << endl;
	    free = start;
	}
    }
#endif
//    cout << endl ;


}  // end of colorset::allocate_more_memory


//***********************************************************************
// color_set::snarf_more_memory
// 
// This routine allocates more memory from the operating system and places
// it into the appropriate size class.  
// It allocates NUM_PAGES_TO_SNARF_AT_A_TIME pages worth of objects.

void color_set::snarf_more_memory()
{
    INT_32 i;
    char *memory_area;
    UINT_32 object_size = 1 << size_class;
    INT_32 num_objects_snarfed;

    if (object_size > GC_PAGE_SIZE) {
	if (object_size > NUM_PAGES_TO_SNARF_AT_A_TIME * GC_PAGE_SIZE) {
	    // We are allocating a large object and that object takes more
	    // pages than NUM_PAGES_TO_SNARF_AT_A_TIME
	    num_objects_snarfed = 1;
	    number_of_objects_per_snarf = num_objects_snarfed;
	    number_of_snarfed_objects_allocated = 0;
#ifdef HARD_REAL_TIME
	    local_high_water_mark = gc.high_water_mark;
	    // object_size should be a power of 2
	    // the starting point from which to alloc 
	    memory_area = gc.high_water_mark; 
	    gc.high_water_mark += object_size;
#else
	    // we need to malloc one extra page because heap_round is going
	    // to round the pointer up to the start of the next logical page.
	    void *malloc_pointer = malloc(GC_PAGE_SIZE+object_size);
	    if (malloc_pointer == NULL) {
		cerr << "System heap exhausted.  Can not malloc additional memory." << endl;
		abort();
	    }
#ifndef NDEBUG
	    max_incremental_memory_extent = ((char *)malloc_pointer) + GC_PAGE_SIZE+object_size;
#endif

#ifndef HARD_REAL_TIME
	    gc.record_malloc_pointer(malloc_pointer);
#endif
	    local_high_water_mark = heap_round(malloc_pointer);
	    memory_area = local_high_water_mark;
#endif
	    for(i=0;i<(object_size / GC_PAGE_SIZE);i++) {
		// set the object manager for the pages used for this
		// large object.  The object manager is a 
		// multi_page object manager.
		gc.set_object_manager(memory_area+(i*GC_PAGE_SIZE),
				      new object_manager(i,
							 size_class));
	    }
#ifdef TEST_FRAGMENTATION
	    object_manager *om =
	      (gc.get_object_manager(memory_area));
	    om->num_objects_on_page = 1;
#endif	    
	} else {
	    // We are allocating large objects, but they fit in the amount
	    // of memory that we are snarfing.
	    num_objects_snarfed = NUM_PAGES_TO_SNARF_AT_A_TIME*GC_PAGE_SIZE 
	                          / object_size;
	    number_of_objects_per_snarf = num_objects_snarfed;
	    number_of_snarfed_objects_allocated = 0;
#ifdef HARD_REAL_TIME
	    local_high_water_mark = gc.high_water_mark;
	    memory_area = gc.high_water_mark;
	    gc.high_water_mark += NUM_PAGES_TO_SNARF_AT_A_TIME*GC_PAGE_SIZE;
#else
	    // we need to malloc one extra page because heap_round is going
	    // to round the pointer up to the start of the next logical page.
	    void *malloc_pointer = 
	      malloc((NUM_PAGES_TO_SNARF_AT_A_TIME+1) * GC_PAGE_SIZE);
	    if (malloc_pointer == NULL) {
		cerr << "System heap exhausted.  Can not malloc additional memory." << endl;
		abort();
	    }
#ifndef NDEBUG
	    max_incremental_memory_extent = ((char *)malloc_pointer) + (NUM_PAGES_TO_SNARF_AT_A_TIME+1) * GC_PAGE_SIZE;
#endif

#ifndef HARD_REAL_TIME
	    gc.record_malloc_pointer(malloc_pointer);
#endif
	    local_high_water_mark = heap_round(malloc_pointer);
	    memory_area = local_high_water_mark;
#endif
	    for(i=0;i<NUM_PAGES_TO_SNARF_AT_A_TIME;i++) {
		// set the object manager for the pages used for these
		// objects.
		gc.set_object_manager(memory_area+(i*GC_PAGE_SIZE),
				      new object_manager(i%(object_size/GC_PAGE_SIZE),
							 size_class));
	    }
#ifdef TEST_FRAGMENTATION
	    object_manager *om =
	      (gc.get_object_manager(memory_area));
	    om->num_objects_on_page = 1;
#endif	    
	}
    } else {
	// If we don't have any empty space left on the pages we
	//    previously allocated, allocate some more pages.
	//    Otherwise, carve one more object out of the space
	//    previously allocated.

	if (number_of_snarfed_objects_allocated >= number_of_objects_per_snarf)
        {
	    // We are allocating small objects.
	    // figure out how many small objects to allocate
	    num_objects_snarfed = NUM_PAGES_TO_SNARF_AT_A_TIME*GC_PAGE_SIZE 
	      / object_size;
#ifdef HARD_REAL_TIME
	    local_high_water_mark = gc.high_water_mark;
	    memory_area = gc.high_water_mark;
	    gc.high_water_mark += NUM_PAGES_TO_SNARF_AT_A_TIME*GC_PAGE_SIZE;
#else
	    // we need to malloc one extra page because heap_round is going
	    // to round the pointer up to the start of the next logical page.
	    void *malloc_pointer = 
	      malloc((NUM_PAGES_TO_SNARF_AT_A_TIME+1) * GC_PAGE_SIZE);
	    if (malloc_pointer == NULL) {
		cerr << "System heap exhausted.  Can not malloc additional memory." << endl;
		abort();
	    }
#ifndef NDEBUG
	    max_incremental_memory_extent = ((char *)malloc_pointer) + (NUM_PAGES_TO_SNARF_AT_A_TIME+1) * GC_PAGE_SIZE;
#endif

#ifndef HARD_REAL_TIME
	    gc.record_malloc_pointer(malloc_pointer);
#endif
	    local_high_water_mark = heap_round(malloc_pointer);
	    memory_area = local_high_water_mark;
#endif
	    for(i=0;i<NUM_PAGES_TO_SNARF_AT_A_TIME;i++) {
		// set the object manager for the pages used for these
		// large objects.  The object manager is a 
		// multi_page object manager
		gc.set_object_manager(memory_area+(i*GC_PAGE_SIZE),
				      my_object_manager);
	    }
	    number_of_objects_per_snarf = num_objects_snarfed;
	    number_of_snarfed_objects_allocated = 0;
	}
    }

    allocate_one_object(object_size);
}


#ifdef DEBUG_LEVEL2
//***********************************************************************
// color_set::in_which_color_list
//
// This routine is included for debugging purposes. It looks for the object
// in this color set.
// 
// The return value 1 means it is in black
//                  2                gray    
//                  3                white
//                  4                free
// 0 means the object is not in this list.
int color_set::in_which_color_list(gc_object_base *obj) {
    gc_object_base *ptr;
    int flag = 0;
    int i;

    for(i= 0,ptr = gray; ;ptr=ptr->get_next(),i++) {
	if (ptr == gray) {
	    flag = 2;
	}
	if (ptr == scan){
	    flag = 1;
	}
#ifdef ALLOCATE_BLACK
	if (ptr == free) {
	    flag = 4;
	}
	if (ptr == white) {
	    flag = 3;
	}
#else
	if (ptr == white) {
	    flag = 3;
	}
	if (ptr == free) {
	    flag = 4;
	}
#endif
	if (ptr == tail) {
	    flag = 0;
	    break;
	}
	if (ptr == obj){
	    break;
	}
    }
    cout << "Object position: " << i << "\n";
    return(flag);
}
#endif // 

#ifndef INLINES
#include "colorset1.ci" 
#include "colorset2.ci" 
#endif
