#ifndef gcclient_h
#define gcclient_h

#include "config.hh"
#include "langtype.h"
#include "gcserver.h"

/* This file contains prototypes for functions that the client code is     */
/* compelled to supply.  The garbage collector will call these functions.  */

/* The macro LINK_TYPE is used to conditionally compile the code   */
/* either for inlining, no inlining, or foreign function calling. */

#undef LINK_TYPE
#ifdef __cplusplus

#ifdef FOREIGN
#define LINK_TYPE extern "C"
#else

#ifdef INLINES
#define LINK_TYPE inline
#else
#define LINK_TYPE 
#endif

#endif /* of ifdef FOREIGN */

#else /* not compiling with c++ compiler */
#define LINK_TYPE
#endif /* of ifdef __cplusplus */


/* ROOT SET FUNCTIONS */

/* The interface to the root set in the client code is implemented as an     */
/* iterator over the pointers into the garbage collected heap.  A call to    */
/* root_reset sets the iterator to the first pointer in the root set, and    */
/* a call to root_next returns the next pointer from the root set and        */
/* increments the iterator.  root_next returns NULL (0) when the entire root */
/* set has been scanned.                                                     */
/*                                                                           */
/* Since it is possible for some roots to have their values change less      */
/* often than other roots, the language interface provides for three root    */
/* scanning primitives.  One lets the collector scan the stable roots first, */
/* the second scans the intermediate roots, and the last scans the unstable  */
/* roots.                                                                    */
/*                                                                           */
/* NB: It is very important that the client do *no* allocation during a call */
/* by the collector to one of the root scanning routines.                    */

LINK_TYPE void stable_root_reset( void );
LINK_TYPE gc_obj_addr stable_root_next( void );

LINK_TYPE void quasistable_root_reset( void );
LINK_TYPE gc_obj_addr quasistable_root_next( void );

LINK_TYPE void unstable_root_reset( void );
LINK_TYPE gc_obj_addr unstable_root_next( void );

/* is_root returns true if pos_ptr_addr is the address of a pointer contained
   in a root variable. */

LINK_TYPE INT_32 is_root(pos_ptr_addr ptr);


/* on_stack returns true if the object pointed to is on the stack. */

LINK_TYPE INT_32 on_stack(gc_obj_addr ptr);


/* init_gcclient is called when the garbage collector is initialized,
   so that the client can initialize any of its root parameters or
   whatever.  If this functionality is not needed, define this as a
   null function. */

LINK_TYPE void init_gcclient(void);


/* POINTER FINDING FUNCTIONS

 The function find_pointers will be called with a pointer into the user's 
 object.  The user is then responsible for iteratively calling the function 
 gc_next_object (declared in gcserver.h) with a pointer to each of the
 pointers in the pointed-to object.

 For example:

 Suppose the user object has two pointers (call them p1 and p2).  The user
 would then make the following *two* calls to gc_next_object:

	gc_next_object(p1);
	gc_next_object(p2);

*/

LINK_TYPE void find_pointers(gc_obj_addr pointer_to_client_object);
#ifdef  DEBUG_LEVEL2
LINK_TYPE
void find_pointers_from_black(gc_obj_addr pointer_to_client_object);
LINK_TYPE
void find_pointers_from_non_black(gc_obj_addr pointer_to_client_object);
#endif
/* MANAGING THE STORED INTO LIST

 Calls to the write barrier might cause us to save a pointer to one of the 
 pointers in the user's object.  Because the user's pointer might not be
 a valid C pointer, the following function is to be supplied.  It takes
 as its argument, a pointer to the user's pointer and returns a C pointer
 version of the user's pointer.

 The convention is that a gc_obj_addr of NULL means "No Object", 
 which this function may return if a word is rewritten (after the
 write barrier remembers it) with a value that is not a pointer
 type, or, in C, has value NULL. */

LINK_TYPE gc_obj_addr cast_and_deref_ptr(pos_ptr_addr ptr);

#endif
