/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "obj.h"
#include "linktype.h"
#include "vinsns.h"
#include "runtim.h"
#include <string.h>

#define arg(i) LEXREF0(i)
/* #define DEBUG_0 */

extern struct part_descr ctcprims_part;
extern struct module_descr ctcprims;

/*
    These functions work together to implement immediate-expressions
    (which are optimizations of closure-threaded code such that
    instead of holding a pointer to a closure which represents an
    expression, you hold a description of the expression itself)
*/

/*  Returns YES if the given obj is an "immediate" expression */

/*  Currently, an "immediate expression" is anything but
    a closure (since a closure is presumed to represent
    additional closure-threaded code)

   (Note that this means it is faster to represent a closure
    literal in CTC as a load from a TLV than as a %literal
    ctc fragment)
*/

bool immediate_expr( obj expr )
{
    return CLOSURE_P(expr) ? NO : YES;
}

/*  Returns the value denoted by an immediate expression  */


/*
    If the immediate expression is a <TopLevelVar>, then
    the value is the current value of the TLV (ie, a tl_ref)

    If the immediate expression is a <FixNum> with the top
    two bits `10', then the value is current value of the
    lexical variable represented by the frame and slot
    number encoded in the low 28 bits of the fixnum,
    14 bits each:

         +-----+-------------------+----------------+-----+
         | 1 0 |    frame number   |    slot        | 0 0 |
         +-----+-------------------+----------------+-----+

    Any other kind of immediate expression evaluates to itself.
*/

obj eval_immediate_expr( obj expr )
{
    if (TLV_P(expr))
    {
    obj x = tlv_value(expr);
	
	if (EQ(x,UNBOUND_OBJ))
	{
	    scheme_error( "~a: TLV is unbound", 1, tlv_name(expr) );
	    return FALSE_OBJ;
	}
	return x;
    }
    else if (OBJ_ISA_FIXNUM(expr) && ((VAL(expr) >> 30) == 2))
    {
    unsigned f;
    obj envt = REG0;
    
	f = (VAL(expr) >> 16) & 0x3FFF;
	while (f > 0)
	{
	    envt = enclosing_envt( envt );
	    f--;
	}
	return gvec_read( envt, (VAL(expr) & 0xFFFF) + SLOT(1) );
    }
    return expr;
}

#define RETURN_IMMEDIATE_EXPR(expr) do { REG0 = eval_immediate_expr(expr); \
				  	 RETURN(1); } while(0)

#define TAIL(expr) do { obj e = expr; if (immediate_expr(e))\
					{ REG0 = eval_immediate_expr(e); \
					  RETURN(1); } \
					else \
					  APPLY(1,e); } while (0)

/*  The info used to bind some registers is:
	arg(0)	minimum number of args
	arg(1)	more args than minimum permitted?
	arg(2)	binding context ("name")
*/
	
obj bind_regs( obj rtenvt )
{
int min_args = fx2int( arg(0) );
int num_args = arg_count_reg;

    if (EQ(arg(1),FALSE_OBJ))
    {
	/* rest arguments not permitted; must be exact match */
	if (num_args != min_args)
	{
	    scheme_error( "~a: exactly ~d arguments expected, ~d received",
	    		  3,
			  arg(2),
			  int2fx( min_args ),
			  int2fx( num_args ) );
	}
    }
    else
    {
    obj rest = NIL_OBJ;
    
	/* rest arguments are permitted; collect them */
	
	if (num_args < min_args)
	{
	    scheme_error( "~a: at least ~d arguments expected, ~d received",
	    		  3,
			  arg(2),
			  int2fx( min_args ),
			  int2fx( num_args ) );
	}
	
	while (num_args > min_args)
	{
	    num_args--;
	    rest = cons( reg_ref(num_args), rest );
	}
	reg_set( min_args, rest );
	num_args++;
    }
    
    /* fake the use of the BEGIN_BIND(num) macro...
       (in particular, `new_envt' is declared by BEGIN_BIND(),
       so we declare our own, but hack it specially for
       the runtime envt
    */
    {
    obj new_envt = alloc( SLOT(num_args+1), bindingenvt_class );
    unsigned i;
    
	gvec_write_fresh( new_envt, SLOT(0), rtenvt );

	switch (num_args)
	{
	    default:	for (i=10; i<num_args; i++)
			    BIND_ARG( i, REG(i) );
	    case 10:	BIND_ARG( 9, REG9 );
	    case 9:	BIND_ARG( 8, REG8 );
	    case 8:	BIND_ARG( 7, REG7 );
	    case 7:	BIND_ARG( 6, REG6 );
	    case 6:	BIND_ARG( 5, REG5 );
	    case 5:	BIND_ARG( 4, REG4 );
	    case 4:	BIND_ARG( 3, REG3 );
	    case 3:	BIND_ARG( 2, REG2 );
	    case 2:	BIND_ARG( 1, REG1 );
	    case 1:	BIND_ARG( 0, REG0 );
	    case 0:	/* do nothing */ ;
	}
	return new_envt;
    }
}

/************************** %if **************************/

PROLOGUE(if_ctc)

BEGIN_FWD(if_ctc)
  FWD_MONOTONE(if_ctc_0)
  FWD_MONOTONE(if_ctc_1)
  FWD_MONOTONE(if_ctc_2)
END_FWD(if_ctc)

MONOTONE(if_ctc_0)
{
    if (immediate_expr(arg(0)))
    {
        REG1 = eval_immediate_expr(arg(0));
	JUMP(2,if_ctc_2);
    }
    else
    {
	SAVE_CONT1(if_ctc_1);
	APPLY(1,arg(0));
    }
}

MONOTONE(if_ctc_1)
{
    REG1 = REG0;
    RESTORE_CONT1();	/* get the run-time envt. back */
    JUMP(2,if_ctc_2);    
}

MONOTONE(if_ctc_2)
{
obj branch;

    branch = EQ(REG1,FALSE_OBJ) ? arg(2) : arg(1);
    TAIL(branch);
}

EPILOGUE(if_ctc)

BEGIN_BACK(if_ctc)
  BACK_MONOTONE(if_ctc_0)
  BACK_MONOTONE(if_ctc_1)
  BACK_MONOTONE(if_ctc_2)
END_BACK(if_ctc)

static struct function_descr if_ctc_descr = {
	JUMP_TABLE( if_ctc ),
	"%if",
	&ctcprims_part };

/************************** %combo **************************/
/*
    arg(0)	head expr
    arg(1)	multi
*/

PROLOGUE(combo_ctc)

BEGIN_FWD(combo_ctc)
  FWD_MONOTONE(combo_ctc_0)
  FWD_MONOTONE(combo_ctc_1)
  FWD_MONOTONE(combo_ctc_2)
  FWD_MONOTONE(combo_ctc_3)
END_FWD(combo_ctc)

MONOTONE(combo_ctc_0)
{
obj head_expr = arg(0);

    if (immediate_expr(head_expr))
    {
	REG1 = eval_immediate_expr(head_expr);
	JUMP( 2, combo_ctc_2 );
    }
    else
    {
	SAVE_CONT1( combo_ctc_1 );
	APPLY( 1, head_expr );
    }
}

MONOTONE(combo_ctc_1)
{
    REG1 = REG0;
    RESTORE_CONT1();
    JUMP( 2, combo_ctc_2 );
}

MONOTONE(combo_ctc_2)
{
    SAVE_CONT2( combo_ctc_3 );
    APPLY( 1, arg(1) );
}

MONOTONE(combo_ctc_3)
{
obj closure;

    closure = PARTCONT_REG(1);
    RESTORE_CONT_REG();
    APPLY( arg_count_reg, closure );
}

EPILOGUE(combo_ctc)

BEGIN_BACK(combo_ctc)
   BACK_MONOTONE(combo_ctc_0)
   BACK_MONOTONE(combo_ctc_1)
   BACK_MONOTONE(combo_ctc_2)
   BACK_MONOTONE(combo_ctc_3)
END_BACK(combo_ctc)

static struct function_descr combo_ctc_descr = {
	JUMP_TABLE( combo_ctc ),
	"%combo",
	&ctcprims_part };

/************************** %procedure **************************/
/*
   PROCEDURE

    arg(0)  runtime environment

    ENCLOSING ENVT
    arg(0)  number of arguments expected (minimum)
    arg(1)  "rest" flag
    arg(2)  procedure name (or #f if anonymous)
    arg(3)  body procedure
               What to do for this procedure.
	       This position is tail

  The PROCEDURE primitive is the analogue of a closure.  The
  "template"-like information is stored in the enclosing binding
  envt, which will typically be the BE of the %lambda which created
  us.

  If the "rest" flag is not #f, then this primitive will build
  any arguments above and beyond the minimum (arg(0)) into
  a proper list (ie, it will be the empty list if there are exactly
  as many arguments provided as the minimum) and place the list
  in the last slot in the run-time binding environment created
  for the arguments.

  The procedure name is pure debugging information.
*/

PROLOGUE(procedure_ctc)

BEGIN_FWD(procedure_ctc)
    FWD_MONOTONE(procedure_ctc_0)
END_FWD(procedure_ctc)

MONOTONE(procedure_ctc_0)
{
obj rtenvt;

#ifdef STEP_DUMP
extern int do_step_dump;
extern FILE *step_dump_file;
extern void touch_step_dump_file( void );

    if (do_step_dump)
    {
	touch_step_dump_file();
	fprintf( step_dump_file, "%%procedure " );
	fprinto( step_dump_file, LEXREF1(2) );
	fprintf( step_dump_file, "\n" );
    }
#endif /* STEP_DUMP */

    rtenvt = arg(0);
    envt_reg = enclosing_envt( envt_reg );
    
    REG0 = bind_regs( rtenvt );
    TAIL( arg(3) );
} 

EPILOGUE(procedure_ctc)

BEGIN_BACK(procedure_ctc)
    BACK_MONOTONE(procedure_ctc_0)
END_BACK(procedure_ctc)

static struct function_descr procedure_ctc_descr = {
	JUMP_TABLE( procedure_ctc ),
	"%procedure",
	&ctcprims_part };




/************************** %lambda **************************/
/*
   LAMBDA

    arg(0)  the template
   
   which contains a pointer to a <BindingEnvt> that looks like:
    arg(0)  number of arguments expected (minimum)
    arg(1)  "rest" flag
    arg(2)  procedure name (or #f if anonymous)
    arg(3)  body procedure
               What to do for this procedure.
	       This position is tail
*/

PROLOGUE(lambda_ctc)

BEGIN_FWD(lambda_ctc)
    FWD_MONOTONE(lambda_ctc_0)
END_FWD(lambda_ctc)

MONOTONE(lambda_ctc_0)
{
    REG0 = make_closure( REG0, arg(0) );
    RETURN(1);
}

EPILOGUE(lambda_ctc)

BEGIN_BACK(lambda_ctc)
    BACK_MONOTONE(lambda_ctc_0)
END_BACK(lambda_ctc)

static struct function_descr lambda_ctc_descr = {
	JUMP_TABLE(lambda_ctc),
	"%lambda",
	&ctcprims_part };

/************************** %bind **************************/
/*
    arg(0)	minimum number of things
    arg(1)	more permitted?
    arg(2)	location ("name")
    arg(3)	initializers
    arg(4)	body
*/

PROLOGUE(bind_ctc)

BEGIN_FWD(bind_ctc)
  FWD_MONOTONE(bind_ctc_0)
  FWD_MONOTONE(bind_ctc_1)
END_FWD(bind_ctc)

MONOTONE(bind_ctc_0)
{
  if (immediate_expr(arg(3)))
    {
      obj rtenvt = REG0;

      REG0 = eval_immediate_expr( arg(3) );
      REG0 = bind_regs( rtenvt );
      TAIL( arg(4) );
    }
  else
    {
      SAVE_CONT1( bind_ctc_1 );
      APPLY( 1, arg(3) );		/* call the %multi initializers */
    }
}

MONOTONE(bind_ctc_1)
{
    REG0 = bind_regs( PARTCONT_REG(0) );
    RESTORE_CONT_REG();
    TAIL( arg(4) );
}

EPILOGUE(bind_ctc)

BEGIN_BACK(bind_ctc)
  BACK_MONOTONE(bind_ctc_0)
  BACK_MONOTONE(bind_ctc_1)
END_BACK(bind_ctc)


static struct function_descr bind_ctc_descr = {
	JUMP_TABLE(bind_ctc),
	"%bind",
	&ctcprims_part };

/************************** %seq **************************/
/*
    Binding Envt:
	    [0] expr-list
*/

PROLOGUE(seq_ctc)

BEGIN_FWD(seq_ctc)
  FWD_MONOTONE(seq_ctc_0)
  FWD_MONOTONE(seq_ctc_1)
  FWD_MONOTONE(seq_ctc_2)
END_FWD(seq_ctc)

MONOTONE(seq_ctc_0)
{
    REG1 = arg(0);
    if (NULL_P(REG1))
    {
	REG0 = NOVALUE_OBJ;
	RETURN(0);
    }
    JUMP( 2, seq_ctc_1 );
}

MONOTONE(seq_ctc_1)
{
obj stmt;

    if (NULL_P(pair_cdr(REG1)))
    {
	TAIL(pair_car(REG1));
    }
    stmt = pair_car(REG1);
    REG1 = pair_cdr(REG1);
    
    if (immediate_expr(stmt))
    {
	/* result is dropped on floor, and since it happens that no
	   immediate exprs have side effects, do nothing */
	JUMP( 2, seq_ctc_1 );
    }
    else
    {
	SAVE_CONT2( seq_ctc_2 );
	APPLY( 1, stmt );
    }
}

MONOTONE(seq_ctc_2)
{
    RESTORE_CONT2();
    BJUMP( 2, seq_ctc_1 );
}

EPILOGUE(seq_ctc)

BEGIN_BACK(seq_ctc)
  BACK_MONOTONE(seq_ctc_0)
  BACK_MONOTONE(seq_ctc_1)
  BACK_MONOTONE(seq_ctc_2)
END_BACK(seq_ctc)

static struct function_descr seq_ctc_descr = {
	JUMP_TABLE(seq_ctc),
	"%seq",
	&ctcprims_part };

/************************** %lex-set! **************************/
/*
    Binding Envt:
	    [0] frame
	    [1] slot
	    [2] expr
*/

PROLOGUE(lex_set_ctc)

BEGIN_FWD(lex_set_ctc)
  FWD_MONOTONE(lex_set_ctc_0)
  FWD_MONOTONE(lex_set_ctc_1)
  FWD_MONOTONE(lex_set_ctc_2)
END_FWD(lex_set_ctc)

MONOTONE(lex_set_ctc_0)
{
    if (immediate_expr(arg(2)))
    {
	REG1 = eval_immediate_expr(arg(2));
	JUMP( 2, lex_set_ctc_2 );
    }
    else
    {
	SAVE_CONT1(lex_set_ctc_1);
	APPLY( 1, arg(2) );
    }
}

MONOTONE(lex_set_ctc_1)
{
    REG1 = REG0;
    RESTORE_CONT1();
    JUMP( 1, lex_set_ctc_2 );
}

MONOTONE(lex_set_ctc_2)
{
int i;

    for (i=fx2int(arg(0)); i>0; i--)
	REG0 = enclosing_envt( REG0 );
    gvec_write( REG0, FXWORDS_TO_RIBYTES(ADD1(arg(1))), REG1 );
    REG0 = REG1;
    RETURN(1);
}

EPILOGUE(lex_set_ctc)

BEGIN_BACK(lex_set_ctc)
  BACK_MONOTONE(lex_set_ctc_0)
  BACK_MONOTONE(lex_set_ctc_1)
  BACK_MONOTONE(lex_set_ctc_2)
END_BACK(lex_set_ctc)

static struct function_descr lex_set_ctc_descr = {
	JUMP_TABLE(lex_set_ctc),
	"%lex-set!",
	&ctcprims_part };

/************************** %tl-set! **************************/
/*
    Binding Envt:
	    [0] tl-var
	    [1] expr
*/

PROLOGUE(tl_set_ctc)

BEGIN_FWD(tl_set_ctc)
  FWD_MONOTONE(tl_set_ctc_0)
  FWD_MONOTONE(tl_set_ctc_1)
  FWD_MONOTONE(tl_set_ctc_2)
END_FWD(tl_set_ctc)

MONOTONE(tl_set_ctc_0)
{
    if (immediate_expr(arg(1)))
    {
	REG0 = eval_immediate_expr(arg(1));
	JUMP( 1, tl_set_ctc_2 );
    }
    else
    {
	SAVE_CONT0(tl_set_ctc_1);
	APPLY( 1, arg(1) );
    }
}

MONOTONE(tl_set_ctc_1)
{
    RESTORE_CONT0();
    JUMP( 1, tl_set_ctc_2 );
}

MONOTONE(tl_set_ctc_2)
{
    tlv_set_value( arg(0), REG0 );
    RETURN(1);
}

EPILOGUE(tl_set_ctc)

BEGIN_BACK(tl_set_ctc)
  BACK_MONOTONE(tl_set_ctc_0)
  BACK_MONOTONE(tl_set_ctc_1)
  BACK_MONOTONE(tl_set_ctc_2)
END_BACK(tl_set_ctc)

static struct function_descr tl_set_ctc_descr = {
	JUMP_TABLE(tl_set_ctc),
	"%tl-set!",
	&ctcprims_part };

/************************** %root-ref **************************/
/*
    arg(0)   module name
    arg(1)   slot
*/

static obj *find_root( void )
{
struct module_descr **m;
UINT_32 ix;

    for (m=master_table; *m; m++)
    {
	if (strcmp(string_text(arg(0)),(*m)->name) == 0)
	{
	    ix = fx2int(arg(1));
	    if (ix < (*m)->num_roots)
	    {
#ifdef DEBUG_0
		printf( "found root: %s[%d]\n", string_text(arg(0)), ix );
#endif /* DEBUG_0 */
		return &(*m)->root_vector[ix];
	    }
	    scheme_error( "%root-ref: Module `~a' doesn't have a root# ~d", 
			  2, arg(0), arg(1) );
	}
    }
    scheme_error( "%root-ref: Module `~a' not found", 1, arg(0) );
    return NULL;
}

PROLOGUE(root_ref_ctc)

BEGIN_FWD(root_ref_ctc)
    FWD_MONOTONE(root_ref_ctc_0)
END_FWD(root_ref_ctc)

MONOTONE(root_ref_ctc_0)
{
    REG0 = *find_root();
    RETURN(1);
}

EPILOGUE(root_ref_ctc)

BEGIN_BACK(root_ref_ctc)
    BACK_MONOTONE(root_ref_ctc_0)
END_BACK(root_ref_ctc)

static struct function_descr root_ref_ctc_descr = {
	JUMP_TABLE(root_ref_ctc),
	"%root-ref",
	&ctcprims_part };


/************************** %root-set **************************/
/*
    arg(0)	module name
    arg(1)	slot
    arg(2)	expr
*/

PROLOGUE(root_set_ctc)

BEGIN_FWD(root_set_ctc)
  FWD_MONOTONE(root_set_ctc_0)
  FWD_MONOTONE(root_set_ctc_1)
  FWD_MONOTONE(root_set_ctc_2)
END_FWD(root_set_ctc)

MONOTONE(root_set_ctc_0)
{
    if (immediate_expr(arg(2)))
    {
	REG0 = eval_immediate_expr(arg(2));
	JUMP( 1, root_set_ctc_2 );
    }
    else
    {
	SAVE_CONT0(root_set_ctc_1);
	APPLY( 1, arg(2) );
    }
}

MONOTONE(root_set_ctc_1)
{
    RESTORE_CONT0();
    JUMP( 1, root_set_ctc_2 );
}

MONOTONE(root_set_ctc_2)
{
    *find_root() = REG0;
    RETURN(1);
}

EPILOGUE(root_set_ctc)

BEGIN_BACK(root_set_ctc)
  BACK_MONOTONE(root_set_ctc_0)
  BACK_MONOTONE(root_set_ctc_1)
  BACK_MONOTONE(root_set_ctc_2)
END_BACK(root_set_ctc)

static struct function_descr root_set_ctc_descr = {
	JUMP_TABLE(root_set_ctc),
	"%root-set!",
	&ctcprims_part };

/************************** %literal **************************/
/*
    Binding Envt:
	    [0] value
*/

PROLOGUE(literal_ctc)

BEGIN_FWD(literal_ctc)
    FWD_MONOTONE(literal_ctc_0)
END_FWD(literal_ctc)

MONOTONE(literal_ctc_0)
{
    REG0 = arg(0);
    RETURN(1);
}

EPILOGUE(literal_ctc)

BEGIN_BACK(literal_ctc)
   BACK_MONOTONE(literal_ctc_0)
END_BACK(literal_ctc)

static struct function_descr literal_ctc_descr = {
	JUMP_TABLE(literal_ctc),
	"%literal",
	&ctcprims_part };

/************************** %multi **************************/

#define arg_list REG1

PROLOGUE(multi_ctc)

BEGIN_FWD(multi_ctc)
  FWD_MONOTONE(multi_ctc_0)
  FWD_MONOTONE(multi_ctc_1)
  FWD_MONOTONE(multi_ctc_1r)
  FWD_MONOTONE(multi_ctc_2)
  FWD_MONOTONE(multi_ctc_2r)
  FWD_MONOTONE(multi_ctc_3)
  FWD_MONOTONE(multi_ctc_3r)
  FWD_MONOTONE(multi_ctc_n)
  FWD_MONOTONE(multi_ctc_nr)
END_FWD(multi_ctc)

MONOTONE(multi_ctc_0)
{
obj arg_expr;

    arg_list = arg(0);	/* get the arg-list from the envt */
    
    if (NULL_P(arg_list))
    {
	REG0 = NOVALUE_OBJ;
	RETURN(0);
    }
    assert( PAIR_P(arg_list) );
    arg_expr = pair_car(arg_list);
    arg_list = pair_cdr(arg_list);
    
    if (immediate_expr(arg_expr))
    {
        REG2 = eval_immediate_expr(arg_expr);
	JUMP( 3, multi_ctc_1 );
    }
    else
    {
	SAVE_CONT2( multi_ctc_1r );
	APPLY( 1, arg_expr );
    }
}

MONOTONE( multi_ctc_1r )	/* arg[0] is in REG0 */
{
    REG2 = REG0;
    RESTORE_CONT2();
    JUMP( 3, multi_ctc_1 );
}

MONOTONE( multi_ctc_1 )	/* arg[0] is in REG2 */
{
obj arg_expr;

    if (NULL_P(arg_list))
    {
	REG0 = REG2;
	RETURN(1);
    }
    assert( PAIR_P(arg_list) );
    arg_expr = pair_car(arg_list);
    arg_list = pair_cdr(arg_list);
    
    if (immediate_expr(arg_expr))
    {
        REG3 = eval_immediate_expr(arg_expr);
	JUMP( 4, multi_ctc_2 );
    }
    else
    {
	SAVE_CONT3( multi_ctc_2r );
	APPLY( 1, arg_expr );
    }
}

MONOTONE( multi_ctc_2r )	/* arg[1] is in REG0 */
{
    REG3 = REG0;
    RESTORE_CONT3();
    JUMP( 4, multi_ctc_2 );
}

MONOTONE( multi_ctc_2 )
{
obj arg_expr;

    if (NULL_P(arg_list))
    {
	REG0 = REG2;
	REG1 = REG3;
	RETURN(2);
    }
    assert( PAIR_P(arg_list) );
    arg_expr = pair_car(arg_list);
    arg_list = pair_cdr(arg_list);

    if (immediate_expr(arg_expr))
    {
        REG4 = eval_immediate_expr(arg_expr);
	JUMP( 5, multi_ctc_3 );
    }
    else
    {
	SAVE_CONT4( multi_ctc_3r );
	APPLY( 1, arg_expr );
    }
}

MONOTONE( multi_ctc_3r )	/* arg[2] is in REG0 */
{
    REG4 = REG0;
    RESTORE_CONT4();
    JUMP( 5, multi_ctc_3 );
}

MONOTONE( multi_ctc_3 )
{
obj arg_expr;

    if (NULL_P(arg_list))
    {
	REG0 = REG2;
	REG1 = REG3;
	REG2 = REG4;
	RETURN(3);
    }
    assert( PAIR_P(arg_list) );
    arg_expr = pair_car(arg_list);
    arg_list = pair_cdr(arg_list);

    if (immediate_expr(arg_expr))
    {
        REG5 = eval_immediate_expr(arg_expr);
	JUMP( 6, multi_ctc_n );
    }
    else
    {
	SAVE_CONT5( multi_ctc_nr );
	APPLY( 1, arg_expr );
    }
}

MONOTONE( multi_ctc_nr )	/* arg is in REG0 */
{
obj temp = REG0;
unsigned n;

    n = restore_arb();
    switch (n)
    {
	case 5:		REG5 = temp; break;
	case 6:		REG6 = temp; break;
	case 7:		REG7 = temp; break;
	case 8:		REG8 = temp; break;
	case 9:		REG9 = temp; break;
	default:	REG(n) = temp; break;
    }
    JUMP( n+1, multi_ctc_n );
}

MONOTONE( multi_ctc_n )
{
obj arg_expr;
unsigned i, n = arg_count_reg;

    if (NULL_P(arg_list))
    {
	assert(n >= 6);
	REG0 = REG2;
	REG1 = REG3;
	REG2 = REG4;
	REG3 = REG5;

	for (i=6; i<n; i++)
	    reg_set( i-2, reg_ref(i) );
	RETURN(n-2);
    }
    assert( PAIR_P(arg_list) );
    arg_expr = pair_car(arg_list);
    arg_list = pair_cdr(arg_list);

    if (immediate_expr(arg_expr))
    {
        reg_set( n, eval_immediate_expr(arg_expr) );
	BJUMP( n+1, multi_ctc_n );
    }
    else
    {
	SAVE_CONT( n, multi_ctc_nr );
	APPLY( 1, arg_expr );
    }
}

EPILOGUE(multi_ctc);

BEGIN_BACK(multi_ctc)
  BACK_MONOTONE(multi_ctc_0)
  BACK_MONOTONE(multi_ctc_1)
  BACK_MONOTONE(multi_ctc_1r)
  BACK_MONOTONE(multi_ctc_2)
  BACK_MONOTONE(multi_ctc_2r)
  BACK_MONOTONE(multi_ctc_3)
  BACK_MONOTONE(multi_ctc_3r)
  BACK_MONOTONE(multi_ctc_n)
  BACK_MONOTONE(multi_ctc_nr)
END_BACK(multi_ctc)

static struct function_descr multi_ctc_descr = {
	JUMP_TABLE(multi_ctc),
	"%multi",
	&ctcprims_part };

/****** raw glue `%ctc-eval' ******/

#define envt REG0
#define ctc REG1

static char ctc_eval_22_name[] = "%ctc-eval";
#define FUNCTION ctc_eval_22_name

PROLOGUE(ctc_eval_22)

BEGIN_FWD(ctc_eval_22)
  FWD_MONOTONE(ctc_eval_22_00)
END_FWD(ctc_eval_22)

MONOTONE(ctc_eval_22_00)
{
extern bool immediate_expr( obj expr );
extern obj eval_immediate_expr( obj expr );

    if (immediate_expr(ctc))
    { 
	REG0 = eval_immediate_expr(ctc); 
	RETURN(1); 
    }
    else 
	APPLY( 1, ctc );
}

EPILOGUE(ctc_eval_22)

BEGIN_BACK(ctc_eval_22)
  BACK_MONOTONE(ctc_eval_22_00)
END_BACK(ctc_eval_22)

static struct function_descr ctc_eval_22_descr = {
	JUMP_TABLE( ctc_eval_22 ),
	ctc_eval_22_name,
	&ctcprims_part };
#undef FUNCTION

#undef envt
#undef ctc

/************************** %code **************************/
/*
    This code is shared by all pieces of closure-threaded
    code, and serves as the entry point.  The template
    is expected to hold (as LITERAL(0)) a pointer to a 
    <BindingEnvt> which is used to specify the binding of arguments:

    INFO ENVT
    arg(0)  number of arguments expected (minimum)
    arg(1)  "rest" flag
    arg(2)  procedure name (or #f if anonymous)
    arg(3)  body procedure
               What to do for this procedure.
	       This position is tail

  A template whose code pointer refers to this procedure is
  a Closure-Threaded-Code Template, and serves the same role.
*/

PROLOGUE(code_ctc)

BEGIN_FWD(code_ctc)
    FWD_MONOTONE(code_ctc_0)
END_FWD(code_ctc)

MONOTONE(code_ctc_0)
{
obj rtenvt;

    rtenvt = envt_reg;
    envt_reg = LITERAL(0);

#ifdef STEP_DUMP
    {
    extern int do_step_dump;
    extern FILE *step_dump_file;
    extern void touch_step_dump_file( void );
    
	if (do_step_dump)
	{
	    touch_step_dump_file();
	    fprintf( step_dump_file, "%%code " );
	    fprinto( step_dump_file, LEXREF0(2) );
	    fprintf( step_dump_file, "\n" );
	}
    }
#endif /* STEP_DUMP */

    REG0 = bind_regs( rtenvt );
    TAIL( arg(3) );
}

EPILOGUE(code_ctc)

BEGIN_BACK(code_ctc)
    BACK_MONOTONE(code_ctc_0)
END_BACK(code_ctc)

static struct function_descr code_descr = {
	JUMP_TABLE( code_ctc ),
	"%code",
	&ctcprims_part };

/****************************************************/

extern struct function_descr finish_run_descr;
extern struct function_descr continue_intr_descr;

static struct function_descr *(fn_tab[]) = {
    &if_ctc_descr,		/* [0] */
    &multi_ctc_descr,		/* [1] */
    &combo_ctc_descr,		/* [2] */
    &literal_ctc_descr,		/* [3] */
    &procedure_ctc_descr,	/* [4] */
    &bind_ctc_descr,		/* [5] */
    &seq_ctc_descr,		/* [6] */
    &lex_set_ctc_descr,		/* [7] */
    &tl_set_ctc_descr,		/* [8] */
    &lambda_ctc_descr,		/* [9] */
    &root_ref_ctc_descr,	/* [10] */
    &root_set_ctc_descr,	/* [11] */
    &finish_run_descr,		/* [12] used by `entry.c' */
    &continue_intr_descr,	/* [13] used by `entry.c' */
    &ctc_eval_22_descr,		/* [14] */
    &code_descr,		/* [15] */
    NULL };
struct part_descr ctcprims_part = { 
    1097, &ctcprims, fn_tab, "ctcprims" };

extern struct part_descr htgeneric_part;

static struct part_descr *(tab[]) = {
    &ctcprims_part,
    &htgeneric_part,
    NULL };

#ifdef PROVIDE_ROOT_NAMES
static struct root_info root_descrs[] = {
    { 0, "$system-link" },
    { 1, "$system-initial" },
    { 3, "$symbol-table" },
    { 5, "$address-table" },
    { 8, "$backup-boots" },
    { 0, NULL } };
#else
#define root_descrs NULL
#endif

struct module_descr ctcprims = { "ctcprims", tab,
	NUM_RUNTIM_ROOTS,
	runtim_root_vector,
	root_descrs
    };
