/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
/* $Id: interim.c,v 1.4 1994/12/11 00:29:12 donovan Exp $ */

#include "runtim.h"
#include "hashmain.h"
#include "hashfn.h"
#include "gcserver.h"
#include <string.h>
#include "vinsns.h"
#ifdef ATRACE
#include "atrace.h"
#endif
#include <ctype.h>
#include "regs.h"
#include "tables_r.h"

#ifndef INLINES
#include "interim.ci"
#endif

/* #define CASE_INSENSITIVE */

obj runtim_root_vector[NUM_RUNTIM_ROOTS];

/***************** miscellaneous *****************/

bool TEMPLATE_P( obj thing )
{
    return PTR_ISA(thing,template_class);
}

bool BYTE_VECTOR_P( obj thing )
{
    return PTR_ISA(thing,byte_vector_class);
}

bool LONGFLOAT_P( obj thing )
{
    return PTR_ISA(thing,double_float_class);
}

IEEE_64 extract_float( obj thing )
{
    return *(IEEE_64 *)PTR_TO_DATAPTR(thing);
}

obj make_float( IEEE_64 x )
{
obj f = alloc2( double_float_class );

    assert( sizeof(IEEE_64) == SLOT(2) ); /* this permits use of `alloc2' */
    
    *(IEEE_64 *)PTR_TO_DATAPTR(f) = x;
    return f;
}

IEEE_64 get_float( obj arg, const char *fn, const char *arg_name )
{
    if (OBJ_ISA_FIXNUM(arg))
	return fx2int(arg);
    else if (LONGFLOAT_P(arg))
	return extract_float(arg);
    scheme_error( "~a: Invalid argument `~a' ~s", 
		      3, 
		      make_string( fn ),
		      make_string( arg_name ),
		      REG0 );
    return 0.0;
}

obj class_of( obj thing )
{
    if (OBJ_ISA_PTR(thing))
	return CLASSOF_PTR(thing);
    else if (OBJ_ISA_FIXNUM(thing))
	return fixnum_class;
    else if (OBJ_ISA_IMMOB(thing))
    {
	switch (SECONDARY_TAG(thing))
	{
	    case BOOLEAN_TAG:		return boolean_class;
	    case NIL_TAG:		return nil_class;
	    case ASCII_CHAR_TAG:	return ascii_char_class;
	    case UNICODE_CHAR_TAG:	return unicode_char_class;
	    case UNIQUE_OBJ_TAG:	return unique_obj_class;
	}
    }
    return unique_obj_class;
}

obj fluid_assq( obj key )
{
obj s = dynamic_state_reg;
obj c;

    while (PAIR_P(s))
    {
	c = pair_car(s);
	if (PAIR_P(c) && EQ(pair_car(c),key))
	    return c;
	s = pair_cdr(s);
    }
    return FALSE_OBJ;
}

/*  This is the "reverse linking" function
    (the interface to the user, anyway).
    It takes a template, and the address of an insn (monotone)
    within that template, and returns a human-readable
    string representation of the address.
*/

obj addr_to_name( obj templ, obj addr )
{
struct function_descr *f;
jump_addr *m;
unsigned i;
char temp[1024];
obj x;

    assert( !OBJ_ISA_PTR(addr) );
    assert( TEMPLATE_P(templ) );
    
    x = objecttable_lookup( address_table, addr, addr );
    if (EQ(x,FALSE_OBJ))
    {
        x = gvec_read( templ, SLOT(1) );	/* get linkage data */
	if (OBJ_ISA_PTR(x))
	{
	    /* this is an UNSWIZZLED TEMPLATE! */
	    sprintf( temp, "%s.%d[%d]",
		     string_text( gvec_read( x, SLOT(0) ) ),
		     fx2int( gvec_read( x, SLOT(1) ) ),
		     fx2int( gvec_read( templ, SLOT(0) ) ) );
	}
	else
	{
	    f = (struct function_descr *)
			OBJ_TO_RAW_PTR(gvec_read( templ, SLOT(1) ));
	    if (f)
	    {
		if (f->in_part->name)
		    sprintf( temp, "%s.%s", f->in_part->name, f->name );
		else
		    strcpy( temp, f->name );
		for (i=0, m=f->monotones; *m; m++, i++)
		    if (EQ(JUMP_ADDR_TO_OBJ(*m),addr))
		    {
			if (i)
			    sprintf( temp + strlen(temp), " [%u]", i );
			break;
		    }
		if (!*m)
		{
		    fprintf( stderr, 
			"addr_to_name error: %#x not in template %#x\n",
			VAL(addr),
			VAL(templ) );
		    sprintf( temp, " [?%#x]", VAL(addr) );
		}
	    }
	    else
		sprintf( temp, "[?%#x]", VAL(addr) );
	}
	x = make_string( temp );
	hashtable_install( address_table, addr, addr, x );
    }
    return x;
}
    

/***************** vectors *****************/

bool VECTOR_P( obj thing )
{
    return PTR_ISA(thing,vector_class);
}

obj make_empty_vector( UINT_32 length )
{
    return gvec_alloc( length, vector_class );
}

/***************** strings *****************/

bool STRING_P( obj thing )
{
    return PTR_ISA(thing,string_class);
}

const char *string_text( obj str )
{
    return (const char *)PTR_TO_DATAPTR(str);
}

bool string_eq( obj str1, obj str2 )
{
INT_32 n = SIZEOF_PTR(str1);
const UINT_32 *s1 = (const UINT_32 *)PTR_TO_DATAPTR(str1);
const UINT_32 *s2 = (const UINT_32 *)PTR_TO_DATAPTR(str2);

    if (n != SIZEOF_PTR(str2))
	return NO;

    while (n > 0)
    {
	if (*s1++ != *s2++)
	    return NO;
	n -= 4;
    }
    return YES;
}

bool string_ci_eq( obj str1, obj str2 )
{
    return (string_ci_cmp( str1, str2 ) == 0) ? YES : NO;
}

int string_cmp( obj str1, obj str2 )
{
    return strcmp( string_text(str1), string_text(str2) );
}

int string_ci_cmp( obj str1, obj str2 )
{
const unsigned char *s1 = (const unsigned char *)string_text(str1);
const unsigned char *s2 = (const unsigned char *)string_text(str2);
unsigned char c1, c2;

    while (1)
    {
        c1 = *s1++;
	c2 = *s2++;
	if (isupper(c1))
	  c1 = tolower(c1);
	if (isupper(c2))
	  c2 = tolower(c2);
	if (c1 < c2)
	  {
	    return -1;
	  }
	else if (c1 > c2)
	  {
	    return 1;
	  }
	else if (!c1)
	  {
	    return 0;
	  }
    }
}

obj make_string( const char *text )
{
size_t len;
obj str;

    if (text)
    {
	len = strlen(text);
	str = bvec_alloc( len+1, string_class );
	memcpy( PTR_TO_DATAPTR(str), text, len );
	return str;
    }
    else
	return FALSE_OBJ;
}

obj string_hash( obj str )
{
    assert( STRING_P(str) );
    return raw_bytes_hash( PTR_TO_DATAPTR(str), SIZEOF_PTR(str) );
}

UINT_32 string_length( obj str )
{
    assert( STRING_P(str) );
    return SIZEOF_PTR(str) - 1;
}

/* we cast the string text pointer to an 'unsigned char *'
   so that characters we extract do not get sign extended
                                          -- dmk 94.06.21 */

obj string_ref( obj str, UINT_32 index )
{
    assert( index < string_length(str) );
    return MAKE_ASCII_CHAR( ((unsigned char *)string_text(str))[index] );
}

void string_set( obj str, UINT_32 index, obj ch )
{
    assert( index < string_length(str) );
    assert( OBJ_ISA_ASCII_CHAR(ch) );
    ((char *)string_text(str))[index] = GET_IMMEDIATE_VALUE(ch);
}

void string_lower_case( obj str )
{
char *p;

    p = (char *)string_text(str);
    while (*p)
    {
	if (isupper(*p))
	    *p = tolower(*p);
	p++;
    }
}

/***************** symbols *****************/

bool SYMBOL_P( obj thing )
{
    return PTR_ISA(thing,symbol_class);
}

/*
    Note that the hash value for a <Symbol>
    is the rehash of the hash value for the
    string mapping to the symbol
*/

obj intern( obj str )
{
obj hash = string_hash(str);
obj value;


    assert( STRING_P(str) );
    assert( OBJ_ISA_FIXNUM(hash) );

    value = stringtable_lookup( symbol_table, hash, str );
    if (EQ(value,FALSE_OBJ))
    {
	value = make2( symbol_class, str, rehash_fixnum( hash ) );
	hashtable_install( symbol_table, hash, str, value );
    }
/* printf( "intern: %s ==> %#x\n", string_text(str), VAL(value) );
*/
    return value;
}

obj lookup_symbol( const char *str )
{
obj ss = make_string(str);

#ifndef CASE_SENSITIVE
    string_lower_case( ss );
#endif
    return intern( ss );
}

obj symbol_hash( obj symbol )
{
    assert( SYMBOL_P(symbol) );
    return gvec_read( symbol, SLOT(1) );
}

const char *symbol_text( obj symbol )
{
    assert( SYMBOL_P(symbol) );
    return string_text( gvec_read(symbol,SLOT(0)) );
}

obj symbol_str( obj symbol )
{
    assert( SYMBOL_P(symbol) );
    return gvec_read(symbol,SLOT(0));
}


/***************** top-level environments & vars *****************/

bool TLE_P( obj thing )
{
    return PTR_ISA(thing,tle_class);
}


obj make_tlv( obj name )
{
    return make2( tlv_class, name, UNBOUND_OBJ );
}

obj tle_lookup_tlv( obj tle, obj name )
{
obj h, v, tbl;

    assert( TLE_P(tle) );
    tbl = gvec_read( tle, SLOT(0) );
    h = symbol_hash( name );
    v = objecttable_lookup( tbl, h, name );
    if (EQ(v,FALSE_OBJ))
    {
	v = make_tlv( name );
	hashtable_install( tbl, h, name, v );
    }
    return v;
}

obj tlv_name( obj tlv )
{
    assert( TLV_P(tlv) );
    return gvec_read( tlv, SLOT(0) );
}

obj tlv_value( obj tlv )
{
obj v;

    assert( TLV_P(tlv) );
    v = gvec_read( tlv, SLOT(1) );
    if (EQ(v,UNBOUND_OBJ))
    {
	scheme_error( "~a: not bound", 1, tlv );
    }
    return v;
}

void tlv_set_value( obj tlv, obj value )
{
    assert( TLV_P(tlv) );
    gvec_write( tlv, SLOT(1), value );
}


/***************** classes *****************/

bool CLASS_P( obj thing )
{
    return PTR_ISA(thing,class_class);
}

obj make_class( obj name, bool is_gvec )
{
obj c = gvec_alloc( 28, class_class );

    assert( EQ(name,FALSE_OBJ) || SYMBOL_P(name) );
    gvec_write_fresh( c, SLOT(0), name );
    gvec_write_fresh_non_ptr( c, SLOT(1), is_gvec ? ZERO : int2fx(1) );
/*gvec_write_fresh_non_ptr( c, SLOT(1), is_gvec ? FALSE_OBJ : TRUE_OBJ );*/
    gvec_write_fresh_non_ptr( c, SLOT(2), ZERO );
    return c;
}

unsigned class_image_mode( obj a_class )
{
    return fx2int( gvec_read( a_class, SLOT(2) ) );
}

bool class_is_gvec( obj a_class )
{
    assert( CLASS_P(a_class) );
    return EQ(gvec_read(a_class,SLOT(1)),FALSE_OBJ);
}

obj class_name( obj a_class )
{
    assert( CLASS_P(a_class) );
    return gvec_read( a_class, SLOT(0) );
}

obj class_supers( obj a_class )
{
    assert( CLASS_P(a_class) );
    return gvec_read( a_class, SLOT(3) );
}

bool subclass_p( obj class1, obj class2 )
{
    while (1)
    {
	if (EQ(class1, class2))
	    return YES;
    	class1 = class_supers(class1);
	if (!PAIR_P(class1))
	    return NO;
	class1 = pair_car( class1 );	/* only look up the first chain */
    }
}

bool obj_isa_instance_of( obj thing, obj a_class )
{
    return subclass_p( class_of(thing), a_class );
}

/** INITIALIZATION **/

static void add_system_def( obj name, obj value )
{
    hashtable_install( system_link_table, 
    			string_hash(name), 
			name, 
			value );
}

static void system_def_class( obj the_class, const char *name )
{
#if 1
    add_system_def( make_string( name ), the_class );
#else
    add_system_def( symbol_str( class_name( the_class ) ),
    	 	    the_class );
#endif
}

void init_runtim( bool verbose )
{
struct module_descr **p;
unsigned n;
obj *r;
char *gc_argv[3];

    /* set all roots to FALSE */
    
    for (p=master_table; *p; p++)
    {
	r = (*p)->root_vector;
	n = (*p)->num_roots;
	while (n > 0)
	{
	    *r++ = FALSE_OBJ;
	    n--;
	}
    }
    
    gc_argv[0] = "rs";
    gc_argv[1] = verbose ? NULL : "-q";
    gc_argv[2] = NULL;
    init_gc( verbose ? 1 : 2, gc_argv );

    init_stack_cache();
#ifdef ATRACE
    atrace_init();
#endif
    class_class = ZERO;
    class_class = make_class( FALSE_OBJ, YES );

    partcont_class = ZERO;
    
    {
    obj *dest = &PTR_TO_HDRPTR(class_class)->pob_class;
    obj value = class_class;
    
	write_barrier_NDP( PTR_TO_GCPTR(class_class),
			    dest,
			    PTR_TO_GCPTR(value) );
	*dest = value;
    }
    
    string_class = make_class( FALSE_OBJ, NO );

    table_class = make_class( FALSE_OBJ, YES );
    string_ci_table_class = make_class( FALSE_OBJ, YES );
    string_table_class = make_class( FALSE_OBJ, YES );
    table_bucket_class = make_class( FALSE_OBJ, YES );

    vector_class = make_class( FALSE_OBJ, YES );
    symbol_class = make_class( FALSE_OBJ, YES );
#ifndef CASE_SENSITIVE
    symbol_table = make_table( string_ci_table_class, 10 );
#else
    symbol_table = make_table( string_table_class, 10 );
#endif
    
    eq_table_class = make_class( lookup_symbol("<eq-table>"), YES );
    object_table_class = make_class( lookup_symbol("<object-table>"), YES );
    
    system_link_table = make_table( string_table_class, 4 );
    address_table = make_table( eq_table_class, 5 );

    gvec_write( class_class, SLOT(0), lookup_symbol("<<Class>>") );
    system_def_class( class_class, "<<Class>>" );

    gvec_write( string_class, SLOT(0), lookup_symbol("<String>") );
    
    gvec_write( table_class, SLOT(0), lookup_symbol("<table>") );
    
	gvec_write( string_ci_table_class,
		    SLOT(0), 
		    lookup_symbol("<string-ci-table>") );
	gvec_write( string_table_class,
		    SLOT(0), 
		    lookup_symbol("<string-table>") );
    
    gvec_write( table_bucket_class, SLOT(0), lookup_symbol("<table-bucket>") );
    
    gvec_write( vector_class, SLOT(0), lookup_symbol("<Vector>") );
    gvec_write( symbol_class, SLOT(0), lookup_symbol("<Symbol>") );
    
    system_def_class( string_class, "<String>" );
    system_def_class( table_class, "<table>" );
    system_def_class( string_table_class, "<string-table>" );
    system_def_class( string_ci_table_class, "<string-ci-table>" );
    system_def_class( eq_table_class, "<eq-table>" );
    system_def_class( object_table_class, "<object-table>" );
    system_def_class( table_bucket_class, "<table-bucket>" );
    system_def_class( vector_class, "<Vector>" );
    system_def_class( symbol_class, "<Symbol>" );
    
    pair_class = make_class( lookup_symbol("<Pair>"), YES );
    system_def_class( pair_class, "<Pair>" );
    
    tle_class = make_class( lookup_symbol("<TopLevelEnvt>"), YES );
    system_def_class( tle_class, "<TopLevelEnvt>" );
    
    tlv_class = make_class( lookup_symbol("<TopLevelVar>"), YES );
    system_def_class( tlv_class, "<TopLevelVar>" );
    
    bindingenvt_class = make_class( lookup_symbol("<BindingEnvt>"), YES );
    system_def_class( bindingenvt_class, "<BindingEnvt>" );
/*
    system_def_class( make_class( lookup_symbol("<Method>"), YES ) );
    system_def_class( make_class( lookup_symbol("<Slot-Descriptor>"), YES ) );
    system_def_class( make_class( lookup_symbol("<Object>"), YES ) );
*/    
    template_class = make_class( lookup_symbol("<Template>"), YES );
    system_def_class( template_class, "<Template>" );
    
    closure_class = make_class( lookup_symbol("<Closure>"), YES );
    system_def_class( closure_class, "<Closure>" );

/* experimental... */

    double_float_class = make_class( lookup_symbol("<LongFloat>"), NO );
    system_def_class( double_float_class, "<LongFloat>" );

#ifdef ATRACE
    atrace_flush(0);
#endif
}


#ifdef PLATFORM_IS_LITTLE_ENDIAN
#include "ntohd.ci"
#endif
