#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;;option  mnemonic: description
;;------  ------------------------
;;    ~a  any: display the argument (as for humans).
;;    ~s  slashified: write the argument (as for parsers).
;;    ~d  decimal: the integer argument is output in decimal format.
;;    ~x  hexadecimal: the integer argument is output in hexadecimal format.
;;    ~c  character: the next argument is displayed as a character.
;;    ~_  space: output a space character.
;;    ~%  newline: output a newline character.
;;    ~~  tilde: output a tilde.
;;    ~t  tab: output a tab character.
;;    ~>  tab: the next arg is an integer # of tab stops.
;;    ~b  binary: the integer argument is output in binary format.
;;    ~o  octal: the integer argument is output in octal format.
;;    ~p  plural: if the argument is greater than 1, print a lower case 's'.
;;    ~h  hash: create a user hash number for the arg object
;;    ~r  raw: the 32 bits of the value are printed in hex
;
; future:
;;    ~g  glorify: pretty print the argument (typically an s-expression).

(define (format/a port arg-list)
    (if (null? arg-list)
	#f
	(begin
	    (%display (car arg-list) port)
	    (cdr arg-list))))

(define (format/h port arg-list)
    (if (null? arg-list)
	#f
	(begin
	    (%write-str port "#*")
	    (%write-int port (get-user-hash (car arg-list)))
	    (cdr arg-list))))

(define (format/p port arg-list)
    (if (null? arg-list)
	#f
	(begin
	    (if (> (car arg-list) 1)
		(%write-char port #\s))
	    (cdr arg-list))))

(define (format/s port arg-list)
    (if (null? arg-list)
	#f
	(begin
	    (%write (car arg-list) port)
	    (cdr arg-list))))

(define (format/d port arg-list)
    (if (null? arg-list)
	#f
	(begin
	    (write-dec (car arg-list) port)
	    (cdr arg-list))))


(define (format/x port arg-list)
    (if (null? arg-list)
	#f
	(begin
	    (write-hex (car arg-list) port)
	    (cdr arg-list))))

(define (format/r port arg-list)
    (if (null? arg-list)
	#f
	(begin
	    (%write-raw port (car arg-list))
	    (cdr arg-list))))

(define (format/b port arg-list)
    (if (null? arg-list)
	#f
	(begin
	    (write-bin (car arg-list) port)
	    (cdr arg-list))))

(define (format/o port arg-list)
    (if (null? arg-list)
	#f
	(begin
	    (write-oct (car arg-list) port)
	    (cdr arg-list))))

(define (format/c port arg-list)
    (if (null? arg-list)
	#f
	(begin
	    (%write-char port (car arg-list))
	    (cdr arg-list))))

(define (format/> port arg-list)
    (if (null? arg-list)
	#f
	(begin
	    (write-tab4 (car arg-list) port)
	    (cdr arg-list))))

(define (format/_ port arg-list)
    (%write-char port #\space)
    arg-list)

(define (format/% port arg-list)
    (%write-char port #\newline)
    arg-list)

(define (format/~ port arg-list)
    (%write-char port #\~)
    arg-list)

(define (format/t port arg-list)
    (%write-char port #\tab)
    arg-list)

; Note:  this is arranged so that users
; can create their own formatter functions (a la NeXTSTEP)
; Since you want scope control, a global access point is not
; desired.  Instead, there is a function "%format" which takes
; a list of formatters, which format calls with the 
; standard list.  Then you can wrap that up with whatever list
; you want! (plus, *default-formatters* is the default list,
; so you don't have to rewrite the standard stuff)
;

(define *default-formatters*
    (list (cons #\a format/a)
	  (cons #\s format/s)
	  (cons #\d format/d)
	  (cons #\x format/x)
	  (cons #\b format/b)
	  (cons #\o format/o)
	  (cons #\c format/c)
	  (cons #\_ format/_)
	  (cons #\% format/%)
	  (cons #\~ format/~)
	  (cons #\t format/t)
	  (cons #\> format/>)
	  (cons #\p format/p)
	  (cons #\h format/h)
	  (cons #\r format/r)))

(define (%format dest format-str arg-list formatters)
    (type-check format-str string? <String>)
    (let ((port (if dest (if (eq? dest #t) 
			     (current-output-port) 
			     dest)
			 (open-output-string))))
	(let loop ((control (parse-format-string format-str)) 
		   (args arg-list))
	    (if (pair? control)
		(if (string? (car control))
		    (begin
			(%write-str port (car control))
			(loop (cdr control) args))
		    (let ((proc (assq (car control) formatters)))
			(if proc
			    (let ((args ((cdr proc) port args)))
				(if args
				    (loop (cdr control) args)
				    (abort 'format
					   "Missing argument for ~~~a"
					   (car control))))
			    (abort 'format
				   "Unrecognized control code ~~~a"
				   (car control)))))))
	(if dest
	    (values)
	    (close-output-port port))))

(%define-glue (parse-format-string str)
{
obj substr, prev, first, next;
const char *begin, *s;

    prev = first = cons( FALSE_OBJ, NIL_OBJ );
    begin = s = string_text(REG0);
    while (*s)
    {
        if (s[0] == '~' && s[1])
	{
	    if (begin != s)
	    {
		substr = bvec_alloc( s - begin + 1, string_class );
		memcpy( PTR_TO_DATAPTR(substr), (void*)begin, s - begin );
		next = cons( substr, NIL_OBJ );
		gvec_write_fresh_ptr( prev, SLOT(1), next );
		prev = next;
	    }
	    next = cons( MAKE_ASCII_CHAR( s[1] ), NIL_OBJ );
	    gvec_write_fresh_ptr( prev, SLOT(1), next );
	    prev = next;
	    s += 2;
	    begin = s;
	}
	else
	    s++;
    }
    if (begin != s)
    {
	substr = bvec_alloc( s - begin + 1, string_class );
	memcpy( PTR_TO_DATAPTR(substr), (void*)begin, s - begin );
	next = cons( substr, NIL_OBJ );
	gvec_write_fresh_ptr( prev, SLOT(1), next );
    }
    REG0 = pair_cdr( first );
    RETURN(1);
})

(define (format dest format-str . arg-list)
    (%format dest format-str arg-list *default-formatters*))

(export format %format *default-formatters* parse-format-string)
