#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;; some shorthand for making tokens...

(define-syntax (make-token type data)
    (make <Token> type: type data: data))
    
(define-syntax (make-triv-token type)
    (make <Token> type: type data: #f))

;; some general facilities...

(define (lexical-error port message . args)
    (fluid-let ((*PORT* port)
    		(*rest-of-line* (read-line port)))
	(apply abort 'scan-token message args)))

;; the defining properties for identifiers.
;; may want to make these faster...

;; is the character 
(define (valid-id-initial? ch)
    (or (char-alphabetic? ch)
	(char-numeric? ch)
	(memq ch '(#\! #\$ #\% #\& #\* #\/ #\: #\< #\= #\> #\?
	           #\~ #\_ #\^))))

(define (valid-id-continued? ch)
    (or (valid-id-initial? ch)
	(memq ch '(#\- #\+ #\.))))

; this procedure returns #t if the character is whitespace
; whitespace is defined as a space, newline, or tab
(define (white-space? ch)
    (memq ch '(#\space #\newline #\tab)))


;; this procedure returns #t if character is a valid delimeter
;; a delimiter is defined as <whitespace> | ( | ) | " | ;
(define (delimiter? ch)
    (or (white-space? ch) 
	(memq ch '(#\( #\) #\" #\;))
	(eof-object? ch)))

(define (not-delimiter? ch)
    (not (delimiter? ch)))

; this converts a string into a number by using string->number

(define (number-string->number-token number p)
    (let ((n (string->number number)))
	(if n
	    (make-token '<literal> n)
	    (lexical-error p "Badly formatted number: ~s" number))))


;; This procedure scans comments.  It reads everything until EOL or EOF is 
;; reached.  Once it reaches EOL or EOF, it calls scan-token again to parse the
;; next token.

(define (scan-comment p c)
    (let loop ()
	(let ((ch (read-char p)))
	    (if (or (eq? ch #\newline) (eof-object? ch))
		(scan-token p)
                (loop)))))

;; this is used to scan the "." character.  If the "." is followed by a 
;; whitespace, it returns a <dot> token.  Otherwise, it calls scan-number to
;; scan a number.

(define (scan-dot p c)
    (let ((ch (peek-char p)))
       (if (white-space? ch)
	   (make-triv-token '<dot>)
	   (scan-number p c))))

;; This is used to scan multi-line LISP style comments.  It uses a helper 
;; procedure called scan-extended-comment-helper to help it parse imbedded
;; comments.  Imbedded comments are parsed thru recursion.
(define (scan-extended-comment p c)
    (let loop ()
        (let ((ch (read-char p)))
	    (cond 
	       ; nested comment
	       ((eq? ch #\#)
		   (if (eq? (read-char p) #\|)
		       (begin (scan-extended-comment-helper p) (loop))
		       (loop)))

	       ; end of comment
               ((eq? ch #\|)
		   (if (eq? (peek-char p) #\#)
		       (begin (read-char p) (scan-token p))
		       (loop)))

		; EOF
		((eof-object? ch) (scan-token p))

		(else (loop))))))

; This procedure is scan-extended-comment procedure's helper.  It is used to 
; recursively parse imbedded comments

(define (scan-extended-comment-helper p) 
    (let loop ()
	(let ((ch (read-char p)))
	   (cond
	      ((eof-object? ch) (scan-token p))
	      ((eq? ch #\|)
	        (if (eq? (read-char p) #\#)
		    ()
		    (loop)))
              (else (loop))))))

;; This scans in a number.  It can handle ints, reals, and other number
;; systems.  It works by collectng all the valid characters & converts them 
;; a token by calling chars->number-token.  chars->number-token uses
;; string->integer to convert the characters into a number.
(define (scan-number p c)
    (let ((number (list->string (collect p c 
	 (lambda (ch) 
	    (or (char-numeric? ch) 
		(memq ch '(#\a #\b #\c #\d #\e #\f #\A #\B #\C #\D #\E #\F  
			   #\i #\o #\x))
		(eq? ch #\.)))))))
         (if (string-ci=? number "...")
	     (make-token '<symbol> '...)
	     (number-string->number-token number p))))

;; This procedure is used to scan the special math operators + and -.
;; There are two possible cases of the + and - signs:
;; 1. It is the mathmatical operator addition or subtraction, in which case it
;;    must be followed by white space.  In this case, we return the symbol + or
;;    -. 
;; 2. It can prefix a number, indicating the sign of the number.  In this case
;;    we call scan-number.

(define (scan-math-operator p c)
    (let ((ch (peek-char p)))
      (if (delimiter? ch)
          (list->symbol-token (list c))
          (scan-number p c))))

(define (scan-string p c)
    (let (((chars <seq>) (make-seq)))
	(define (unterminated)
	    (lexical-error p "Unterminated string: ~s" (seq->list chars)))
	(let loop ()
	    (let ((c (read-char p)))
		(if (char? c)
		    (if (eq? c #\\)
			;; an escape character.. process it
			(let ((c (read-char p)))
			    (case c
				((#\n) (seq-add! chars #\newline))
				((#\t) (seq-add! chars #\tab))
				(else (if (char? c)
					  (seq-add! chars c)
					  (unterminated))))
			    (loop))
			;; not an escape character...
			(if (eq? c #\")
			    ;; but the end of the string, so make the
			    ;; characters into a string
			    (make-token '<literal> 
					(list->string (seq->list chars)))
			    ;; not an escape or the EOS, so a regular char
			    (begin
				(seq-add! chars c)
				(loop))))
		    ;; not a character at all; must be EOF
		    (unterminated))))))

(define (string-ci-assoc str lst)
    (let loop ((l lst))
	(if (null? l)
	    #f
	    (let ((entry (car l)))
		(if (string-ci=? str (car entry))
		    entry
		    (loop (cdr l)))))))

(define $special-sharps
    '(("t" . #t)
      ("f" . #f)
      ("rest" . '#rest)
      ("key" . '#key)
      ("undef" . '#undef)
      ("uninit" . '#uninit)
      ("unbound" . '#unbound)
      ("none" . '#none)))

;; This scans all tokens that begin w/ a #<keyword>
(define (special_sharp p)
    (let ((ch (read-char p)))
       (if (eq? ch #\()
	   (make-triv-token '<open-vector>)
           (let* ((sometoken (list->string (collect p ch not-delimiter?)))
	   	  (maps-to (string-ci-assoc sometoken $special-sharps)))
	      (if maps-to
		  (make-token '<literal> (cdr maps-to))
                  (lexical-error p "Bad special: #~a" sometoken))))))

(define $special-chars '(("nul" . 0) ("soh" . 1) ("stx" . 2) ("etx" . 3) 
			("eot" . 4) ("enq" . 5) ("ack" . 6) ("bel" . 7) 
			("bs" . 8) ("tab" . 9) ("newline" . 10) ("vt" . 11) 
			("np" . 12) ("cr" . 13) ("so" . 14) ("si" . 15) 
			("dle" . 16) ("dc1" . 17) ("dc2" . 18) ("dc3" . 19) 
			("dc4" . 20) ("nak" . 21) ("syn" . 22) ("etb" . 23) 
			("can" . 24) ("em" . 25) ("sub" . 26) ("esc" . 27) 
			("fs" . 28) ("gs" . 29) ("rs" . 30) ("us" . 31) 
			("space" . 32)))

;; This scans all special symbols that begin w/ #\
;; i.e. char consts.  e.g. #\c #\space

(define (scan-char-consts sometoken p)
    (make-token '<literal> (integer->char (grok-char-literal* sometoken p))))
    
;; grok-char-literal* returns the ASCII code
;; of the character grokked.

(define (grok-char-literal* sometoken p)
    (if (eq? (string-length sometoken) 1) 
	(%bvec-read sometoken 0)
	(let ((maps-to (string-ci-assoc sometoken $special-chars)))
	    (if maps-to
		maps-to
		(lexical-error p "Bad special: \#\\~s" sometoken)))))

(define (grok-char-literal sometoken p)
    (if (and (> (string-length sometoken) 2)
    	     (eq? (string-ref sometoken 1) #\-)
	     (char-ci=? (string-ref sometoken 0) #\M))
	(+ (grok-char-literal* (substring sometoken 2) p)
	   128))
	(grok-char-literal* sometoken p))

;; scan-special is used to parse everything that begins with a "#".
(define (scan-special p c)
    (let ((ch (peek-char p)))
        (cond
	   ; other number systems
	   ((memq ch '(#\e #\i #\b #\o #\d #\x)) 
	        (scan-number p c))

	   ; extended (multi-line) comments
           ((eq? ch #\|)
	        (read-char p) 
	        (scan-extended-comment p ch))

	   ; object references
	   ((eq? ch #\*)
	        (make-token '<obj-ref>
			    (string->number 
			     (list->string 
			      (collect p (read-char p) char-numeric?)))))

	   ; character constants
	   ((eq? ch #\\) 
		(read-char p) 
		(set! ch (read-char p)) 
		(scan-char-consts (list->string (collect p ch not-delimiter?)) 
				  p))

	   ; special # tokens
	   (else (special_sharp p)))))


;; convert a list of characters to a symbol token object
;; note that we downcase the characters (YUCK; should have
;; a flag to say if we are being case sensitive or not)

(define (list->symbol-token chs)
    (make-token '<symbol>
    		(string->symbol
		    (list->string
			(map char-downcase chs)))))

;; collect characters from a port p (the character c
;; has already been seen, so is added to the beginning)
;; as long as the more? predicate is true of the 
;; character

(define (collect p c more?)
    (let (((s <seq>) (make-seq)))
	(seq-add! s c)
	(let loop ()
	    (let ((c (peek-char p)))
		(if (and (char? c) (more? c))
		    (begin
			(read-char p)
			(seq-add! s c)
			(loop))
		    (seq->list s))))))

(define (scan-ident p c)
    (list->symbol-token 
	(collect p c valid-id-continued?)))

(define *scanners* #f)

#| initialize the *scanner-table* top-level-variable
   to be a 256-element vector containing all ASCII characters from ASCII  0 to
   ASCII 255.  Each element in the vector contains pointers to the appropriate 
   procedure to scan that character.  This vector essentially maps a
   character to the function that correctly parse that character.

   Mathmatically this is stated: 
	f(character)=function to correctly parse that character.

	e.g. f(a)=function to parse identifier
	     f(3)=function to parse number

   Scanner algorithm:
   We read in a character.  Depending on what character is read in, we then 
   call the procedure to correctly parse that character.  
|#

(define (init-scanners)
    (set! *scanners* 
	    (make-vector 256 
			 (lambda (p c) 
				(abort 'scan-token 
				       "Lexical error at: ~s" 
				       c))))
    (let loop ((i 0))
	(if (< i 256)
	    (begin
		(if (valid-id-initial? (integer->char i))
		    (vector-set! *scanners* i scan-ident))
		(loop (+ i 1)))))
    (init #\0 #\9 scan-number)
    (init1 #\. scan-dot)
    (init1 #\" scan-string)
    (init1 #\# scan-special)
    (init1 #\; scan-comment)
    (inits '(#\+ #\-) scan-math-operator)
    
    (init1 #\( (lambda (p c) (make-triv-token '<open-paren>)))
    (init1 #\) (lambda (p c) (make-triv-token '<close-paren>)))
    (init1 #\' (lambda (p c) (make-triv-token '<quote>)))
    (init1 #\` (lambda (p c) (make-triv-token '<quasiquote>)))
    (init1 #\, (lambda (p c) 
		(if (eq? (peek-char p) #\@)
		    (begin
			(read-char p)
			(make-triv-token '<unquote-splicing>))
		    (make-triv-token '<unquote>))))

    (inits '(#\space #\newline #\tab)
	   (lambda (p c)
		(scan-token p)))

    (define (init1 ch s)
	(vector-set! *scanners* (char->integer ch) s))
	
    (define (inits l s)
	(for-each
	    (lambda (ch)
		(vector-set! *scanners* (char->integer ch) s))
	    l))

    (define (init from to s)
	(let ((limit (char->integer to)))
	    (do ((i (char->integer from) (add1 i)))
		((> i limit))
	      (vector-set! *scanners* i s)))))

;; scan a single token from an input port p

(define-method %scan-token ((p <InputPort>))
    (if (not *scanners*)
	(init-scanners))
    (let ((c (read-char p)))
	(if (eof-object? c)
	    c
	    ((vector-ref *scanners* (char->integer c)) p c))))
