/*
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
*/
#include "privio.h"
#include "bltin_p.h"
#include "scmio_p.h"

#include <ctype.h>
#include <string.h>

#define NUL ((char)0)
#define DONE_COLLECTING		(-1)		/* used by collect() */
#define COLLECTION_ERROR	(-2)		/* abort a collection */
#define CNTL(ch) ((ch)-0x40)

#define literal_token_type		SLOT(0)
#define symbol_token_type		SLOT(1)
#define number_token_type		SLOT(2)
#define open_paren_token_type		SLOT(3)
#define open_vector_token_type		SLOT(4)
#define close_paren_token_type		SLOT(5)
#define dot_token_type			SLOT(6)
#define quote_token_type		SLOT(7)
#define quasiquote_token_type		SLOT(8)
#define unquote_token_type		SLOT(9)
#define unquote_splicing_token_type	SLOT(10)
#define object_ref_type			SLOT(11)

static inline obj make_token( UINT_32 type, obj data )
{
    return make2( token_class,
    		  gvec_read(token_types,type), 
		  data );
}

static obj makebytes( UINT_32 len, obj bvec_class, const char *src )
{
obj bvec;

    bvec = bvec_alloc( len, bvec_class );
    memcpy( PTR_TO_DATAPTR(bvec), src, len );
    return bvec;
}

static void canonicalize_string( char *text )
{
    while (*text)
      {
	if (isupper(*text))
	  *text = tolower( *text );
	text++;
      }
}

static void lexical_error( const char *msg )
{
    scheme_error( "Lexical error: ~a", 1, make_string(msg) );
}

#define lexical_error0(msg) \
	scheme_error( "Lexical error: " msg, 0 )
#define lexical_error1(msg,arg) \
	scheme_error( "Lexical error: " msg, 1, arg )
#define lexical_error2(msg,arg1,arg2) \
	scheme_error( "Lexical error: " msg, 2, arg1, arg2 )

static struct character_names_s {
    const char *name;
    obj 	value;
} character_names[] = {
    { "newline", MAKE_ASCII_CHAR('\n') },
    { "tab", MAKE_ASCII_CHAR('\t') },
    { "space", MAKE_ASCII_CHAR(' ') },
    { "esc", MAKE_ASCII_CHAR(CNTL('[')) },
    { NULL }
};


static struct {
    const char	*name;
    obj		value;
} named_constants[] = {	/* these can't start with t, f, i, e, b, o, d, or x */
    { "nil", NIL_OBJ },
    { "undef", UNDEFINED_OBJ },
    { "none", NOVALUE_OBJ },
    { "uninit", UNINITIALIZED_OBJ },
    { "unbound", UNBOUND_OBJ },
    { "key", MAKE_UNIQ_OBJ(KEY_TAG) },
    { "rest", MAKE_UNIQ_OBJ(REST_TAG) },
    { NULL } };

#define MAX_LEN_CHARACTER_NAME 32

#define MAX_LEN_NUMBER 128

#define NUMBER_OF_SPECIAL_INITIALS (sizeof(special_initial) / sizeof(int))

static inline bool is_special_initial( char ch )
{
    return strchr( "!$%&*/:<=>?~_^", ch ) ? YES : NO;
}

#define IS_AT(ch)                ((ch) == '@')
#define IS_NEWLINE(ch)           ((ch) == '\n')
#define IS_SPACE(ch)             ((ch) == ' ')
#define IS_SEMICOLON(ch)         ((ch) == ';')
#define IS_OPEN_PAREN(ch)        ((ch) == '(')
#define IS_CLOSE_PAREN(ch)       ((ch) == ')')
#define IS_SHARP(ch)             ((ch) == '#')
#define IS_BACKSLASH(ch)         ((ch) == '\\')
#define IS_SLASH(ch)             ((ch) == '/')
#define IS_DOUBLE_QUOTE(ch)      ((ch) == '"')
#define IS_SINGLE_QUOTE(ch)      ((ch) == '\'')
#define IS_BACKQUOTE(ch)         ((ch) == '`')
#define IS_GREATER_THAN(ch)      ((ch) == '>')
#define IS_COMMA(ch)             ((ch) == ',')
#define IS_DOT(ch)               ((ch) == '.')
#define IS_SEMICOLON(ch)         ((ch) == ';')
#define IS_PLUS(ch)              ((ch) == '+')
#define IS_MINUS(ch)             ((ch) == '-')
#define IS_EQUAL(ch)             ((ch) == '=')
#define IS_T(ch)                 (((ch) == 't') || ((ch) == 'T'))
#define IS_F(ch)                 (((ch) == 'f') || ((ch) == 'F'))
#define IS_WHITESPACE(ch)        (IS_SPACE(ch) || IS_NEWLINE(ch))
#define IS_LETTER(ch)            isalpha(ch)
#define IS_DIGIT(ch)             isdigit(ch)
#define IS_SPECIAL_INITIAL(ch)   is_special_initial(ch)
#define IS_INITIAL(ch)           (IS_LETTER(ch) || IS_SPECIAL_INITIAL(ch))
#define IS_SPECIAL_SUBSEQUENT(ch) (IS_DOT(ch) || IS_PLUS(ch) || IS_MINUS(ch))
#define IS_SUBSEQUENT(ch)        (IS_INITIAL(ch) || IS_DIGIT(ch) \
        || IS_SPECIAL_SUBSEQUENT(ch))
#define IS_DELIMITER(ch)          (IS_WHITESPACE(ch) || IS_OPEN_PAREN(ch) \
        || IS_CLOSE_PAREN(ch) || IS_DOUBLE_QUOTE(ch) || IS_SEMICOLON(ch))
#define EOS '\0' /* end of string */
#define SHARP '#'
#define COMMA ','

static obj (*the_peek_fn)( obj );
static obj (*the_read_fn)( obj );
static obj the_port;

static inline char access_port( obj port, obj (*function)( obj ) )
{
obj scheme_ch;

    assert( EQ( port, the_port ) );
    scheme_ch = function( the_port );
    if (OBJ_ISA_ASCII_CHAR(scheme_ch))
    {
	return (char)GET_IMMEDIATE_VALUE(scheme_ch);
    }
    else
    {
	assert( EOF_OBJECT_P(scheme_ch) );
        return NUL;
    }
}

static char peek_char( obj port )
{
    return access_port( port, the_peek_fn );
}

static char get_char( obj port )
{
    return access_port( port, the_read_fn );
}

#define SHORT_COLLECT_BUF_LEN	1024

static char short_collect_buf[ SHORT_COLLECT_BUF_LEN ];

static void free_collect_buf( char *buf )
{
    if (buf != short_collect_buf)
	free( buf );
}


static bool collect( int (*next)( obj, void * ), obj port, void *state,
		     char **result, unsigned *length )
{
unsigned i, len, left;
int ch;
char *buffer;

    i = 0;
    len = left = SHORT_COLLECT_BUF_LEN;
    buffer = short_collect_buf;
    
    while ((ch = next(port,state)) != DONE_COLLECTING)
    {
	if (ch == COLLECTION_ERROR)
	{
	    free_collect_buf( buffer );
	    return NO;
	}
	buffer[i++] = ch;
	left--;
#ifndef HACK
	if (left == 0)
	{
	unsigned old_len = len, new_len = 2*len;
	
	    left = new_len - len;
	    len = new_len;
	    if (buffer == short_collect_buf)
	    {
	    char *nb;
	    
		nb = (char *)malloc( new_len );
		memcpy( nb, buffer, old_len );
		buffer = nb;
	    }
	    else
	    {
		buffer = (char *)realloc( buffer, new_len );
	    }
	}
#else
	assert( left != 0 );
#endif /* HACK */
    }
    buffer[i] = 0;
    if (length)
	*length = i;
    *result = buffer;
    return YES;
}


/*
   note: limited support for numbers at this point.  Only +/- integers.
*/

static obj number( obj p, char *prefix )
{
char text[MAX_LEN_NUMBER + 1];
int j, ch;

    j = 0;
    while (*prefix)
	text[j++] = *prefix++;

    ch = get_char(p);
    text[j++] = ch;
    while (j < MAX_LEN_NUMBER)
    {
	ch = peek_char(p);
	if (!(IS_DIGIT(ch) || ch == '.'))
	    break;
	text[j++] = get_char(p);
    }
    text[j] = NUL;
    
    /* peculiar identifiers */
    
    if (j == 1 && (text[0] == '+' || text[0] == '-'))
    {
        return make_token( symbol_token_type, lookup_symbol( text ) );
    }
    else if (strchr(text,'.'))
	return make_token( number_token_type, make_float(atof(text)) );
    else
	return make_token( number_token_type, RAWINT_TO_FIXNUM(atoi(text)) );
}

static int char_to_digit( char c )
{
    if (c >= '0' && c <= '9')
	return c - '0';
    if (c >= 'a' && c <= 'z')
	return c - 'a' + 10;
    if (c >= 'A' && c <= 'Z')
	return c - 'A' + 10;
    return -1;
}

static obj number_radix( unsigned radix, obj p )
{
int d;
UINT_32 value, max_before_shift;

    value = 0;
    max_before_shift = 0x3FFFFFFF / radix;
    
    while (1)
    {
	d = char_to_digit( peek_char(p) );
	if (d < 0 || d >= radix)	/* invalid digit */
	    break;
	if (value > max_before_shift)
	{
	    lexical_error( "Numeric constant overflow" );
	}
	value = value * radix + d;
	get_char(p);
    }
    return make_token( number_token_type, int2fx(value) );
}


static int string_ch( obj port, void *state )
{
char ch;

    ch = get_char(port);
    if (ch == '\"') /* " */
	return DONE_COLLECTING;
    if (ch == NUL)
    {
	lexical_error( "Unterminated string" );
	return COLLECTION_ERROR;
    }
    if (ch == '\\')
    {
	ch = get_char(port);
	switch (ch)
	{
	    case 'n':	return '\n';
	    case 't':	return '\t';
	    case 'e':   return CNTL('[');
	    case 'r':   return CNTL('M');
	    case 'b':   return CNTL('H');
	    case '1': case '2': case '3':
	    case '4': case '5': case '6':
	    case '7': case '8': case '9':
	    case '0':
	      { int n;

		n = ch - '0';
		ch = peek_char(port);
		if (ch >= '0' && ch <= '9')
		  {
		    get_char(port);
		    n = (n * 10) + ch - '0';
		    ch = peek_char( port );
		    if (ch >= '0' && ch <= '9')
		      {
			get_char(port);
			n = (n * 10) + ch - '0';
		      }
		  }
		return n;
	      }
	    default:	return ch;
	    case 0:	/* missing escaped char */
			lexical_error( "Missing escaped char" );
	    		return COLLECTION_ERROR;
	}
    }
    else
	return ch;
}

/*
    Having not yet eaten the initial double-quote
*/

static obj scan_string( obj port )
{
unsigned len;
char *text;
obj str;

    get_char(port);	/* skip leading quote */
    
    if (!collect( string_ch, port, NULL, &text, &len ))
	return FALSE_OBJ;

    str = makebytes( len+1, string_class, text );
    free_collect_buf( text );
    return make_token( literal_token_type, str );
}

/*
    Having not yet eaten the initial "{", consume
    a "C Text" object
*/

struct c_text_info {
    int		depth;
    bool	in_string;
};

static int c_text_ch( obj port, void *state )
{
char ch;
struct c_text_info *info = (struct c_text_info *)state;

    if (info->depth < 0)
	return DONE_COLLECTING;

    ch = get_char(port);
    if (ch == NUL)
    {
	lexical_error( "Missing C-text terminator" );
	return COLLECTION_ERROR;
    }
    if (info->in_string)
    {
	if (ch == '\"') /* " */
	    info->in_string = NO;
    }
    else if (ch == '{')
    {
	info->depth++;
    }
    else if (ch == '}')
    {
        info->depth--;
	if (info->depth < 0)
	    return DONE_COLLECTING;
    }
    else if (ch == '\"') /* " */
    {
	info->in_string = YES;
    }
    return ch;
}

static obj scan_c_text( obj port )
{
struct c_text_info state = { 0, NO };
char *text;
obj ct;
unsigned len;

    get_char( port );	/* throw away initial brace */
    
    if (!collect( c_text_ch, port, &state, &text, &len ))
	return FALSE_OBJ;
    
    ct = make1( c_text_class, makebytes( len+1, string_class, text ) );
    free_collect_buf( text );
    return make_token( literal_token_type, ct );
}


/*
    Scan an identifier
*/

static int ident_ch( obj port, void *state )
{
char ch = peek_char(port);

    if (ch != NUL && IS_SUBSEQUENT(ch))
	return get_char(port);
    return DONE_COLLECTING;
}

static obj scan_ident( obj port )
{
obj sym;
char *text;

    if (!collect( ident_ch, port, NULL, &text, NULL ))
	return FALSE_OBJ;

    sym = lookup_symbol( text );
    free_collect_buf( text );
    return make_token( symbol_token_type, sym );
}

/*
    function to get the next character in a collection,
    where we are collecting the characters in a <character-name>
    (that is, a #\x construct, where x is alphabetic)
    
    Note that upper-case chars are converted to lower case
    right here.
*/

static int character_name_ch( obj port, void *state )
{
char ch;

    ch = peek_char(port);
    if (!ch || IS_DELIMITER(ch))
	return DONE_COLLECTING;
    ch = get_char(port);
    return isupper(ch) ? tolower(ch) : ch;
}

/*
    have already scanned past "#\"
*/

static obj character( obj p )
{
char ch, *text;
unsigned i;

    ch = peek_char(p);
    if (isalpha(ch))
    {
        if (!collect( character_name_ch, p, NULL, &text, NULL ))
	    return FALSE_OBJ;
	for (i=0; character_names[i].name; i++)
	    if (strcmp( text, character_names[i].name ) == 0)
	    {
		free_collect_buf( text );
		return make_token( literal_token_type, 
				   character_names[i].value );
	    }
	if (strlen(text) == 1)
	{
	/* note that we access "ch" instead of "text[0]".  That
	   is because text has been lower-case'd to compare against
	   character names -- dmk 93.08.20 */
	obj it = make_token( literal_token_type, MAKE_ASCII_CHAR(ch) );
	
	    free_collect_buf( text );
	    return it;
	}
	{
	obj name = make_string(text);
	    free_collect_buf(text);
	    lexical_error1( "#\\~a: no character by that name", name );
	}
	return FALSE_OBJ;
    }
    else
    {
	ch = get_char(p);
	return make_token( literal_token_type, MAKE_ASCII_CHAR(ch) );
    }
}


static int character_digit_ch( obj port, void *state )
{
char ch;

    ch = peek_char(port);
    if (!ch || !IS_DIGIT(ch))
	return DONE_COLLECTING;
    return get_char(port);
}

static obj num_character( obj p )
{
char *text;
obj it;

    if (!collect( character_digit_ch, p, NULL, &text, NULL ))
	return FALSE_OBJ;
    it = make_token( literal_token_type, MAKE_ASCII_CHAR(atoi(text)) );
    free_collect_buf( text );
    return it;
}


static obj an_obj_ref( obj p )
{
char *text;
obj it;

    if (!collect( character_digit_ch, p, NULL, &text, NULL ))
	return FALSE_OBJ;
    it = make_token( object_ref_type, int2fx(atoi(text)) );
    free_collect_buf( text );
    return it;
}


/*
    scan a special object, which is something that starts
    with '#', having already eaten the '#'
*/

static obj special( obj port )
{
char ch;

    ch = get_char( port );	/* look at what comes after */
    
    switch (ch)
    {
	case 't':	return make_token( literal_token_type, TRUE_OBJ );
	case 'f':	return make_token( literal_token_type, FALSE_OBJ );
	case 'b':	/* binary number */
			return number_radix( 2, port );

	case 'o':	/* octal number */
			return number_radix( 8, port );

	case 'd':	/* decimal number */
			return number_radix( 10, port );

	case 'x':	/* hex number */
			return number_radix( 16, port );

	case 'i':	/* inexact number */
			lexical_error( "#i not supported" );
			return FALSE_OBJ;
	case 'e':	/* exact number */
			lexical_error( "#e not supported" );
			return FALSE_OBJ;
	case '*':	/* a way of referring directly to objects
			   using ``user-hash'' number (see objref.scm) */
			return an_obj_ref( port );

	case '\\':	return character(port);
	case '/':	return num_character(port);
	case '(':	return make_token( open_vector_token_type, 
					   FALSE_OBJ );
    }

    if (isalpha(ch))
    {
    char *text;
    unsigned i;
    
	if (!collect( ident_ch, port, NULL, &text, NULL ))
	    return FALSE_OBJ;
#ifndef CASE_SENSE
        canonicalize_string( text );
#endif
	for (i=0; named_constants[i].name; i++)
	    if (named_constants[i].name[0] == ch
	        && (strcmp(text, named_constants[i].name+1) == 0))
	    {
		free_collect_buf( text );
		return make_token( literal_token_type, 
				   named_constants[i].value );
	    }
	{
	obj c = make_string( text );
	
	    free_collect_buf(text);
	    lexical_error2( "#~c~a : not a valid constant", 
			    MAKE_ASCII_CHAR(ch), 
			    c );
	}
	return FALSE_OBJ;
    }
    lexical_error2( "#~c : bad token (~s can't follow #)",
    	MAKE_ASCII_CHAR(ch),
    	MAKE_ASCII_CHAR(ch) );
    return FALSE_OBJ;
}

static void nestable_comment( obj port )
{
int nesting = 1;

    /* note that there is a DFA (FSM) encoded
       in the labels and jumps of this while loop.
       All it does is look for "#|", in which case
       it increments the counter, and (simultaneously)
       for "|#", in which case it decrements the
       counter.  Note that the sequence " #|#" is taken
       to be a start-comment only, but "|#|#" is two
       end-comments.
    */
    
plain: /* in the middle of some (nesting of them) comments */
    switch (get_char(port))
    {
	case '|':	goto after_bar;
	case '#':	goto after_sharp;
	case 0:		goto premature_end;
	default:	goto plain;
    }

after_sharp:
    switch (get_char(port))
    {
	case '|':	/* found a start-comment */
			nesting++;
			goto plain;
	case '#':	goto after_sharp;	/* still after a sharp */
	case 0:		goto premature_end;
	default:	goto plain;
    }

after_bar:
    switch (get_char(port))
    {
	case '|':	goto after_bar;
	case '#':	/* found an end-comment */
			nesting--;
			if (nesting == 0)
			    return;
			goto plain;
	case 0:		goto premature_end;
	default:	goto plain;
    }

premature_end:
    lexical_error( "End of input inside #|...|# comment" );
    return;
}

obj inport_scan_token( obj port, obj (*peek)( obj ), obj (*read)( obj ) )
{
char ch;

    the_peek_fn = peek;
    the_read_fn = read;
    the_port = port;
    
again:
    ch = peek_char( port );
    
    /* end of the input stream?  */
    if (!ch)		
    {
	/* then return EOF object */
	return eof_object;	
    }

    else if (isspace(ch))
    {
	get_char(port);
    	goto again;
    }

    else if (isdigit(ch) || ch == '+' || ch == '-')
    	return number( port, "" );

    else if (ch == ';')
    {
	while (ch && ch != '\n')
	    ch = get_char(port);
	goto again;
    }
    
    else if (ch == '\"') /* " */
	return scan_string( port );
	
    else if (ch == '{')
        return scan_c_text( port );
    
    else if (ch == '#')
    {
	get_char( port );		/* eat it */
	ch = peek_char( port );
	
	/* handle long nestable comments specially */
	if (ch == '|')
	{
	    get_char(port);		/* eat initial '|' */
	    nestable_comment( port );
	    goto again;
	}
	
	return special(port);
    }
    
    else if (IS_INITIAL(ch))
	return scan_ident(port);

    else if (ch == '.')
    {
	get_char(port);
	ch = peek_char(port);
	if (isdigit(ch))
	    return number( port, "." );
	else if (ch == '.')
	{
	    get_char(port);		/* eat the 2nd '.' */
	    ch = get_char(port);	/* get the 3rd */
	    if (ch != '.')
	    {
		lexical_error( "token starts with '..' but is not '...'" );
	    }
	    return make_token( symbol_token_type, lookup_symbol( "..." ) );
	}
    	return make_token( dot_token_type, FALSE_OBJ );
    }

    else if (ch == '(')
    {
	get_char(port);
    	return make_token( open_paren_token_type, FALSE_OBJ );
    }
    
    else if (ch == ')')
    {
	get_char(port);
    	return make_token( close_paren_token_type, FALSE_OBJ );
    }
    
    else if (ch == '\'')
    {
	get_char(port);
    	return make_token( quote_token_type, FALSE_OBJ );
    }

    else if (ch == '`')
    {
	get_char(port);
    	return make_token( quasiquote_token_type, FALSE_OBJ );
    }

    else if (ch == ',')
    {
	get_char(port);
	ch = peek_char(port);
	if (ch == '@')
	{
	    get_char( port );	/* eat the @ */
	    return make_token( unquote_splicing_token_type, FALSE_OBJ );
	}
    	return make_token( unquote_token_type, FALSE_OBJ );
    }
	
    /* unrecognized start-of-token */

    ch = get_char(port);
    lexical_error1( "~s: Bad start of token", MAKE_ASCII_CHAR(ch) );
    return MAKE_ASCII_CHAR(ch);
}


