#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
#|
(define-generic-fn name)
(define-generic-fn set-name!)

(define-generic-fn state)
(define-generic-fn return-values)

(define-generic-fn set-state!)
(define-generic-fn set-return-values!)
|#

(define-class <Thread> (<Object>)
    thread-state
    thread-return-values
    thread-name
    (suspend-count 0))

(define (thread? thing)
    (instance? thing <thread>))

(define-method write ((self <Thread>) port)
    (format port "#[Thread ~a]" (thread-name self)))
    
(define-method display ((self <Thread>) port)
    (format port "#[Thread ~a]" (thread-name self)))
    
(define *ready-queue* '())
(define *current-thread* '())
(define *time-slice* 50)

(define (get-current-thread)
    *current-thread*)

(define (get-time-slice)
    *time-slice*)
    
(define (set-time-slice! x)
    (set! *time-slice* x))

(define (with-interrupts-disabled thunk)
    (%disable-interrupts)
    (let ((result (thunk)))
	(%reenable-interrupts)
	result))

;; begin-suspend executes a critical section which ends
;; with the thread suspended.  after the body finishes,
;; the thread context is saved and the thread removed
;; from the ready queue

(define-syntax (begin-suspend . body)
    (begin
	(%disable-interrupts)
	(begin . body)
	(%save-and-suspend-current-thread)
	(values)))

(define-syntax (begin-critical . body)
    (begin
	(%disable-interrupts)
	(begin . body)
	(%reenable-interrupts)
	(values)))

(define (get-ready-threads)
    (if (null? *ready-queue*)
	'()
	(with-interrupts-disabled
	    (lambda ()
		(let loop ((i (cdr *ready-queue*)) 
			   (r (list (car *ready-queue*))))
		    (if (eq? i *ready-queue*)
			(reverse r)
			(loop (cdr i) (cons (car i) r))))))))
	    	
#|
   this threads system is based on Lightweight Continuations,
   which are just <PartialContinuation> objects -- ie, they are
   not wrapped up into closures.
   
   Hence, you have to invoke them using special primitives
   (ie, %jump-with-timeout)
|#
   
;; when this function is called, the previously executing thread
;; has already suspended execution (and is no longer in the ready queue)

(define (%dispatch-next-thread)
    (set! *current-thread* (car *ready-queue*))
    (if (eq? (suspend-count *current-thread*) 0)
	(%jump-with-timeout (thread-state *current-thread*) *time-slice*)
	(begin
	    ;; delete this thread from the ready queue --
	    ;; it's suspended
	    (set-car! *ready-queue* (cadr *ready-queue*))
	    (set-cdr! *ready-queue* (cddr *ready-queue*))
	    (%dispatch-next-thread))))

(define (%threads-did-unblock queue threads)
    (for-each
	%mark-thread-ready
	threads)
    (values))	; want enqueue! to return nothing,
    		; and it tail-calls us when a queue unblocks

(define (make-thread thunk)
    (let ((t (make <Thread> thread-return-values: #f)))
	(call-with-current-continuation
	    (lambda (exit)
		(%call-with-lightweight-continuation
		    (lambda (state)
			(set-thread-state! t state)
			(exit t)))
		(bind ((#rest result (thunk)))
		    (set-thread-return-values! t result)
		    (format #t "~s finishes with: ~s\n" t result)
		    (%suspend-current-thread))))))

(define (%thread-yield)
    (%timeout)
    (values))

(define *sleep-list* '())

(define (idle)
    (with-interrupts-disabled
	(lambda ()
	    ;; check for any threads being ready now
	    (let ((now (time-now)))
		(let loop ((i *sleep-list*) (prev #f))
		    (if (not (null? i))
			(if (time>=? now (caar i))
			    ;; the (car i) thing has expired
			    (begin
				(%mark-thread-ready (cdar i))
				(if prev
				    (set-cdr! prev (cdr i))
				    (set! *sleep-list* (cdr i)))
				(loop (cdr i) prev))
			    ;; the (car i) thing is still sleeping
			    (loop (cdr i) i)))))))
    ;; if, after all that, nothing is ready,
    ;; then do an OS-level sleep for 50ms
    (if (eq? *ready-queue* (cdr *ready-queue*))
	(sleep 0.05))
    (%thread-yield)
    (idle))

(define (suspend-thread thread)
    (type-check thread thread? <thread>)
    (if (eq? thread *current-thread*)
	(begin-suspend
	    (set-suspend-count! thread 1))
	(begin-critical
	    (set-suspend-count! thread (+ (suspend-count thread) 1)))))

(define (resume-thread thread)
    (type-check thread thread? <thread>)
    (case (suspend-count thread)
	((0) (abort 'resume-thread "thread not suspended: ~s" thread))
	((1) (set-suspend-count! thread 0)
	     (%mark-thread-ready)
	     0)
	(else
	     (set-suspend-count! thread (- (suspend-count thread) 1))
	     (suspend-count thread))))
    
;; time is an interval.  use seconds->interval to convert from a number

(define (thread-sleep interval)
    (if (number? interval)
	(set! interval (seconds->interval interval))
	(type-check interval interval? <interval>))
    (let ((time-event (cons (time+ (time-now) interval) *current-thread*)))
      (begin-suspend
       (set! *sleep-list* 
	   (cons time-event *sleep-list*)))))

(define ($init-threads)
    (let ((idle (make-thread idle)))
	(set! *ready-queue* (cons idle '()))
	(set-cdr! *ready-queue* *ready-queue*)
	(set! *sleep-list* '())
	(set-thread-name! idle "idle")))

(define (thread-done? t)
    (if (thread-return-values t)
    	#t
	#f))

; It would be nice for this interface to support the notion
; of returning whenever all the threads finish, but it's not
; clear how to implement that (esp. if threads spawn other
; threads, possibly detaching them)

(define ($run-threads . threads)
    ($init-threads)
    (for-each %mark-thread-ready threads)
    (vector-set! exception-vector 3 %timeout)
    (%dispatch-next-thread))

(define (show-threads)
   (for-each
     (lambda (t)
           (format #t "#*~h ~s\n" t t))
     (get-ready-threads)))


