#|
Copyright (C) 1994  Donovan Kolbly, Paul Wilson

This file is part of RScheme (v0.5.0, 94.12.31)

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

The author may be reached at Donovan.Kolbly@TKG.com
or dkolbly@cs.utexas.edu or c/o wilson@cs.utexas.edu
|#
;;$Id: paths.scm,v 1.2 1994/11/10 01:46:04 donovan Exp $

;============================= pathname.scm =============================

(define-class <file-name> (<Object>)
    directory
    filename
    extension)

(define-class <directory-name> (<Object>)
    rooted-at
    steps)

(define-class <root-dir> (<Object>)
    root-name)

(define (root-dir? x)
    (isa? x <root-dir>))
    
(define $system-root #f)
(define $user-root #f)

(define (init-root-paths)
    (set! $system-root (make <root-dir> root-name: "/"))
    (set! $user-root (make <root-dir> root-name: "~/")))

(run-named init-root-paths)
    
(define-generic-fn pathname->string)

(define (open-input-path* p)
    (open-input-file* (pathname->string p)))

(define (open-input-path p)
    (open-input-file (pathname->string p)))

(define (open-output-path* p)
    (open-output-file* (pathname->string p)))

(define (open-output-path p)
    (open-output-file (pathname->string p)))

(define (file-within-dir p)
    (if (extension p)
	(string-append (filename p) "." (extension p))
	(filename p)))

(define-method pathname->string ((p <file-name>))
    (string-append 
	(if (directory p)
	    (pathname->string (directory p))
	    "")
	(filename p)
	(if (extension p)
	    (string-append "." (extension p))
	    "")))

(define-method pathname->string ((self <root-dir>))
    (root-name self))

(define-method pathname->string ((p <directory-name>))
    (let loop ((i (steps p))
	       (r (if (rooted-at p) 
	       	      (list (pathname->string (rooted-at p)))
		      '())))
	(if (null? i)
	    (if (null? r)
		"./"
		(apply string-append (reverse r)))
	    (if (eq? (car i) 'up)
		(loop (cdr i) (cons "/" (cons ".." r)))
		(loop (cdr i) (cons "/" (cons (car i) r)))))))


(define-method pathname->string ((self <String>))
    self)

(define-method write ((self <file-name>) port)
    (format port "#[<file> ~a]" (pathname->string self)))
    
(define-method write ((self <root-dir>) port)
    (format port "#[<root> ~a]" (root-name self)))

(define-method display ((thing <file-name>) port)
    (display (pathname->string thing) port))
    
(define-method write ((thing <directory-name>) port)
    (format port "#[<dir> ~a]" (pathname->string thing)))
    
(define-method display ((thing <directory-name>) port)
    (display (pathname->string thing) port))


; note that "/foo" relative to "../bleen" gives "/foo"
; (which is what one should expect)

(define (append-dirs p1 p2)
    (if (rooted-at p2)		;; if p2 is rooted, then forget relativity
	p2
	(let loop ((i (append (steps p1) (steps p2))) 
		   (r '()))
	    (if (null? i)
		(make <directory-name> rooted-at: (rooted-at p1) 
				       steps: (reverse r))
		(case (car i)
		    ((up) (if (and (pair? r)
				   (not (eq? (car r) 'up)))
			      (loop (cdr i) (cdr r))
			      (loop (cdr i) (cons 'up r))))
		    (else (loop (cdr i) (cons (car i) r))))))))


(define (append-path dir path)
    (make <file-name> 
	 filename: (filename path)
	 extension: (extension path)
	 directory: (if (directory path)
			(append-dirs dir (directory path))
			dir)))

(define (string->file str)
    (let* ((parsed (parse-fname str))
	   (s (simplify parsed)))
	(make <file-name> 
	    filename: (cadr s)
	    extension: (caddr s)
	    directory: (if (null? (cddr parsed))
			   #f
			   (make <directory-name> 
	    			rooted-at: (caar s)
				steps: (cdar s))))))

(define (string->dir str)
    (let ((s (simplify (parse-fname (string-append str "/")))))
	(make <directory-name> 
		rooted-at: (caar s)
		steps: (cdar s))))

; returns the NAME of the last directory in the given directory
; or #f if the last directory is not named (ie, it is "..")
;   for example:
;	(last-directory (string->dir "/tmp/foo")) => "foo"
;	(last-directory (string->dir "../bleen/foo")) => "foo"
;	(last-directory (string->dir "../..")) => #f

(define (last-directory dir)
    (let ((p (steps dir)))
	(if (null? p)
	    #f
	    (let ((l (last p)))
		(if (string? l)
		    l
		    #f)))))

#|
./foo/../bar => ((self "foo" up) "bar" #f)
("." "foo" "..") "bar"


/foo/../tmp/x.y => ((root "foo" up "tmp") "x" "y")
("" "foo" ".." "tmp") "x.y"

foo/bar/ => (("foo" "bar") #f #f)
("foo" "bar" "")

../foo/bar.scm => ((up "foo") "bar" "scm")
(".." "foo") "bar.scm"

|#

(define (parse-tail* tail-chars)
    (let loop ((i tail-chars) (r '()))
	(if (pair? i)
	    (if (eq? (car i) #\.)
		(list (list->string (reverse (cdr i)))
		      (list->string r))
		(loop (cdr i) (cons (car i) r)))
	    (list (list->string r) #f))))

(define (parse-dirs* path-chars)
    (let loop ((i path-chars) (r '()))
	(if (pair? i)
	    (if (eq? (car i) #\/)
		(cons (list->string (reverse r)) (parse-dirs* (cdr i)))
		(loop (cdr i) (cons (car i) r)))
	    (parse-tail* r))))

(define (parse-fname fname)
    (parse-dirs* (string->list fname)))

; simplify takes the output from parse-fname and:
    ; removes "."
    ; removes ""
    ; removes "something/.."
    ; and replaces ".." with up
    ; and leading "" with $system-root
    ; and leading "~" with $user-root
    ; and breaks into ((path ...) file ext) format
    ; and makes sure it starts with either #f or a <root-dir>

(define (simplify pathname)
    (let loop ((i pathname) (r '()))
	(if (null? (cddr i))
	    (let ((dirs (reverse r)))
		(if (and (pair? dirs) (root-dir? (car dirs)))
		    (cons dirs i)
		    (cons (cons #f dirs) i)))
	    (cond 
		((equal? (car i) "..")
	    		(if (and (pair? r) (string? (car r)))
			    (loop (cdr i) (cdr r))
			    (loop (cdr i) (cons 'up r))))
		((equal? (car i) ".")
			(loop (cdr i) r))
		((equal? (car i) "")
			(if (eq? i pathname)
			    (loop (cdr i) (list $system-root))
			    (loop (cdr i) r)))
		((equal? (car i) "~")
			(if (eq? i pathname)
			    (loop (cdr i) (list $user-root))
			    (loop (cdr i) r)))
		(else (loop (cdr i) (cons (car i) r)))))))

(define (make-try-extensions exit extensions fn)
    (let ((basename (filename fn))
    	  (ext (extension fn)))
	(if ext
	    ; an extension was specified, so try it only
	    (lambda (dir)
		(let* ((try (make <file-name>
				filename: basename
				extension: ext
				directory: dir))
		       (f (open-input-path* try)))
		    (if f (exit (cons f try))
		          #f)))
	    
	    ; no extension was specified, so return a closure
	    ; that will try each of the extensions in the given
	    ; list (in addition to no extension)
	    
	    (let ((try-extensions (cons #f extensions)))
		(lambda (dir)
		    (for-each
			    ; try a particular extension
			(lambda (ext)
			    (let* ((try (make <file-name>
						filename: basename
						extension: ext
						directory: dir))
				    (f (open-input-path* try)))
				(if f
				    (exit (cons f try))
				    #f)))
			try-extensions))))))
	    
(define (locate current-dir search-path extensions filename)
    (call-with-current-continuation
	(lambda (exit)
	    (let* ((fn (string->file filename)))
		(for-each
		
		    ; look in a particular directory for the file
		    ; (make-try-extensions returns a procedure of
		    ;  one argument that will search the directory
		    ;  which is the argument)
		    
		    (make-try-extensions exit extensions fn)

		    ; decide what directories to look in
		    (if (directory fn)
			; there IS a path specification, so interpret it
			; relative to the current-dir
			(list (append-dirs current-dir (directory fn)))
			; there is no dir specification, so use the search
			; path starting with the current dir.
			(cons current-dir search-path)))))))

(define (current-directory)
    (string->dir (getwd)))

;; eliminate a common prefix of two directories
;;
;;  For example:
;;      d1 = /tmp/foo_dir/
;;      d2 = /tmp/quux_dir/
;;  Compute how to get from d1 to d2
;;  (dir-from-to d1 d2) ==> ../quux_dir/

;;	To get from		To			    Use
;;	-----------		----------------------	    -----------
;;      /tmp/foo_dir		/tmp/foo_dir/blech_dir	==> blech_dir/
;;      /tmp/foo_dir/blech_dir	/tmp/foo_dir		==> ../
;;	../blah/foo/		../blah/quux/		==> ../quux
;;	/tmp/foo_dir		../blah			==> $(WD/..)/blah
;;	./			../foo			==> ../foo
;;	blech_dir/		./			==> ../
;;	blech_dir/yy		foo_dir/xx		==> ../../

(define (dir-from-to d1 d2)
    (if (and (rooted-at d2) (not (rooted-at d1)))
	d2
	(if (and (rooted-at d1) (not (rooted-at d2)))
	    (dir-from-to d1 (append-dirs (current-directory) d2))
	    (if (and (rooted-at d1) 
		     (rooted-at d2) 
		     (not (equal? (car (steps d1)) 
		     	          (car (steps d2)))))
	    	d2
		(let loop ((from (steps d1)) (to (steps d2)))
		    (if (null? from)
			(make <directory-name> steps: to)
			(if (and (pair? to) 
				    (equal? (car from) (car to)))
			    (loop (cdr from) (cdr to))
			    (make <directory-name> 
				steps: (append (map (lambda (x) 
							'up) 
						    from) 
						to)))))))))
						
;; rather crude for now...

(define (path=? x y)
    (string=? (pathname->string x) (pathname->string y)))

(export <file-name>
	<directory-name>
	<root-dir>
	string->file
	string->dir
	current-directory
	pathname->string
	path=?
	locate
	dir-from-to
	file-within-dir
	open-output-path* open-output-path 
	open-input-path* open-input-path
	append-dirs append-path last-directory
	filename extension directory)

